/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Timer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Interceptor;
import org.ow2.petals.component.framework.api.PetalsComponent;
import org.ow2.petals.component.framework.api.Wsdl;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.util.PropertiesUtil;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.ComponentInterceptor;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablelong;
import org.ow2.petals.component.framework.mbean.InterceptorManager;
import org.ow2.petals.component.framework.mbean.MBeanHelper;
import org.ow2.petals.component.framework.mbean.MBeanImplementor;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.component.framework.process.JBIAcceptorManager;
import org.ow2.petals.component.framework.process.JBIProcessorManager;
import org.ow2.petals.component.framework.process.async.AsyncMessageManager;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.ow2.petals.component.framework.util.WSDLUtilImpl;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;

/**
 * An abstract class that can be used to easily create a new JBI Component.
 * <p>
 * It implements the Component interface.<br />
 * It allow to manage directly endpoints.<br />
 * To do this, simply put into the resources of the project a valid wsdl named
 * 'component.wsdl'
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 */
public abstract class AbstractComponent implements PetalsComponent, MBeanImplementor {

    public static final String METHOD_RELOAD_PLACEHOLDERS = "reloadPlaceHolders";

    /**
     * Default value of the sampling duration used by response time probe.
     */
    private static final long DEFAULT_MONITORING_SAMPLING_PERIOD = 300;

    /**
     * The custom place holders read from a property file to be able to configure service units.
     */
    private final Properties customPlaceHolders = new Properties();

    /**
     * The Component's Context
     */
    protected ComponentContext context;

    /**
     * The message exchange acceptor manager
     */
    private JBIAcceptorManager acceptorManager;

    /**
     * The asynchronous message handler
     */
    private AsyncMessageManager asyncManager;

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    /**
     * The component specific configuration in extensions form.
     */
    private ConfigurationExtensions componentExtensions;

    /**
     * The list of native service provider endpoints, managed directly by the component
     */
    private final Map<ServiceEndpointKey, ServiceEndpoint> nativeServiceEndpoints = new HashMap<ServiceEndpointKey, ServiceEndpoint>();

    /**
     * Manage the interceptors at runtime, exposed as MBean
     */
    private InterceptorManager interceptorManager;

    /**
     * All the component interceptors. Key is the name of the interceptor
     */
    private Map<String, Interceptor> interceptors;

    /**
     * The component configuration
     */
    private Jbi jbiComponentConfiguration;

    /**
     * The component's logger
     */
    private Logger logger;

    /**
     * The description and the document of the native WSDL managed internally by
     * the component
     */
    private Wsdl nativeWsdl;

    /**
     * The processor manager in charge of handling JBI Message Exchanges
     */
    private JBIProcessorManager processorManager;

    /**
     * Manage runtime configuration notifications, exposed as MBean
     */
    private RuntimeConfigurationNotifier runtimeConfigurationNotifier;

    /**
     * The component's service unit manager
     */
    private AbstractServiceUnitManager suManager;

    /**
     * The timer used as sampler by response time probes
     */
    private Timer probesTimer;

    /**
     * The monitoring MBean
     */
    private Monitoring monitoringMbean;

    /**
     * The object name of the monitoring MBean
     */
    private ObjectName monitoringMbeanName;

    public AsyncMessageManager getAsyncManager() {
        return this.asyncManager;
    }

    @Override
    public DeliveryChannel getChannel() {
        return this.channel;
    }

    public Jbi getJbiComponentDescriptor() {
        return this.jbiComponentConfiguration;
    }

    public ConfigurationExtensions getComponentExtensions() {
        return this.componentExtensions;
    }

    @Override
    public ComponentContext getContext() {
        return this.context;
    }

    @Override
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    @Override
    public Interceptor getInterceptor(final String name) {
        Interceptor interceptor = null;

        if (this.interceptors != null) {
            interceptor = this.interceptors.get(name);
        }
        return interceptor;
    }

    @Override
    public Map<String, Interceptor> getInterceptors() {
        return this.interceptors;
    }

    @Override
    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    @Override
    public final Logger getLogger() {
        return this.logger;
    }

    /**
     * Get the nativeWsdl
     * 
     * @return the nativeWsdl
     */
    public Wsdl getNativeWsdl() {
        return this.nativeWsdl;
    }

    /**
     * Even though calls to this method can happen concurrently, there is no reason they arrive when the collections
     * read in it (i.e., nativeServiceEndpoints) are modified, so no need for it to be a concurrent map
     */
    @Override
    public Document getServiceDescription(final ServiceEndpoint endpoint) {
        if (endpoint == null) {
            return null;
        }

        final ServiceEndpointKey key = new ServiceEndpointKey(endpoint);

        if (this.nativeServiceEndpoints.containsKey(key)
                && this.nativeWsdl != null
                && WSDLUtilImpl.isDescriptionContaining(this.nativeWsdl.getDescription(), endpoint)) {
            return this.nativeWsdl.getDocument();
        }

        return this.suManager.getServiceDescription(key);
    }

    @Override
    public AbstractServiceUnitManager getServiceUnitManager() {
        return this.suManager;
    }

    @Override
    public final void init(final ComponentContext compContext) throws JBIException {
        this.context = compContext;
        this.logger = this.context.getLogger("", this.getResourceBundleName());
        this.logger.info("Init component...");
        this.channel = this.context.getDeliveryChannel();
        // Get the component configuration
        final File jbiXmlFile = new File(compContext.getInstallRoot() + File.separator
                + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);
        try {
            // CDK configuration
            try (final FileInputStream jbiInputStream = new FileInputStream(jbiXmlFile)) {
                this.jbiComponentConfiguration = CDKJBIDescriptorBuilder.getInstance()
                        .buildJavaJBIDescriptor(jbiInputStream);
            } catch (final FileNotFoundException e) {
                throw new JBIException(e);
            } catch (final IOException e) {
                this.logger.log(Level.WARNING, "Error while closing jbi.xml file", e);
            }

            // Read place holders (content of the property file used as variable substitutions to configure service
            // units)
            this.loadPlaceHolders();

            // Component specific configuration
            final String propertiesFile = this.jbiComponentConfiguration.getComponent().getPropertiesFile().getValue();
            if (!StringHelper.isNullOrEmpty(propertiesFile)) {
                // Load configuration extensions of the component
                this.componentExtensions = new ConfigurationExtensions(
                        this.jbiComponentConfiguration.getComponent().getAny(), propertiesFile);
            } else {
                this.componentExtensions = new ConfigurationExtensions(
                        this.jbiComponentConfiguration.getComponent().getAny());
            }

            this.runtimeConfigurationNotifier = new RuntimeConfigurationNotifier(this, jbiXmlFile, this.logger);
            this.loadInterceptors();
            this.interceptorManager = new InterceptorManager(this, jbiXmlFile);

            // Initialize the response time probe sampler. MUST be done before
            // the creation of the monitoring MBean that requires it.
            this.probesTimer = new Timer(this.context.getComponentName() + " - Monitoring sampler",
                    true);

            // create monitoring mbean and probes
            try {
                this.monitoringMbean = this.createMonitoringMBean();
            } catch (final MultipleProbesFactoriesFoundException e1) {
                this.getLogger().log(Level.WARNING,
                        "Technical monitoring deactivated because of the following error", e1);
            } catch (final NoProbesFactoryFoundException e1) {
                this.getLogger().log(Level.WARNING,
                        "Technical monitoring deactivated because of the following error", e1);
            }

            // Necessary resources must be instantiated only once during the
            // life-cycle of the component
            if (this.suManager == null) {
                this.suManager = this.createServiceUnitManager();
            }
            final File wsdlFile = new File(compContext.getWorkspaceRoot() + File.separator
                    + Constants.Component.NATIVE_SERVICE_FOLDER + File.separator
                    + Constants.Component.COMPONENT_WSDL);
            if (wsdlFile.exists()) {
                final Description description = WSDLUtilImpl.createWsdlDescription(
                        wsdlFile.toURI().toURL());
                this.nativeWsdl = new ComponentWsdl(description);
            }

        } catch (final JBIDescriptorException e) {
            throw new JBIException(e);
        } catch (final WSDLException e) {
            throw new JBIException(e);
        } catch (final MalformedURLException e) {
            throw new JBIException(e);
        }
        
        // until here, if something failed before, it's not really a problem in term of resources allocated (see
        // shutdown)

        boolean monitoring = false;
        boolean monitoringMBean = false;
        try {
            this.logger.fine("Calling initialization method on component implementation...");
            this.doInit();
            this.logger.fine("Specific initialization done.");

            // initialize the native service endpoint list managed by the component
            this.activateNativeServiceEndpoints();

            try {
                monitoring = true;
                this.monitoringMbean.init();
            } catch (final ProbeException e) {
                this.getLogger().log(Level.WARNING, "Error initializing probes", e);
            }

            // Since the worker manager instantiates the JBIListeners, we must
            // initialize it after the component specific initialization. This is
            // because the JBIListener implementation can access to the specific
            // component implementation which is done in the previous doInit()
            // call.
            this.processorManager = new JBIProcessorManager(this, this.runtimeConfigurationNotifier, this.logger);
            this.acceptorManager = new JBIAcceptorManager(this, this.processorManager,
                    this.runtimeConfigurationNotifier, this.logger);
            this.asyncManager = new AsyncMessageManager(this, this.processorManager, this.logger);

            // Registers the monitoring MBean
            this.monitoringMbeanName = this.context.getMBeanNames()
                    .createCustomComponentMBeanName("monitoring_" + this.context.getComponentName());

            monitoringMBean = true;
            MBeanHelper.registerMBean(this.context.getMBeanServer(), this.monitoringMbeanName, this.monitoringMbean);

            this.logger.info("Component initialized");
        } catch (final Exception e) {

            if (monitoringMBean) {
                try {
                    MBeanHelper.unregisterMBean(this.context.getMBeanServer(), this.monitoringMbeanName);
                } catch (final Exception ex) {
                    this.logger.log(Level.WARNING,
                            "Error unregistering the component MBean during revert from failed init", e);
                }
            }

            if (monitoring) {
                try {
                    this.monitoringMbean.shutdown();
                } catch (final ProbeException ex) {
                    this.getLogger().log(Level.WARNING, "Error shutting down probes during revert from failed init", e);
                }
            }

            this.deactivateNativeServiceEndpoints();

            try {
                doShutdown();
            } catch (final Exception ex) {
                this.getLogger().log(Level.WARNING, "Error shutting down component during revert from failed init", e);
            }

            throw e;
        }
    }

    /**
     * Override to add more runtime MBean operations: do not forget to reuse the collection!
     */
    @Override
    public Collection<String> getMBeanOperationsNames() {
        final Collection<String> methods = new HashSet<String>();
        methods.add(METHOD_RELOAD_PLACEHOLDERS);
        return methods;
    }

    /**
     * Override to add more runtime MBean attributes
     */
    @Override
    public Collection<String> getMBeanAttributesNames() {
        return new HashSet<>();
    }

    public void reloadPlaceHolders() {

        this.logger.info("Placeholders reloading.");

        loadPlaceHolders();

        final AbstractServiceUnitManager sum = getServiceUnitManager();
        if (sum != null) {
            sum.reloadPlaceHolders();
        }

        this.logger.info("Placeholders reloaded.");
    }

    private void loadPlaceHolders() {
        this.customPlaceHolders.clear();
        final String propertiesFile = this.jbiComponentConfiguration.getComponent().getPropertiesFile().getValue();
        if (!StringHelper.isNullOrEmpty(propertiesFile)) {
            try {
                this.customPlaceHolders.putAll(PropertiesUtil.loadProperties(propertiesFile));
                if (this.logger.isLoggable(Level.FINE)) {
                    this.logger.fine("Loaded placeholders: " + this.customPlaceHolders);
                }
            } catch (final PropertiesException e) {
                this.logger.log(Level.WARNING,
                        "Unable to read place holders values. Fix the problem before deploying service units", e);
            }
        }
    }

    /**
     * Create the monitoring Mbean
     * 
     * @return The monitoring Mbean
     * @throws NoProbesFactoryFoundException
     *             No probe implementation found in the classloader
     * @throws MultipleProbesFactoriesFoundException
     *             Several probe implementations found in the classloader
     */
    protected Monitoring createMonitoringMBean() throws MultipleProbesFactoriesFoundException,
            NoProbesFactoryFoundException {

        return new Monitoring(this.getProbesTimer(), this.getResponseTimeProbeSamplePeriod());
    }

    @Override
    public boolean isExchangeWithConsumerOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    @Override
    public boolean isExchangeWithProviderOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    @Override
    public ServiceEndpoint resolveEndpointReference(final DocumentFragment epr) {
        return null;
    }

    @Override
    public final void shutDown() throws JBIException {

        this.logger.info("Shut down component...");

        try {
            final PEtALSCDKException exception = new PEtALSCDKException("Error stopping the component");

            try {
                MBeanHelper.unregisterMBean(this.context.getMBeanServer(), this.monitoringMbeanName);
            } catch (final Exception e) {
                this.logger.log(Level.WARNING, "Error unregistering the component MBean", e);
            }

            try {
                this.processorManager.shutdown();
            } catch (final Exception e) {
                this.logger.log(Level.WARNING, "Error shutting down the processor manager", e);
            }

            try {
                this.acceptorManager.shutdown();
            } catch (final Exception e) {
                this.logger.log(Level.WARNING, "Error shutting down the acceptor manager", e);
            }

            try {
                this.asyncManager.shutdown();
            } catch (final Exception e) {
                this.logger.log(Level.WARNING, "Error shutting down the async manager", e);
            }

            // Shutdown the probes
            try {
                this.monitoringMbean.shutdown();
            } catch (final ProbeException e) {
                this.getLogger().log(Level.WARNING, "Error shutting down component probes", e);
            }

            // Deregister native service endpoints
            this.deactivateNativeServiceEndpoints();

            this.logger.fine("Calling specific shut down...");
            try {
                this.doShutdown();
                this.logger.fine("Specific shut down done.");
            } catch (final Exception e) {
                this.getLogger().log(Level.WARNING, "Error shutting down component", e);
            }

            // Free the response time probe sampler
            this.probesTimer.cancel();

            exception.throwIfNeeded();
        } finally {
            // remove unnecessary resources
            this.nativeWsdl = null;
            this.channel = null;
            this.runtimeConfigurationNotifier = null;
            this.interceptors = null;
            this.interceptorManager = null;
            this.acceptorManager = null;
            this.processorManager = null;
            this.asyncManager = null;
            CDKJBIDescriptorBuilder.destroyInstance();
        }

        this.logger.info("Component shut down");
    }

    @Override
    public final void start() throws JBIException {
        this.logger.info("Start component...");

        boolean interceptor = false;
        boolean processor = false;
        boolean acceptor = false;
        boolean async = false;
        boolean implementation = false;
        try {
            this.registerRuntimeConfigMBean();

            interceptor = true;
            this.registerInterceptorManagerMBean();

            processor = true;
            this.processorManager.start();

            acceptor = true;
            this.acceptorManager.start();

            async = true;
            this.asyncManager.start();

            this.logger.fine("Calling specific start...");
            implementation = true;
            this.doStart();
            this.logger.fine("Specific start done.");

            // Start the probes. Probes are started in last position to have all
            // components already started
            try {
                this.monitoringMbean.start();
            } catch (final ProbeException e) {
                this.getLogger().log(Level.WARNING, "Error starting probes", e);
            }
        } catch (final Exception e) {
            if (implementation) {
                try {
                    this.doStop();
                } catch (final Exception ex) {
                    this.getLogger().log(Level.WARNING, "Error stopping the component during revert from failed start",
                            ex);
                }
            }

            if (acceptor) {
                try {
                    this.acceptorManager.stop();
                } catch (final Exception ex) {
                    this.getLogger().log(Level.WARNING,
                            "Error stopping the acceptor manager during revert from failed start", ex);
                }
            }

            if (processor) {
                try {
                    this.processorManager.stop();
                } catch (final Exception ex) {
                    this.getLogger().log(Level.WARNING,
                            "Error stopping the processor manager during revert from failed start", ex);
                }
            }

            if (async) {
                try {
                    this.asyncManager.stop();
                } catch (final Exception ex) {
                    this.getLogger().log(Level.WARNING,
                            "Error stopping the async manager during revert from failed start", ex);
                }
            }

            if (interceptor) {
                try {
                    this.unregisterInterceptorManagerMBean();
                } catch (final Exception ex) {
                    this.getLogger().log(Level.WARNING,
                            "Error unregistering the interceptor manager mbean during revert from failed start", ex);
                }
            }

            try {
                this.unregisterRuntimeConfigurationMBean();
            } catch (final Exception ex) {
                this.getLogger().log(Level.WARNING,
                        "Error unregistering the runtime configuration mbean during revert from failed start", ex);
            }

            throw e;
        }

        this.logger.info("Component started");
    }

    @Override
    public final void stop() throws JBIException {
        this.logger.info("Stop component...");

        final PEtALSCDKException exception = new PEtALSCDKException("Error stopping the component");

        // Stop the probes. Probes are stopped in first position to have all
        // components always running if needed
        try {
            this.monitoringMbean.stop();
        } catch (final ProbeException e) {
            exception.addSuppressed(e);
        }

        this.logger.fine("Calling specific stop...");
        try {
            this.doStop();
        } catch (final Exception e) {
            exception.addSuppressed(e);
        }
        this.logger.fine("Specific stop done.");

        // stop first the acceptor Manager as it aggregates the worker manager
        try {
            this.acceptorManager.stop();
        } catch (final Exception e) {
            exception.addSuppressed(e);
        }

        try {
            this.processorManager.stop();
        } catch (final Exception e) {
            exception.addSuppressed(e);
        }

        try {
            this.asyncManager.stop();
        } catch (final Exception e) {
            exception.addSuppressed(e);
        }

        try {
            this.unregisterInterceptorManagerMBean();
        } catch (final Exception e) {
            exception.addSuppressed(e);
        }

        try {
            this.unregisterRuntimeConfigurationMBean();
        } catch (final Exception e) {
            exception.addSuppressed(e);
        }

        exception.throwIfNeeded();

        this.logger.info("Component stopped");
    }

    /**
     * Can be overridden to provide a specific service unit manager
     * 
     * @return the service unit manager used by this component.
     */
    protected abstract AbstractServiceUnitManager createServiceUnitManager();

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doInit(). This method is called at the end of the component
     * classical initialization. When this method is called, the component's
     * delivery channel, context, logger and service unit manager are available.
     * 
     */
    protected void doInit() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doShutdown(). This method is called at the end of the
     * component classical shutdown.
     * 
     */
    protected void doShutdown() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStart(). This method is called at the end of the component
     * classical start.
     * 
     */
    protected void doStart() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStop(). This method is called at the end of the component
     * classical stop.
     * 
     */
    protected void doStop() throws JBIException {
    }

    /**
     * Can be overridden to return a custom resource bundle name
     * 
     * @return the resource bundle name or null if no resource bundle name has
     *         to be used (default value)
     */
    protected String getResourceBundleName() {
        return null;
    }

    /**
     * <p>
     * Activate an endpoint of the native service.
     * </p>
     * <p>
     * <b>Note</b>: A native service endpoint is an endpoint of the native service described by the native WSDL
     * available into the component resources.</b>
     * 
     * @param serviceName
     *            The name of the native service.
     * @param endpointName
     *            The name of native service endpoint to activate
     * @throws JBIException
     */
    private ServiceEndpoint activateNativeServiceEndpoint(final QName serviceName, final String endpointName)
            throws JBIException {

        // Fake endpoint for method Component.getServiceDescription(...) invoked by
        // ComponentContext.activateEndpoint(...)
        final ServiceEndpointKey key = new ServiceEndpointKey(serviceName, endpointName);
        this.nativeServiceEndpoints.put(key, new ComponentServiceEndpoint(serviceName, endpointName));

        final ServiceEndpoint se = this.context.activateEndpoint(serviceName, endpointName);

        // After the activation, we override the fake endpoint with the real value
        this.nativeServiceEndpoints.put(key, se);
        return se;
    }

    /**
     * <p>
     * Activate the endpoints of the native service managed directly by the component, not the endpoints associated to
     * the service units.
     * </p>
     * 
     * @throws JBIException
     */
    private void activateNativeServiceEndpoints() throws JBIException {
        if (this.nativeWsdl != null) {
            for (final Endpoint endpoint : WSDLUtilImpl.getEndpointList(this.nativeWsdl.getDescription())) {
                final QName serviceName = endpoint.getService().getQName();
                final String endpointName = endpoint.getName();
                this.activateNativeServiceEndpoint(serviceName, endpointName);
            }
        }
    }

    /**
     * <p>
     * Deactivate the endpoints of the native service managed directly by the component, not the endpoints associated to
     * the service units.
     * </p>
     */
    private void deactivateNativeServiceEndpoints() {
        final Iterator<ServiceEndpoint> edpts = this.nativeServiceEndpoints.values().iterator();
        while (edpts.hasNext()) {
            // remove from container
            try {
                this.context.deactivateEndpoint(edpts.next());
            } catch (final Exception e) {
                this.getLogger().log(Level.WARNING, "Error deactivating endpoint", e);
            }
            // remove from local cache
            edpts.remove();
        }
    }

    /**
     * Load the interceptors if there is any.
     */
    private final void loadInterceptors() {

        this.interceptors = new HashMap<String, Interceptor>();

        if (this.jbiComponentConfiguration.getComponent().getComponentInterceptors() != null) {
            for (final ComponentInterceptor componentInterceptor : this.jbiComponentConfiguration
                    .getComponent().getComponentInterceptors().getInterceptor()) {
                try {
                    this.logger.info("Loading interceptor : " + componentInterceptor.getName());
                    final Class<?> clazz = Thread.currentThread().getContextClassLoader()
                            .loadClass(componentInterceptor.getClazz());
                    final Interceptor itc = (Interceptor) clazz.newInstance();

                    ConfigurationExtensions configurationExtensions;
                    final String propertiesFile = this.jbiComponentConfiguration.getComponent().getPropertiesFile()
                            .getValue();
                    if (!StringHelper.isNullOrEmpty(propertiesFile)) {
                        configurationExtensions = new ConfigurationExtensions(componentInterceptor.getParam(),
                                propertiesFile);
                    } else {
                        configurationExtensions = new ConfigurationExtensions(componentInterceptor.getParam());
                    }

                    itc.init(this, componentInterceptor.getName(), componentInterceptor.isActive(),
                            configurationExtensions, this.logger);
                    this.interceptors.put(componentInterceptor.getName(), itc);
                } catch (final PEtALSCDKException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (final ClassNotFoundException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (final InstantiationException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (final IllegalAccessException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (final ClassCastException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                }
            }
        }
    }

    /**
     * 
     * @throws JBIException
     */
    private void registerInterceptorManagerMBean() throws JBIException {
        final ObjectName mbeanName = this.context.getMBeanNames().createCustomComponentMBeanName(
                "interceptors_" + this.context.getComponentName());

        this.interceptorManager.setMBeanName(mbeanName);

        MBeanHelper
                .registerMBean(this.context.getMBeanServer(), mbeanName, this.interceptorManager);
    }

    /**
     * Register the {@link RuntimeConfigurationMBean}
     * 
     * @throws JBIException
     */
    private void registerRuntimeConfigMBean() throws JBIException {
        final ObjectName runtimeConfigurationMBeanName = this.context.getMBeanNames()
                .createCustomComponentMBeanName(
                        "runtime_configuration_" + this.context.getComponentName());

        this.runtimeConfigurationNotifier.setMBeanName(runtimeConfigurationMBeanName);

        MBeanHelper.registerMBean(this.context.getMBeanServer(), runtimeConfigurationMBeanName,
                this.runtimeConfigurationNotifier);
    }

    /**
     * 
     * @throws JBIException
     */
    private void unregisterInterceptorManagerMBean() throws JBIException {
        MBeanHelper.unregisterMBean(this.context.getMBeanServer(),
                this.interceptorManager.getMBeanName());
    }

    /**
     * Unregister the runtime configuration MBean
     * 
     * @throws JBIException
     */
    private void unregisterRuntimeConfigurationMBean() throws JBIException {
        MBeanHelper.unregisterMBean(this.context.getMBeanServer(),
                this.runtimeConfigurationNotifier.getMBeanName());
    }
    
    /**
     * Return the JBI processor manager
     * 
     * @return the JBI processor manager
     */
    public JBIProcessorManager getProcessorManager() {
        return this.processorManager;
    }

    /**
     * @return The timer used as sampler by response time probes
     */
    public Timer getProbesTimer() {
        return this.probesTimer;
    }

    public Monitoring getMonitoringBean() {
        return this.monitoringMbean;
    }

    /**
     * @return The sampling duration used by response time probes, in
     *         milliseconds
     */
    public long getResponseTimeProbeSamplePeriod() {
        final Settablelong value = this.getJbiComponentDescriptor().getComponent().getMonitoringSamplingPeriod();
        return (value == null ? DEFAULT_MONITORING_SAMPLING_PERIOD : value.getValue()) * 1000;
    }

    public Properties getPlaceHolders() {
        return this.customPlaceHolders;
    }

    private static class ComponentServiceEndpoint implements ServiceEndpoint {
        private final QName s;

        private final String e;

        public ComponentServiceEndpoint(final QName serviceName, final String endpointName) {
            this.s = serviceName;
            this.e = endpointName;
        }

        @Override
        public DocumentFragment getAsReference(final QName operationName) {
            return null;
        }

        @Override
        public String getEndpointName() {
            return this.e;
        }

        @Override
        public QName[] getInterfaces() {
            return new QName[] {};
        }

        @Override
        public QName getServiceName() {
            return this.s;
        }
    }
}
