/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.ObjectName;

import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.util.PropertiesUtil;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.SettableanyURI;
import org.ow2.petals.component.framework.mbean.ExtensionMBean;
import org.ow2.petals.component.framework.mbean.MBeanHelper;
import org.ow2.petals.component.framework.mbean.MBeanImplementor;
import org.ow2.petals.component.framework.util.WSDLUtilImpl;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.w3c.dom.Element;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;

/**
 * A default bootstrap. 
 * <p>
 * This bootstrap class register an MBean object into the
 * JMX server that will be available only during the bootstrap phase. The
 * developer only needs to override methods {@code getMethodList()} to set the
 * list of method that he want to expose via JMX.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
public class DefaultBootstrap extends JBIBootstrap implements MBeanImplementor {

    private static final String INVALID_CONF_VALUE_MSG = "Invalid value (%s) for the configuration parameter '%s'. Default value used: %d.";

    /**
     * The bootstrap MBean name
     */
    protected ObjectName mbeanName;

    /**
     * Returns the extension MBean name.
     * 
     * @return Returns the extension MBean name, except if the component is
     *         installed (ie. the extension MBean is not registered) in which
     *         case <code>null</code> is returned.
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    @Override
    public ObjectName getExtensionMBeanName() {

        final ObjectName objName;

        final MBeanServer server = this.installContext.getContext().getMBeanServer();
        if (server.isRegistered(this.mbeanName)) {
            objName = this.mbeanName;
        } else {
            objName = null;
        }

        return objName;
    }

    /**
     * Installs the component persisting the configuration done via the
     * extension Mbean.
     * 
     * @see javax.jbi.component.Bootstrap#onInstall()
     */
    @Override
    public void onInstall() throws JBIException {
        this.getLogger().info(getComponentMessageInfo());
        final File jbiXmlFile = new File(this.getInstallContext().getInstallRoot() + File.separator
                + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);

        this.validateComponentSettableConfiguration();

        try {
            try (final FileOutputStream jbiOutputStream = new FileOutputStream(jbiXmlFile)) {
                CDKJBIDescriptorBuilder.getInstance().writeXMLJBIdescriptor(this.getJbiComponentConfiguration(),
                        jbiOutputStream);
            }
        } catch (final IOException e) {
            throw new JBIException("Failed to persist the installation configuration", e);
        } catch (final JBIDescriptorException e) {
            throw new JBIException("Failed to persist the installation configuration", e);
        }
        try {
            this.processNativeWSDLResource(this.getInstallContext().getContext().getWorkspaceRoot());
        } catch (WSDLException e) {
            throw new JBIException("An error occurred while copying the native wsdl resource", e);
        }
    }

    /**
     * Clean resources allocated by the bootstrap
     * 
     * @see javax.jbi.component.Bootstrap#cleanUp()
     */
    @Override
    public void cleanUp() throws JBIException {

        MBeanServer server = installContext.getContext().getMBeanServer();
        if (server.isRegistered(mbeanName)) {
            try {
                server.unregisterMBean(mbeanName);
            } catch (InstanceNotFoundException e) {
                throw new JBIException("Error during bootstrap clean up", e);
            } catch (MBeanRegistrationException e) {
                throw new JBIException("Error during bootstrap clean up", e);
            }
        }

        super.cleanUp();
    }

    /**
     * Initialize the bootstrap registering its Extension MBean into the JMX
     * server.
     * 
     * @see org.ow2.petals.component.framework.JBIBootstrap#doInit()
     */
    @Override
    protected void doInit() throws JBIException {

        this.mbeanName = this.installContext
                .getContext()
                .getMBeanNames()
                .createCustomComponentMBeanName(
                        "bootstrap_" + this.installContext.getComponentName());

        MBeanHelper.registerMBean(this.installContext.getContext().getMBeanServer(),
                this.mbeanName, new ExtensionMBean(this));
    }

    @Override
    public Collection<String> getMBeanOperationsNames() {
        return new ArrayList<String>();
    }

    @Override
    public Collection<String> getMBeanAttributesNames() {
        return new ArrayList<String>();
    }

    /**
     * Validate the Component settable configuration.
     * 
     * @throws PEtALSCDKException
     */
    private void validateComponentSettableConfiguration() throws JBIException {
        final SettableanyURI propertiesFile = this.getJbiComponentConfiguration().getComponent().getPropertiesFile();

        if (!StringHelper.isNullOrEmpty(propertiesFile.getValue())) {
            try {
                PropertiesUtil.loadProperties(propertiesFile.getValue());
            } catch (final PropertiesException e) {
                throw new JBIException(e);
            }
        }
    }   

    /***
     * Copy the native WSDL resources in the work path of the components.
     * 
     * @param installRootPath
     *            Path of the work directory of the component.
     * @throws WSDLException
     */
    private void processNativeWSDLResource(final String workRootPath) throws WSDLException {
        final File nativeDir = new File(workRootPath, Constants.Component.NATIVE_SERVICE_FOLDER);
        if (!nativeDir.exists())
            nativeDir.mkdirs();
        final File nativeWSDLFile = new File(nativeDir, Constants.Component.COMPONENT_WSDL);
        // if it's not here, we copy it!
        // TODO why is it copied and not updated in place like with SUs?
        if (!nativeWSDLFile.exists()) {
            final URL nativeWsdlURL = this.getClass().getClassLoader()
                    .getResource(Constants.Component.COMPONENT_WSDL);
            if (nativeWsdlURL != null) {
                final Description wsdlDescription = WSDLUtilImpl.createWsdlDescription(nativeWsdlURL, true);
                // we need to include the imported documents because we are copying it so paths are differents!
                WSDLUtilImpl.convertDescriptionToFile(wsdlDescription, nativeWSDLFile, true);
            }
        }
    }

    /**
     * Message to indicate component's name and component's version
     * 
     * @return Exemple : <br/>
     *         "Install component [Name : petals-bc-mail, Version : 3.1.4-SNAPSHOT]"
     */
    private String getComponentMessageInfo() {
        String componentVersion = "Version not set";
        final String clazz = this.getJbiComponentConfiguration().getComponent()
                .getJbiListenerClassName();
        try {
            Class<?> clazz2 = Class.forName(clazz, false, this.getClass().getClassLoader());
            if (clazz2 != null) {
                componentVersion = clazz2.getPackage().getImplementationVersion();
                if (StringHelper.isNullOrEmpty(componentVersion))
                    componentVersion = "Version not set";
            }
        } catch (final ClassNotFoundException e) {
            this.getLogger().log(Level.FINE, "Can't retrieve component info", e);
        }
        final String result = "Install component [Name : "
                + this.getInstallContext().getComponentName() + ", Version : " + componentVersion
                + "]";
        return result;
    }

    /**
     * Set a parameter.
     * 
     * @param name
     * @param value
     */
    public void setParam(final String name, String value) {
        for (Element element : this.getJbiComponentConfiguration().getComponent().getAny()) {
            if (element.getLocalName().equals(name)) {
                element.setTextContent(value);
            }
        }
    }

    /**
     * Get a parameter.
     * 
     * @param name
     * @return The value of the parameter as {@link String}
     */
    public String getParam(final String name) {
        for (final Element element : this.getJbiComponentConfiguration().getComponent().getAny()) {
            if (element.getLocalName().equals(name)) {
                return element.getTextContent();
            }
        }

        return null;
    }

    /**
     * <p>
     * Get a parameter as integer.
     * </p>
     * <p>
     * Note: This method <b>must</b> used only with parameter declared as <code>petalsCDK:settableint</code>,
     * <code>petalsCDK:runtimeint</code>, <code>petalsCDK:settablepositivestrictint</code> or
     * <code>petalsCDK:runtimepositivestrictint</code>.
     * </p>
     * 
     * @param name
     *            The parameter name to get
     * @param defaultValue
     *            The default value to use if the parameter is not set, or if the number is not parsable.
     * @return The value of the parameter as integer
     */
    public int getParamAsInteger(final String name, final int defaultValue) {
        int value = defaultValue;

        final String valueAsString = this.getParam(name);
        if (valueAsString != null && !valueAsString.trim().isEmpty()) {
            try {
                value = Integer.parseInt(valueAsString);
            } catch (final NumberFormatException e) {
                // Invalid value, we use the default one
                this.getLogger().warning(String.format(INVALID_CONF_VALUE_MSG, valueAsString, name, defaultValue));
            }
        }

        return value;
    }

    /**
     * <p>
     * Get a parameter as long.
     * </p>
     * <p>
     * Note: This method <b>must</b> used only with parameter declared as <code>petalsCDK:settablelong</code>,
     * <code>petalsCDK:runtimelong</code>, <code>petalsCDK:settablepositivestrictlong</code> or
     * <code>petalsCDK:runtimepositivestrictlong</code>.
     * </p>
     * 
     * @param name
     *            The parameter name to get
     * @param defaultValue
     *            The default value to use if the parameter is not set, or if the number is not parsable.
     * @return The value of the parameter as long
     */
    public long getParamAsLong(final String name, final long defaultValue) {
        long value = defaultValue;

        final String valueAsString = this.getParam(name);
        if (valueAsString != null && !valueAsString.trim().isEmpty()) {
            try {
                value = Long.parseLong(valueAsString);
            } catch (final NumberFormatException e) {
                // Invalid value, we use the default one
                this.getLogger().warning(String.format(INVALID_CONF_VALUE_MSG, valueAsString, name, defaultValue));
            }
        }

        return value;
    }

    /**
     * <p>
     * Get a parameter as boolean.
     * </p>
     * <p>
     * Note: This method <b>must</b> used only with parameter declared as <code>petalsCDK:settableboolean</code>,
     * <code>petalsCDK:runtimeboolean</code>.
     * </p>
     * 
     * @param name
     *            The parameter name to get
     * @param defaultValue
     *            The default value to use if the parameter is not set.
     * @return The value of the parameter as boolean
     */
    public boolean getParamAsBoolean(final String name, final boolean defaultValue) {

        final boolean value;
        final String valueAsString = this.getParam(name);
        if (valueAsString != null && !valueAsString.trim().isEmpty()) {
            value = Boolean.parseBoolean(valueAsString);
        } else {
            value = defaultValue;
        }

        return value;
    }
}
