/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.InstallationContext;
import javax.management.ObjectName;

import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;

/**
 * A simple bootstrap implementation.
 * <p>
 * This bootstrap just copy the JBI extensions in a properties file that will be used by components.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
public class JBIBootstrap implements javax.jbi.component.Bootstrap {

    private static final String LOGGER_BOOTSTRAP_PREFIX = "bootstrap.";

    /**
     * The component installation context
     */
    protected InstallationContext installContext;

    /**
     * The logger
     */
    private Logger logger;

    /**
     * The component configuration
     */
    private Jbi jbiComponentConfiguration;

    /**
     * Returns null by default as this class does not provides JMX extension
     * MBean
     */
    @Override
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    public Jbi getJbiComponentConfiguration() {
        return this.jbiComponentConfiguration;
    }

    @Override
    public void init(final InstallationContext installContext) throws JBIException {
        this.installContext = installContext;
        try {
            // the restriction on the method getLogger has been unactivated in
            // PEtALS container to be able to manage the logger of the bootstrap
            this.logger = this.installContext.getContext().getLogger(
                    LOGGER_BOOTSTRAP_PREFIX + installContext.getComponentName(), null);
        } catch (final IllegalStateException e) {
            // not a PEtALS container
            this.logger = Logger.getLogger(LOGGER_BOOTSTRAP_PREFIX + installContext.getComponentName());
            this.logger.log(Level.WARNING, "Couldn't create a Logger with the JBI API", e);
        }

        if (installContext.isInstall()) {
            // We are in an installation context

            // Get the component configuration
            final File jbiXmlFile = new File(installContext.getInstallRoot() + File.separator
                    + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);
            try {
                try (final FileInputStream jbiInputStream = new FileInputStream(jbiXmlFile)) {
                    this.jbiComponentConfiguration = CDKJBIDescriptorBuilder.getInstance().buildJavaJBIDescriptor(
                            jbiInputStream);
                }
            } catch (final JBIDescriptorException e) {
                throw new JBIException(e);
            } catch (IOException e) {
                throw new JBIException(e);
            }

            // do additional configuration from component implementation. (
            // this method is overridden in the default bootstrap)
            doInit();
        }
    }

    @Override
    public void onInstall() throws JBIException {
        // NOP
    }

    @Override
    public void onUninstall() throws JBIException {
        // NOP
    }

    @Override
    public void cleanUp() throws JBIException {
        CDKJBIDescriptorBuilder.destroyInstance();
    }

    public InstallationContext getInstallContext() {
        return installContext;
    }

    public Logger getLogger() {
        return logger;
    }

    /**
     * Can be overridden.
     */
    protected void doInit() throws JBIException {
        // NOP
    }

}
