/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.bc;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.listener.AbstractExternalListener;

/**
 * This manager is used to enable a binding component to handle external service call. 
 * <p>
 * It manages listeners for external services (start and stop).
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 */
public class ExternalListenerManager {

    /**
     * A map of external listeners
     */
    private final Map<String, List<AbstractExternalListener>> externalListeners = new ConcurrentHashMap<String, List<AbstractExternalListener>>();

    /**
     * The JBI component
     */
    private BindingComponentServiceUnitManager suManager;

    public ExternalListenerManager(final BindingComponentServiceUnitManager suManager) {
        this.suManager = suManager;
    }

    public void createListeners(final String serviceUnitName, final List<Consumes> consumesList)
            throws PEtALSCDKException {
        for (final Consumes consumes : consumesList) {
            addExternalListerner(serviceUnitName, createAndInitExternalListener(consumes));
        }
    }

    /**
     * Start a listener for each given consumes.
     */
    public void startListening(final String serviceUnitName) throws PEtALSCDKException {
        final List<AbstractExternalListener> listeners = this.externalListeners.get(serviceUnitName);
        if (listeners != null) {
            for (final AbstractExternalListener externalListener : listeners) {
                externalListener.start();
            }
        }
    }

    /**
     * Stop all external listeners related to the given SU.
     */
    public void stopListening(final String serviceUnitName) throws PEtALSCDKException {
        final List<AbstractExternalListener> listeners = this.externalListeners.get(serviceUnitName);
        if (listeners != null) {

            final PEtALSCDKException exception = new PEtALSCDKException("Errors while stopping external listeners");

            for (final AbstractExternalListener externalListener : listeners) {
                try {
                    externalListener.stop();
                } catch (final Exception e) {
                    this.suManager.getComponent().getLogger().log(Level.WARNING,
                            "Error while stopping an external listener", e);
                    exception.addSuppressed(e);
                }
            }
            exception.throwIfNeeded();
        }
    }

    /**
     * Remove the SU bound listeners.
     * 
     */
    public void removeListeners(final String serviceUnitName) {
        this.externalListeners.remove(serviceUnitName);
    }

    /**
     * Create and initialize an external listener for the given SU
     * configuration.
     * 
     * @param consumes
     * @return
     * @throws PEtALSCDKException
     */
    private AbstractExternalListener createAndInitExternalListener(final Consumes consumes)
            throws PEtALSCDKException {
        AbstractExternalListener externalListener = this.suManager.createExternalListener();
        if (externalListener != null) {
            externalListener.init(suManager.getComponent());
            externalListener.setConsumes(consumes);
            externalListener.init();
        } else {
            throw new PEtALSCDKException("No ExternalListener can be created. Check if you have "
                    + "provided an ExternalListener in your component implementation.");
        }
        return externalListener;
    }

    /**
     * Add the external listener to the listener map.
     * 
     * @param serviceUnitName
     * @param externalListener
     */
    private void addExternalListerner(final String serviceUnitName, final AbstractExternalListener externalListener) {
        List<AbstractExternalListener> externalListenersForSU = this.externalListeners.get(serviceUnitName);
        if (externalListenersForSU == null) {
            externalListenersForSU = new ArrayList<AbstractExternalListener>();
            this.externalListeners.put(serviceUnitName, externalListenersForSU);
        }

        externalListenersForSU.add(externalListener);

    }

}
