/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.listener;

import java.net.URI;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.FlowAttributesExchangeHelper;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.configuration.SuConfigurationParameters;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.ExchangeProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageParam;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.Param;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.logger.StepLogHelper;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.MonitoringHelper;
import org.ow2.petals.component.framework.process.MessageExchangeProcessor;
import org.ow2.petals.component.framework.process.MessageInterceptorProcessor;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.component.framework.util.ServiceUnitUtil;

/**
 * The abstract methods for a CDK listener. 
 * <p>
 * Both external and JBI listener extend it.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Adrien Louis - EBM WebSourcing
 */
public abstract class AbstractListener {

    /**
     * The JBI component itself
     */
    private AbstractComponent component;

    /**
     * The messageExchange factory used to create MessageExchanges
     */
    protected MessageExchangeFactory factory;

    /**
     * The consumes for External Listeners
     */
    private Consumes consumes;

    /**
     * The provides for JBI listeners
     */
    private Provides provides;

    /**
     * The Consumes or Provides extensions
     */
    private SuConfigurationParameters extensions;

    /**
     * The message interceptor processor; Each listener have its own interceptor
     * processor in order to get better performances.
     */
    private MessageInterceptorProcessor messageInterceptorProcessor;

    /**
     * Creates a new instance of {@link AbstractListener}
     * 
     */
    public AbstractListener() {
        super();
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME <br>
     * - address (interface, service, endpoint) <br>
     * - MEP <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified
     *             in extensions
     */
    public Exchange createConsumeExchange(final Consumes consumes) throws MessagingException {
        return this.createConsumeExchange(consumes, null);
    }

    /**
     * Create a message exchange with the given exchange pattern and all the information set in the specified CONSUME,
     * <br>
     * - address (interface, service, endpoint) <br>
     * - operation (if no operation was given)<br>
     * - MEP (if no mep was given) <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * Note: the MEP is the only thing really needed, that's why it is a parameter, for the rest they can be tweaked by
     * the caller on the exchange once created.
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             The message exchange can't be created, or no pattern specified in extensions, or endpoint not
     *             registered.
     */
    public Exchange createConsumeExchange(final Consumes consumes, final MEPPatternConstants mep)
            throws MessagingException {

        final MEPPatternConstants actualMEP;
        if (mep == null) {
            actualMEP = ServiceUnitUtil.retrievePattern(consumes);
        } else {
            actualMEP = mep;
        }

        // only the MEP is really needed, the rest is optional!
        if (actualMEP == null) {
            throw new MessagingException("No MEP specified in the Consumes nor as a parameter");
        }

        final Exchange exchange = createExchange(consumes.getInterfaceName(), consumes.getServiceName(),
                consumes.getEndpointName(), actualMEP.value());

        exchange.setOperation(consumes.getOperation());
        
        this.setProperties(exchange, consumes);

        if (consumes.getLinkType() != null) {
            exchange.setProperty(Exchange.PROPERTY_PROVIDER_LINKTYPE, consumes.getLinkType().value());
        }

        return exchange;
    }

    /**
     * if service name and endpoint name are not null, they will be used to retrieve a {@link ServiceEndpoint}.
     * 
     * In any case interface name and service name will be set on the exchange.
     */
    public Exchange createExchange(final QName interfaceName, final QName serviceName, final String endpointName,
            final URI mep) throws MessagingException {

        final Exchange exchange;
        if (serviceName != null && endpointName != null) {
            exchange = createExchange(serviceName, endpointName, mep);
            
            if (interfaceName != null) {
                boolean found = false;
                for (final QName name : exchange.getEndpoint().getInterfaces()) {
                    if (interfaceName.equals(name)) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    // let's log that as information, this may means there is a bug in the SU
                    this.getLogger()
                            .warning(String.format(
                                    "Created exchange but endpoint retrieved for %s and %s does not implement the expected interface %s",
                                    serviceName, endpointName, interfaceName));
                }
            }
        } else {
            exchange = createExchange(mep);
        }

        // we set that in any case, it could make sense even if the endpoint is set
        exchange.setInterfaceName(interfaceName);
        exchange.setService(serviceName);

        return exchange;
    }

    /**
     * This create an exchange with the endpoint resolved from the parameters (which must be not <code>null</code>!)
     */
    public Exchange createExchange(final QName serviceName, final String endpointName, final URI mep)
            throws MessagingException {

        if (serviceName == null) {
            throw new MessagingException("serviceName can't be null");
        }

        if (endpointName == null) {
            throw new MessagingException("endpointName can't be null");
        }

        final Exchange exchange = this.createExchange(mep);

        final ServiceEndpoint ep = this.getComponent().getContext().getEndpoint(serviceName, endpointName);
        if (ep == null) {
            throw new MessagingException("The endpoint '" + endpointName + "' is not registered with the service name '"
                    + serviceName.toString() + "'.");
        }

        exchange.setEndpoint(ep);

        return exchange;
    }

    /**
     * Create an exchange implementation matching the given pattern.
     * 
     * @param mep
     *            a specific MEP pattern
     * @return an exchange implementation matching the given pattern
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified
     *             in extensions
     */
    public Exchange createExchange(final MEPPatternConstants mep) throws MessagingException {
        return createExchange(mep.value());
    }

    /**
     * Create an exchange implementation matching the given pattern.
     * 
     * @param mep
     *            a specific MEP pattern
     * @return an exchange implementation matching the given pattern
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified in extensions
     */
    public Exchange createExchange(final URI mep) throws MessagingException {
        if (mep == null) {
            throw new MessagingException("MEP cannot be null when creating an exchange");
        }

        final MessageExchange exchange = this.factory.createExchange(mep);

        return new ExchangeImpl(exchange);
    }

    /**
     * Get the component.
     * 
     * @return
     */
    public AbstractComponent getComponent() {
        return this.component;
    }

    /**
     *
     * @deprecated The direct usage of the Component context is not recommended in the CDK.
     *
     * @return
     */
    @Deprecated
    public ComponentContext getComponentContext() {
        return this.component.getContext();
    }

    public Consumes getConsumes() {
        return this.consumes;
    }

    /**
     * Get the service extensions.
     * 
     * @return Extensions, can be null if no extensions found
     */
    public SuConfigurationParameters getExtensions() {
        return this.extensions;
    }

    public Logger getLogger() {
        return this.component.getLogger();
    }

    public MessageInterceptorProcessor getMessageInterceptorProcessor() {
        return this.messageInterceptorProcessor;
    }

    public Provides getProvides() {
        return this.provides;
    }

    /**
     * Initialize the listener.
     * 
     * @param component
     */
    public final void init(final AbstractComponent component) {
        this.component = component;
        this.messageInterceptorProcessor = new MessageInterceptorProcessor(this.component,
                this.getLogger());
        this.factory = this.component.getChannel().createExchangeFactory();
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/>
     * Process MessageExchange interceptors before sending message. <br/>
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void send(final Exchange exchange) throws MessagingException {
        this.getLogger().fine(
                "Send an asynchronous exchange with id : " + exchange.getExchangeId());

        this.preSend((ExchangeImpl) exchange);

        this.getComponent().getChannel().send(exchange.getMessageExchange());
    }

    /**
     * This contains MONIT and stat related things before a send happen
     */
    private void preSend(final ExchangeImpl exchange) throws MessagingException {

        // Note: accept response can happen if answering an InOut for example
        this.interceptSendOrAcceptResponse(exchange);

        if (exchange.isProviderRole()) {
            // we get back the flow attributes we stored corresponding to the current processing
            final FlowAttributes current = exchange.getCurrentProcessFlowAttributes();

            // this should never be null because as a provider, we can only be answering an exchange!
            assert current != null;

            // we store them so that we can find them back in MessageExchangeProcessor
            // Note: we use our own prefix, because we are the only one relying on it!
            FlowAttributesExchangeHelper.setFlowAttributes(exchange.getMessageExchange(), current,
                    MessageExchangeProcessor.PROVIDER_FLOWATTRIBUTES_PREFIX);

            exchange.setCurrentProcessFlowAttributes(current);

            // it's the second step of the pattern for the provider
            // if it's inOptOut, then there must be either no fault, or a fault but no out (i.e. second step of the
            // pattern)
            // for the others, provider always only acts as the second step of a pattern
            // -> we log a provide flow end/failure
            if (!exchange.isInOptionalOutPattern() || (!exchange.isFaultMessage() || !exchange.isOutMessage())) {

                StepLogHelper.addMonitEndOrFailureTrace(getLogger(), exchange.getMessageExchange(), current);

                MonitoringHelper.updateStatsOnEndOfServiceProviderInvocation(exchange,
                        this.component.getMonitoringBean(), this.getLogger());
            }
        }

        // this is the first sent of the consumer
        // -> we set the flow attributes on the exchange
        if (exchange.isConsumerRole()) {

            FlowAttributes current;
            if (exchange.isNewExchange()) {
                current = PetalsExecutionContext.getFlowAttributes();

                if (current == null) {
                    // we want this to fail during tests!
                    assert false;
                    current = PetalsExecutionContext.initFlowAttributes();
                    getLogger().warning(
                            "Missing flow attributes in ExecutionContext (there must be a bug in the component implementation), "
                                    + "initialising some instead: " + current);
                }
            } else {
                current = exchange.getCurrentProcessFlowAttributes();
                // this should never be null because it's not a new exchange!
                assert current != null;
            }

            // Note: we use the default prefix, because this is shared knowledge between different components
            FlowAttributesExchangeHelper.setFlowAttributes(exchange.getMessageExchange(), current);

            exchange.setCurrentProcessFlowAttributes(current);
        }
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/>
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void sendAsync(final Exchange exchange, final AsyncContext asyncContext) throws MessagingException {

        final Object service = getServiceFrom(exchange);

        if (asyncContext.getTimeToLive() < 0) {
            asyncContext.setTimeToLive(getDefaultTimeoutFrom(service));
        }

        final String refKey = this.component.getAsyncManager().referenceAsyncContext(exchange, asyncContext, service);

        try {
            this.send(exchange);
        } catch (final MessagingException e) {
            this.component.getAsyncManager().revertReferenceAsyncContext(refKey);
            throw e;
        }
    }

    private Object getServiceFrom(final Exchange exchange) {
        final AbstractServiceUnitManager suManager = this.component.getServiceUnitManager();
        if (suManager == null) {
            return null;
        }
        if (exchange.isProviderRole()) {
            // we are a provider and we are answering someone
            // Note: maybe it's the same as this.provides, but anyway, let's use the information from the exchange
            return suManager.getProvidesFromExchange(exchange);
        } else {
            // we send a new exchange on a consume
            // Note: maybe it's the same as this.consumes, but anyway, let's use the information from the exchange
            return suManager.getConsumesFromExchange(exchange);
        }
    }

    /**
     * Careful, getTimeout can return null and unboxing it to long with raise an NPE
     */
    private long getDefaultTimeoutFrom(final Object service) {
        Long defaultTimeout = null;
        if (service instanceof Consumes) {
            defaultTimeout = ((Consumes) service).getTimeout();
        } else if (service instanceof Provides) {
            defaultTimeout = ((Provides) service).getTimeout();
        }
        // note: if service itself is null, the default timeout will also be used
        return defaultTimeout != null ? defaultTimeout : Constants.Component.DEFAULT_SEND_TIMEOUT;
    }

    /**
     * Call sendSync(MessageExchage exchange, long timeoutMS) with a timeout
     * value found in provides or consumes extensions or "0" if not found
     * 
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange) throws MessagingException {
        return this.sendSync(exchange, getDefaultTimeoutFrom(getServiceFrom(exchange)));
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * synchronous way.<br/>
     * Process MessageExchange interceptors before sending message. <br/>
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @param timeoutMS
     *            timeout period, in milliseconds (0 means no timeout); must be
     *            greater than or equal to zero
     * @return true if the message has been sent (if a timeout occurs, return
     *         false)
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange, final long timeoutMS) throws MessagingException {
        boolean bRet = false;

        this.getLogger().fine(
                "Send a synchronous exchange with id : " + exchange.getExchangeId() + " and timeout " + timeoutMS);

        try {
            this.preSend((ExchangeImpl) exchange);

            bRet = this.getComponent().getChannel().sendSync(exchange.getMessageExchange(), timeoutMS);

            if (bRet) {
                this.interceptSendResponse(exchange);
            }
        } catch (final PEtALSCDKException e) {
            throw new MessagingException(e);
        } catch (final JBIException e) {
            throw new MessagingException(e);
        }

        return bRet;
    }

    /**
     * @param consumes
     *            the consumes to set
     */
    public final void setConsumes(final Consumes consumes) {
        this.consumes = consumes;
        this.extensions = getExtensions(consumes);
    }

    /**
     * Set Provides section
     * 
     * @param provides
     *            the provides to set
     */
    public final void setProvides(final Provides provides) {
        this.provides = provides;
        this.extensions = getExtensions(provides);
    }

    /**
     * Get the extensions for the given provides/consumes.
     * 
     * @param service
     * @return The extensions
     */
    private final SuConfigurationParameters getExtensions(final Object service) {
        if (service == null) {
            return null;
        }
        final AbstractServiceUnitManager suManager = this.component.getServiceUnitManager();
        if (suManager == null) {
            return null;
        }
        final ServiceUnitDataHandler suDataHandler = suManager.getSUDataHandler(service);
        if (suDataHandler != null) {
            return suDataHandler.getConfigurationExtensions(service);
        } else {
            return null;
        }
    }

    /**
     * Call the Interceptors before sending the an exchange.
     */
    private void interceptSendOrAcceptResponse(final Exchange exchange) throws MessagingException {
        final AbstractServiceUnitManager suManager = this.component.getServiceUnitManager();
        if (suManager == null) {
            return;
        }
        // TODO isn't there component interceptor that should be trigerred?
        try {
            if (exchange.isConsumerRole()) {
                final Consumes exchangeConsumes = suManager.getConsumesFromExchange(exchange);
                if (exchangeConsumes != null) {
                    this.messageInterceptorProcessor.interceptSend(exchange, exchangeConsumes);
                }
            }

            if (exchange.isProviderRole()) {
                final Provides exchangeProvides = suManager.getProvidesFromExchange(exchange);
                if (exchangeProvides != null) {
                    this.messageInterceptorProcessor.interceptAcceptResponse(exchange, exchangeProvides);
                }
            }
        } catch (final PEtALSCDKException e) {
            throw new MessagingException(e);
        }
    }

    /**
     * Call the Interceptors after a synchronous send.
     */
    private void interceptSendResponse(final Exchange exchange) throws PEtALSCDKException {
        final AbstractServiceUnitManager suManager = this.component.getServiceUnitManager();
        if (suManager == null) {
            return;
        }
        // TODO isn't there component interceptor that should be trigerred?
        // we intercept only if we get a response as a consumer
        if (exchange.isConsumerRole()) {
            final Consumes exchangeConsumes = suManager.getConsumesFromExchange(exchange);
            if (exchangeConsumes != null) {
                this.messageInterceptorProcessor.interceptSendResponse(exchange, exchangeConsumes);
            }
        }
    }

    /**
     * Set exchange and message properties from extensions.
     * 
     * @param exchange
     *            a {@link Exchange}
     * @param extensions
     */
    private final void setProperties(final Exchange exchange, final Consumes consumes) {
        final ExchangeProperties exchangeProperties = consumes.getExchangeProperties();
        if (exchangeProperties != null) {
            for (final Param exchangeProperty : exchangeProperties.getExchangeProperty()) {
                exchange.setProperty(exchangeProperty.getName(), exchangeProperty.getValue());
            }
        }

        final MessageProperties messageProperties = consumes.getMessageProperties();
        if (messageProperties != null) {
            for (final MessageParam messageProperty : messageProperties.getMessageProperty()) {
                final NormalizedMessage message = ((ExchangeImpl) exchange).getMessageExchange()
                        .getMessage(messageProperty.getMessage().name());
                message.setProperty(messageProperty.getName(), messageProperty.getValue());
            }
        }
    }
}
