/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.mbean;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.ReflectionException;
import javax.management.RuntimeOperationsException;

import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.jbidescriptor.generated.Component;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimeint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimelong;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositiveint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivelong;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivestrictint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivestrictlong;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settableint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablelong;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositiveint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositivelong;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositivestrictint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositivestrictlong;

/**
 * Abstract dynamic MBean used as base for configuration MBean.
 * 
 * @author Christophe DENEUX - Linagora
 */
public abstract class AbstractConfigurationMBean implements DynamicMBean {

    /**
     * Name of the attribute {@link Component}.processorMaxPoolSize. It is check
     * when loading the current class.
     * 
     */
    public final static String PROCESSOR_MAX_POOL_SIZE_ATTR_NAME = "processorMaxPoolSize";

    /**
     * Name of the attribute {@link Component}.processorPoolSize. It is check
     * when loading the current class.
     * 
     */
    public final static String PROCESSOR_POOL_SIZE_ATTR_NAME = "processorPoolSize";

    protected final Logger logger;

    protected final Jbi componentJbiDescriptor;

    private final static Map<String, Class<?>> primitiveClasses = new HashMap<String, Class<?>>(8);

    static {
        primitiveClasses.put(Boolean.TYPE.toString(), Boolean.TYPE);
        primitiveClasses.put(Character.TYPE.toString(), Character.TYPE);
        primitiveClasses.put(Byte.TYPE.toString(), Byte.TYPE);
        primitiveClasses.put(Short.TYPE.toString(), Short.TYPE);
        primitiveClasses.put(Integer.TYPE.toString(), Integer.TYPE);
        primitiveClasses.put(Long.TYPE.toString(), Long.TYPE);
        primitiveClasses.put(Float.TYPE.toString(), Float.TYPE);
        primitiveClasses.put(Double.TYPE.toString(), Double.TYPE);
    }

    static {
        try {
            // We check that the attribute names are the right ones.
            Component.class.getDeclaredField(PROCESSOR_MAX_POOL_SIZE_ATTR_NAME);
            Component.class.getDeclaredField(PROCESSOR_POOL_SIZE_ATTR_NAME);
        } catch (final Exception e) {
            throw new RuntimeException(e);
        }
    }

    private static class AttributeSetterGetter {

        private String attributeName;

        private Method getter;

        private Method setter;

        public AttributeSetterGetter(final String attributeName, final Method getter, final Method setter) {
            this.attributeName = attributeName;
            this.getter = getter;
            this.setter = setter;
        }
    }

    private final List<AttributeSetterGetter> customAttrs = new ArrayList<AttributeSetterGetter>();

    private final MBeanImplementor implementor;

    public AbstractConfigurationMBean(final Jbi componentJbiDescriptor, final Logger logger,
            final MBeanImplementor implementor) {
        this.componentJbiDescriptor = componentJbiDescriptor;
        this.logger = logger;
        this.implementor = implementor;
        
        try {
            final Method[] methodList = implementor.getClass().getMethods();

            // get the custom attributes list
            for (final String attr : implementor.getMBeanAttributesNames()) {
                Method setter = null;
                Method getter = null;
                for (final Method method : methodList) {
                    if (method.getName().equalsIgnoreCase("set" + attr)) {
                        setter = method;
                    } else if (method.getName().equalsIgnoreCase("get" + attr)
                            || method.getName().equalsIgnoreCase("is" + attr)) {
                        getter = method;
                    }
                }

                // TODO get also the description for the attribute
                this.customAttrs.add(new AttributeSetterGetter(attr, getter, setter));
            }
        } catch (final Exception e) {
            this.logger.log(Level.WARNING, "Failed to initialize fields of the configuration MBean", e);
        }
    }

    /**
     * Dynamically get the operations for the MBean.
     */
    private MBeanOperationInfo[] getOperationsInfos() {

        final Collection<String> methods = implementor.getMBeanOperationsNames();

        final List<MBeanOperationInfo> operations = new ArrayList<MBeanOperationInfo>();
        for (final Method method : implementor.getClass().getMethods()) {
            final String name = method.getName();

            // add method if it has been defined by the user
            if (methods.contains(name)) {
                // TODO get also the description for the method
                final MBeanOperationInfo oper = new MBeanOperationInfo("Operation description", method);
                operations.add(oper);
            }
        }

        return operations.toArray(new MBeanOperationInfo[operations.size()]);
    }

    /**
     * Find the classes of the signature.
     */
    private static Class<?>[] findSignatureClasses(final String[] signature, final ClassLoader loader)
            throws ReflectionException {
        final int length = signature.length;
        final Class<?>[] tab = new Class[length];

        if (length == 0) {
            return tab;
        }

        try {
            for (int i = 0; i < length; i++) {
                // try to get the primitive class
                final Class<?> primCla = primitiveClasses.get(signature[i]);
                if (primCla != null) {
                    tab[i] = primCla;
                } else {
                    // try to get the class if not primitive one
                    tab[i] = Class.forName(signature[i], false, loader);
                }
            }
        } catch (final ClassNotFoundException e) {
            throw new ReflectionException(e, "The parameter class could not be found");
        } catch (final RuntimeException e) {
            throw e;
        }
        return tab;
    }

    @Override
    public synchronized Object invoke(final String actionName, final Object[] params, final String[] signature)
            throws MBeanException, ReflectionException {

        Object result = null;
        Class<?>[] parameters = null;

        final Class<?> objClass = implementor.getClass();
        final ClassLoader loader = objClass.getClassLoader();
        if (signature != null) {
            parameters = findSignatureClasses(signature, loader);
        }
        // Query the metadata service to get the right method
        //
        final Method method;
        try {
            method = objClass.getMethod(actionName, parameters);
        } catch (final NoSuchMethodException | SecurityException e) {
            throw new ReflectionException(e, "The operation with name " + actionName + " could not be found");
        }

        final ClassLoader ccl = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(loader);
        try {
            // invoke method
            result = method.invoke(implementor, params);
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e,
                    "IllegalAccessException" + " occured trying to invoke operation " + actionName);
        } catch (final RuntimeException e) {
            throw new RuntimeOperationsException(e,
                    "RuntimeException" + " occured trying to invoke operation " + actionName);
        } catch (final InvocationTargetException e) {
            this.logger.log(Level.WARNING, "Exception thrown in operation " + actionName, e);
            throw new MBeanException(e, "Exception thrown in operation " + actionName);
        } finally {
            Thread.currentThread().setContextClassLoader(ccl);
        }

        this.saveConfiguration();

        return result;
    }

    @Override
    public synchronized MBeanInfo getMBeanInfo() {
        // get the attributes that can be hot-changed
        final MBeanAttributeInfo[] attrs = this.getAttributesInfos();

        // get the methods that reload configuration
        final MBeanOperationInfo[] ops = this.getOperationsInfos();

        // create MBeanInfo with no constructors
        return new MBeanInfo(this.getClass().getName(), "Property Manager MBean", attrs, null, ops, null);
    }

    @Override
    public synchronized AttributeList getAttributes(final String[] attributes) {
        final AttributeList list = new AttributeList();

        for (final String attribute : attributes) {
            try {
                list.add(new Attribute(attribute, this.getAttribute(attribute)));
            } catch (final AttributeNotFoundException | ReflectionException e) {
                this.logger.log(Level.WARNING, "Failed to get attribute '" + attribute + "' in the configuration MBean",
                        e);
            }
        }

        return list;
    }

    private static Field findInheritedField(final Class<?> clazz, final String fieldName)
            throws NoSuchFieldException {

        Class<?> currentClazz = clazz;

        while (currentClazz != null) {
            try {
                return currentClazz.getDeclaredField(fieldName);
            } catch (final NoSuchFieldException e) {
                currentClazz = currentClazz.getSuperclass();
                if (currentClazz == null) {
                    // No more superclass
                    throw e;
                }
            }
        }
        // Statement unreachable
        return null;

    }

    @Override
    public synchronized void setAttribute(final Attribute attribute)
            throws AttributeNotFoundException, InvalidAttributeValueException, ReflectionException {

        try {
            final Field field = Component.class.getDeclaredField(attribute.getName());
            final Class<?> fieldType = field.getType();

            if (Runtimeint.class.isAssignableFrom(fieldType) || Settableint.class.isAssignableFrom(fieldType)) {

                if (!(attribute.getValue() instanceof Integer)) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '" + attribute.getName() + "': The value must be an integer.");
                }

                final int value = (int) attribute.getValue();

                final String fieldName = field.getName();
                if (fieldName.equals(PROCESSOR_MAX_POOL_SIZE_ATTR_NAME)) {
                    final int processorPoolSizeAttrValue = (int) this.getAttribute(PROCESSOR_POOL_SIZE_ATTR_NAME);
                    if (value <= processorPoolSizeAttrValue) {
                        throw new InvalidAttributeValueException("Invalid value for attribute '" + attribute.getName()
                                + "': The value must be superior or equal to the processor core pool size.");
                    }
                } else if (fieldName.equals(PROCESSOR_POOL_SIZE_ATTR_NAME)) {
                    final int processorMaxPoolSizeAttrValue = (int) this
                            .getAttribute(PROCESSOR_MAX_POOL_SIZE_ATTR_NAME);
                    if (value > processorMaxPoolSizeAttrValue) {
                        throw new InvalidAttributeValueException("Invalid value for attribute '" + attribute.getName()
                                + "': The value must be inferior or equal to the processor max pool size ("
                                + processorMaxPoolSizeAttrValue + ").");
                    }
                }

                if ((Runtimepositiveint.class.isAssignableFrom(fieldType)
                        || Settablepositiveint.class.isAssignableFrom(fieldType)) && value < 0) {
                    throw new InvalidAttributeValueException("Invalid value for attribute '" + attribute.getName()
                            + "': The value must be superior or equal to 0.");
                }

                if ((Runtimepositivestrictint.class.isAssignableFrom(fieldType)
                        || Settablepositivestrictint.class.isAssignableFrom(fieldType)) && value <= 0) {
                    throw new InvalidAttributeValueException("Invalid value for attribute '" + attribute.getName()
                            + "': The value must be superior or equal to 1.");
                }
            } else if (Runtimelong.class.isAssignableFrom(fieldType)
                    || Settablelong.class.isAssignableFrom(fieldType)) {

                if (!(attribute.getValue() instanceof Long)) {
                    throw new InvalidAttributeValueException(
                            "Invalid value for attribute '" + attribute.getName() + "': The value must be a long.");
                }

                final long value = (long) attribute.getValue();

                if ((Runtimepositivelong.class.isAssignableFrom(fieldType)
                        || Settablepositivelong.class.isAssignableFrom(fieldType)) && value < 0) {
                    throw new InvalidAttributeValueException("Invalid value for attribute '" + attribute.getName()
                            + "': The value must be superior or equal to 0.");
                }

                if ((Runtimepositivestrictlong.class.isAssignableFrom(fieldType)
                        || Settablepositivestrictlong.class.isAssignableFrom(fieldType)) && value < 0) {
                    throw new InvalidAttributeValueException("Invalid value for attribute '" + attribute.getName()
                            + "': The value must be superior or equal to 0.");
                }
            }

            field.setAccessible(true);
            final Object settableObject = field.get(this.componentJbiDescriptor.getComponent());
            
            // get the previous value
            String methodName = "getValue";
            if (fieldType.getSimpleName().endsWith("boolean")) {
                methodName = "isValue";
            }
            final Method getMethod = settableObject.getClass().getMethod(methodName);
            final Object oldValue = getMethod.invoke(settableObject);
            
            // set the value
            final Field primitiveField = findInheritedField(fieldType, "value");
            final Method setMethod = settableObject.getClass().getMethod("setValue",
                    new Class[] { primitiveField.getType() });
            setMethod.invoke(settableObject, new Object[] { attribute.getValue() });
            
            this.saveConfiguration();
            this.notifyConfigurationChange(attribute.getName() + "modified", attribute.getName(),
                    oldValue, attribute.getValue());
            
        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute.getName() + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            // try to set a custom attribute
            this.setCustomAttribute(attribute);
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e);
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e);
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e);
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e);
        }
    }
    
    private void setCustomAttribute(final Attribute attribute)
            throws ReflectionException, AttributeNotFoundException, InvalidAttributeValueException {

        try {
            for (final AttributeSetterGetter attributeSetterGetter : this.customAttrs) {
                if (attributeSetterGetter.attributeName.equals(attribute.getName())) {
                    attributeSetterGetter.setter.invoke(implementor, attribute.getValue());
                    return;
                }
            }
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            if (e.getCause() instanceof InvalidAttributeValueException) {
                throw (InvalidAttributeValueException) e.getCause();
            } else {
                throw new ReflectionException(e, e.getMessage());
            }
        }

        // Custom attribute not found
        throw new AttributeNotFoundException("Attribute '" + attribute + "' does not exist");
    }
    
    private Object getCustomAttribute(final String attributeName)
            throws ReflectionException, AttributeNotFoundException {
        try {
            for (final AttributeSetterGetter attributeSetterGetter : this.customAttrs) {
                if (attributeSetterGetter.attributeName.equals(attributeName)) {
                    return attributeSetterGetter.getter.invoke(implementor);
                }
            }
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e, e.getMessage());
        }

        // Custom attribute not found
        throw new AttributeNotFoundException("Attribute '" + attributeName + "' does not exist");
    }
    protected void saveConfiguration() {
        // NOP
    }
    
    protected void notifyConfigurationChange(final String msg, final String attrName,
            final Object oldVal, final Object newVal) {
        // NOP
    }

    @Override
    public synchronized Object getAttribute(final String attribute)
            throws AttributeNotFoundException, ReflectionException {
        Object attributeObject = null;

        try {
            final Field field = Component.class.getDeclaredField(attribute);
            field.setAccessible(true);
            final Object gettableObject = field.get(this.componentJbiDescriptor.getComponent());
            if (gettableObject != null) {
                String methodName = "getValue";
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    methodName = "isValue";
                }
                final Method method = gettableObject.getClass().getMethod(methodName);
                attributeObject = method.invoke(gettableObject);
            }

        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            // try to get a custom attribute
            attributeObject = this.getCustomAttribute(attribute);
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e, e.getMessage());
        }

        return attributeObject;
    }

    @Override
    public synchronized AttributeList setAttributes(final AttributeList list) {
        final AttributeList attributeList = new AttributeList();

        for (final Object object : list) {
            final Attribute attribute = (Attribute) object;
            try {
                this.setAttribute(attribute);
            } catch (final Exception e) {
                this.logger.log(Level.WARNING,
                        "Failed to set attribute '" + attribute.getName() + "' in the configuration MBean", e);
            }
        }

        return attributeList;
    }

    /**
     * Dynamically get attributes for the MBean. The attributes are the ones
     * from the properties file.
     * 
     * @return
     */
    private MBeanAttributeInfo[] getAttributesInfos() {
        final Component componentConfiguration = this.componentJbiDescriptor.getComponent();

        final Collection<MBeanAttributeInfo> attrs = new ArrayList<>();

        try {
            // get the settable fields
            for (final Field field : Component.class.getDeclaredFields()) {
                if (this.isSettableField(field)) {
                    field.setAccessible(true);
                    final Object settableObject = field.get(componentConfiguration);
                    if (settableObject != null) {
                        Method getter;
                        if (field.getType().getSimpleName().endsWith("boolean")) {
                            getter = settableObject.getClass().getMethod("isValue");
                        } else {
                            getter = settableObject.getClass().getMethod("getValue");
                        }
                        final Field primitiveField = findInheritedField(field.getType(), "value");
                        final Method setter = settableObject.getClass().getMethod("setValue",
                                new Class[] { primitiveField.getType() });

                        attrs.add(
                                new MBeanAttributeInfo(field.getName(), "Property " + field.getName(), getter, setter));
                    }
                }
            }

            for (final AttributeSetterGetter attr : customAttrs) {
                attrs.add(new MBeanAttributeInfo(attr.attributeName, "Property " + attr.attributeName, attr.getter,
                        attr.setter));
            }
        } catch (final Exception e) {
            this.logger.log(Level.WARNING, "Failed to initialize fields of the configuration MBean", e);
        }

        return attrs.toArray(new MBeanAttributeInfo[attrs.size()]);
    }

    /**
     * @return <code>true</code> if the field declared in the jbi descriptor is meant to be modified by the MBean
     */
    protected abstract boolean isSettableField(final Field field);
}
