/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.mbean;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.management.MBeanException;
import javax.management.ObjectName;

import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Interceptor;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.ComponentInterceptor;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;

/**
 * Add and engage interceptors at runtime.
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
public class InterceptorManager implements InterceptorManagerMBean {

    private final AbstractComponent component;

    private final Logger logger;

    private final File jbiComponentFile;

    private ObjectName mBeanName;

    public InterceptorManager(final AbstractComponent component, final File jbiComponentFile) {
        this.component = component;
        this.jbiComponentFile = jbiComponentFile;
        this.logger = component.getLogger();
    }

    @Override
    public void addInterceptor(final String className, final String interceptorName) throws MBeanException {
        this.logger.info("Add interceptor " + interceptorName);

        if (this.component.getInterceptor(interceptorName) != null) {
            throw new MBeanException(new Exception("This interceptor is already registered"));
        }

        final Interceptor itc;
        try {
            final Class<?> clazz = Thread.currentThread().getContextClassLoader().loadClass(className);
            itc = (Interceptor) clazz.newInstance();
            itc.init(this.component, interceptorName, false, new HashMap<String, String>(),
                    this.logger);
        } catch (final PEtALSCDKException | InstantiationException | IllegalAccessException | ClassNotFoundException
                | ClassCastException e) {
            throw new MBeanException(e);
        }
        final ComponentInterceptor componentInterceptor = new ComponentInterceptor();
        componentInterceptor.setActive(false);
        componentInterceptor.setClazz(className);
        componentInterceptor.setName(interceptorName);

        final Map<String, Interceptor> interceptors = this.component.getInterceptors();
        interceptors.put(interceptorName, itc);
        final List<ComponentInterceptor> componentInterceptors = this.component.getJbiComponentDescriptor()
                .getComponent().getComponentInterceptors().getInterceptor();
        componentInterceptors.add(componentInterceptor);
        this.save();
    }

    @Override
    public void engageInterceptor(final String interceptorName) throws MBeanException {
        this.logger.info("Engage interceptor " + interceptorName);

        final Interceptor interceptor = this.component.getInterceptor(interceptorName);
        if (interceptor == null) {
            throw new MBeanException(new Exception("Failed to retrieve interceptor named '"
                    + interceptorName + "'"));
        }

        if (!interceptor.isActive()) {
            final AbstractServiceUnitManager suManager = this.component.getServiceUnitManager();
            if (suManager != null) {
                for (final ServiceUnitDataHandler suDataHandler : suManager.getServiceUnitDataHandlers()) {
                    for (final MessageInterceptorManager manager : suDataHandler.getMessageInterceptorManagers()) {
                        manager.addGlobalInterceptor(interceptor);
                    }
                }
            }

            interceptor.setActive(true);
            final List<ComponentInterceptor> componentInterceptors = this.component.getJbiComponentDescriptor()
                    .getComponent().getComponentInterceptors().getInterceptor();
            for (final ComponentInterceptor componentInterceptor : componentInterceptors) {
                if (componentInterceptor.getName().equals(interceptor.getName())) {
                    componentInterceptor.setActive(true);
                    this.save();
                    break;
                }
            }
        }
    }

    /**
     * 
     * @param mbeanName
     */
    public void setMBeanName(final ObjectName mbeanName) {
        this.mBeanName = mbeanName;
    }

    /**
     * 
     * @return
     */
    public ObjectName getMBeanName() {
        return this.mBeanName;
    }

    private void save() {
        try {
            try (final FileOutputStream outputStream = new FileOutputStream(jbiComponentFile)) {
                CDKJBIDescriptorBuilder.getInstance().writeXMLJBIdescriptor(this.component.getJbiComponentDescriptor(),
                        outputStream);
            }
        } catch (final JBIDescriptorException | IOException e) {
            this.logger.log(Level.WARNING, "Failed to persist modification of the Interceptor manager MBean", e);
        }
    }

}
