/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.mbean;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Logger;

import javax.management.AttributeChangeNotification;
import javax.management.DynamicMBean;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationEmitter;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;

import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;

/**
 * <p>
 * A runtime configuration MBean which sends notifications to subscribers.
 * </p>
 * <p>
 * It is used to change configuration values at runtime or to reload configuration.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 */
public class RuntimeConfigurationNotifier extends AbstractConfigurationMBean implements
        DynamicMBean, NotificationEmitter {

    private final NotificationBroadcasterSupport broadcasterSupport;

    private final AtomicLong seqNumber;

    private final File jbiComponentDescriptorFile;

    private ObjectName mbeanName;

    public RuntimeConfigurationNotifier(final AbstractComponent component, final File jbiComponentDescriptorFile,
            final Logger logger) {
        super(component.getJbiComponentDescriptor(), logger, component);
        this.jbiComponentDescriptorFile = jbiComponentDescriptorFile;
        this.broadcasterSupport = new NotificationBroadcasterSupport();
        this.seqNumber = new AtomicLong(0);
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] { new MBeanNotificationInfo(
                new String[] { AttributeChangeNotification.ATTRIBUTE_CHANGE },
                AttributeChangeNotification.class.getName(),
                "Change runtime configuration attribute") };
    }

    @Override
    public void addNotificationListener(final NotificationListener listener,
            final NotificationFilter filter, final Object handback) {
        this.broadcasterSupport.addNotificationListener(listener, filter, handback);
    }

    @Override
    public void removeNotificationListener(final NotificationListener listener)
            throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener);
    }

    @Override
    public void removeNotificationListener(final NotificationListener listener,
            final NotificationFilter filter, final Object handback)
            throws ListenerNotFoundException {
        this.broadcasterSupport.removeNotificationListener(listener, filter, handback);
    }

    public ObjectName getMBeanName() {
        return this.mbeanName;
    }

    public void setMBeanName(final ObjectName mbeanName) {
        this.mbeanName = mbeanName;
    }

    @Override
    protected void saveConfiguration() {
        try (final OutputStream jbiComponentFileAsOutputStream = new FileOutputStream(
                this.jbiComponentDescriptorFile)) {
            CDKJBIDescriptorBuilder.getInstance().writeXMLJBIdescriptor(this.componentJbiDescriptor,
                    jbiComponentFileAsOutputStream);
        } catch (final FileNotFoundException e) {
            this.logger.log(Level.WARNING, "Failed to persist modification of the runtime configuration MBean : "
                    + e.getClass().getSimpleName(), e);
        } catch (final JBIDescriptorException e) {
            this.logger.log(Level.WARNING, "Failed to persist modification of the runtime configuration MBean : "
                    + e.getClass().getSimpleName(), e);
        } catch (final IOException e) {
            this.logger.log(Level.WARNING, "Can't close jbi component file output stream", e);
        }
    }

    @Override
    protected void notifyConfigurationChange(final String msg, final String attrName,
            final Object oldVal, final Object newVal) {

        final Notification notif = new AttributeChangeNotification(this, this.seqNumber
                .getAndIncrement(), System.currentTimeMillis(), msg, attrName, String.class
                .getName(), oldVal, newVal);
        this.broadcasterSupport.sendNotification(notif);
    }

    @Override
    protected boolean isSettableField(final Field field) {
        return field.getType().getSimpleName().startsWith("Runtime");
    }
}
