/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.message;

import java.io.InputStream;
import java.net.URI;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractListener;
import org.ow2.petals.component.framework.process.MessageExchangeProcessor;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.lang.ExceptionHelper;
import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * This is a {@link MessageExchange} Wrapper.
 * 
 * It simplify the manipulation of the underlying {@link MessageExchange}.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Victor Noël - Linagora
 */
public class ExchangeImpl implements Exchange {

    private final MessageExchange exchange;

    /**
     * Meant to be used by the {@link MessageExchangeProcessor} and {@link AbstractListener} only.
     */
    private FlowAttributes currentProcessFA;

    /**
     * 
     * Creates a new instance of {@link ExchangeImpl}
     * 
     * @param exchange
     */
    public ExchangeImpl(final MessageExchange exchange) {
        super();
        this.exchange = exchange;
    }

    @Override
    public boolean amIOwner() {
        try {
            exchange.getRole();
            return true;
        } catch (final IllegalStateException ex) {
            return false;
        }
    }

    /**
     * Copy attachment to a normalized message.
     * 
     * @param attachments
     *            Attachments to copy. Must be not-null.
     * @param normalizedMesg
     *            Normalized message in which attachments must be copied. Must
     *            be not-null.
     * @throws MessagingException
     *             An error occurs when adding an attachement to the normalized
     *             message.
     */
    final private static void addAttachments(final Map<String, DataHandler> attachments,
            final NormalizedMessage normalizedMesg) throws MessagingException {
        for (final Entry<String, DataHandler> entry : attachments.entrySet()) {
            normalizedMesg.addAttachment(entry.getKey(), entry.getValue());
        }
    }

    /**
     * Copy properties to a normalized message.
     * 
     * @param properties
     *            Properties to copy. Must be not-null.
     * @param normalizedMesg
     *            Normalized message in which properties must be copied. Must be
     *            not-null.
     */
    final private static void addMessageProperties(final Map<String, Object> properties,
            final NormalizedMessage normalizedMsg) {
        for (final Entry<String, Object> entry : properties.entrySet()) {
            normalizedMsg.setProperty(entry.getKey(), entry.getValue());
        }
    }

    @Override
    public void addInMessageAttachment(final String id, final DataHandler content)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        inMess.addAttachment(id, content);
        this.setInMessage(inMess);
    }

    @Override
    public void addOutMessageAttachment(final String id, final DataHandler content)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        outMess.addAttachment(id, content);
        this.setOutMessage(outMess);
    }

    @Override
    public Fault createFault() throws MessagingException {
        return this.exchange.createFault();
    }

    /**
     * Creates a message.
     * <p>
     * This method is not exposed by the {@code Exchange} interface.
     * 
     * @return
     * @throws MessagingException
     */
    public NormalizedMessage createMessage() throws MessagingException {
        return this.exchange.createMessage();
    }

    @Override
    public ServiceEndpoint getEndpoint() {
        return this.exchange.getEndpoint();
    }

    @Override
    public String getEndpointName() {
        final ServiceEndpoint ep = this.exchange.getEndpoint();
        return ep == null ? null : ep.getEndpointName();
    }

    @Override
    public Exception getError() {
        return this.exchange.getError();
    }

    @Override
    public String getExchangeId() {
        return this.exchange.getExchangeId();
    }

    @Override
    public Fault getFault() {
        return this.exchange.getFault();
    }

    @Override
    public NormalizedMessage getInMessage() {
        NormalizedMessage message = this.exchange.getMessage(IN_MESSAGE_NAME);
        if (message == null) {
            try {
                message = this.exchange.createMessage();
                this.exchange.setMessage(message, IN_MESSAGE_NAME);
            } catch (final MessagingException e) {
                // the purpose of getInMessage seems to be able to work more easily with message
                // introducing a MessagingException would be bothering and like that
                // we can nicely find bugs when they happen!
                throw new UncheckedException("Can't create or set IN message", e);
            }
        }
        return message;
    }

    @Override
    public DataHandler getInMessageAttachment(final String name) throws MessagingException {

        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        return inMessage.getAttachment(name);
    }

    @Override
    @SuppressWarnings("unchecked")
    public Set<String> getInMessageAttachmentNames() throws MessagingException {

        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        return (Set<String>) inMessage.getAttachmentNames();
    }

    @Override
    public Set<DataHandler> getInMessageAttachments() throws MessagingException {
        final Set<DataHandler> attachments = new HashSet<DataHandler>();

        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        final Set<?> names = inMessage.getAttachmentNames();
        for (final Object attachId : names) {
            final DataHandler dh = inMessage.getAttachment((String) attachId);
            if (dh != null) {
                attachments.add(dh);
            }
        }
        return attachments;
    }

    @Override
    public Document getInMessageContentAsDocument() throws MessagingException {
        return this.getInMessageContentAsDocument(true);
    }

    @Override
    public Document getInMessageContentAsDocument(final boolean forkSource)
            throws MessagingException {
        final NormalizedMessage inMessage = this.getInMessage();
        Document result = null;
        final Source source = inMessage.getContent();
        try {
            if (source != null) {
                result = SourceUtil.createDocument(source, forkSource);
            }
        } catch (PEtALSCDKException e) {
            throw new MessagingException(e);
        }

        return result;
    }

    @Override
    public Source getInMessageContentAsSource() throws MessagingException {
        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }
        return inMessage.getContent();
    }

    @Override
    public Object getInMessageProperty(final String name) throws MessagingException {
        final NormalizedMessage nm = this.getInMessage();
        if (nm == null) {
            throw new MessagingException("In message cannot be null");
        }
        return nm.getProperty(name);
    }

    @Override
    @SuppressWarnings("unchecked")
    public Set<String> getInMessagePropertyNames() throws MessagingException {
        final NormalizedMessage nm = this.getInMessage();
        if (nm == null) {
            throw new MessagingException("In message cannot be null");
        }
        return (Set<String>) nm.getPropertyNames();
    }

    @Override
    public Subject getInSubject() throws MessagingException {
        final NormalizedMessage inMessage = this.getInMessage();
        if (inMessage == null) {
            throw new MessagingException("In message cannot be null");
        }

        return inMessage.getSecuritySubject();
    }

    @Override
    public QName getInterfaceName() {
        return this.exchange.getInterfaceName();
    }

    /**
     * Return the wrapped Message Exchange.
     */
    @Override
    public MessageExchange getMessageExchange() {
        return this.exchange;
    }

    @Override
    public QName getOperation() {
        return this.exchange.getOperation();
    }

    @Override
    public String getOperationName() {
        final QName operation = this.exchange.getOperation();
        if (operation == null) {
            return null;
        }
        return operation.getLocalPart();
    }

    @Override
    public NormalizedMessage getOutMessage() {
        NormalizedMessage message = this.exchange.getMessage(OUT_MESSAGE_NAME);
        if (message == null) {
            try {
                message = this.exchange.createMessage();
                this.exchange.setMessage(message, OUT_MESSAGE_NAME);
            } catch (final MessagingException e) {
                // the purpose of getOutMessage seems to be able to work more easily with message
                // introducing a MessagingException would be bothering and like that
                // we can nicely find bugs when they happen!
                throw new UncheckedException("Can't create or set OUT message", e);
            }
        }
        return message;
    }

    @Override
    public DataHandler getOutMessageAttachment(final String name) throws MessagingException {

        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        return outMessage.getAttachment(name);
    }

    @Override
    @SuppressWarnings("unchecked")
    public Set<String> getOutMessageAttachmentNames() throws MessagingException {

        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        return (Set<String>) outMessage.getAttachmentNames();
    }

    @Override
    public Set<DataHandler> getOutMessageAttachments() throws MessagingException {
        final Set<DataHandler> attachments = new HashSet<DataHandler>();

        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        final Set<?> names = outMessage.getAttachmentNames();
        for (final Object attachId : names) {
            final DataHandler dh = outMessage.getAttachment((String) attachId);
            if (dh != null) {
                attachments.add(dh);
            }
        }
        return attachments;
    }

    @Override
    public Document getOutMessageContentAsDocument() throws MessagingException {
        return this.getOutMessageContentAsDocument(true);
    }

    @Override
    public Document getOutMessageContentAsDocument(final boolean forkSource)
            throws MessagingException {
        final NormalizedMessage outMessage = this.getOutMessage();
        Document result = null;
        final Source source = outMessage.getContent();
        try {
            if (source != null) {
                result = SourceUtil.createDocument(source, forkSource);
            }
        } catch (PEtALSCDKException e) {
            throw new MessagingException(e);
        }

        return result;
    }

    @Override
    public Source getOutMessageContentAsSource() throws MessagingException {
        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }
        return outMessage.getContent();
    }

    @Override
    public Object getOutMessageProperty(final String name) throws MessagingException {
        final NormalizedMessage nm = this.getOutMessage();
        if (nm == null) {
            throw new MessagingException("Out message cannot be null");
        }
        return nm.getProperty(name);
    }


    @Override
    public Set<String> getOutMessagePropertyNames() throws MessagingException {
        final NormalizedMessage nm = this.getOutMessage();
        if (nm == null) {
            throw new MessagingException("Out message cannot be null");
        }
        @SuppressWarnings("unchecked")
        final Set<String> propertyNames = nm.getPropertyNames();
        return propertyNames;
    }

    @Override
    public Subject getOutSubject() throws MessagingException {
        final NormalizedMessage outMessage = this.getOutMessage();
        if (outMessage == null) {
            throw new MessagingException("Out message cannot be null");
        }

        return outMessage.getSecuritySubject();
    }

    @Override
    public URI getPattern() {
        return this.exchange.getPattern();
    }

    @Override
    public Object getProperty(final String name) {
        return this.exchange.getProperty(name);
    }

    @Override
    public Set<String> getPropertyNames() {
        @SuppressWarnings("unchecked")
        final Set<String> propertyNames = this.exchange.getPropertyNames();
        return propertyNames;
    }

    @Override
    public MessageExchange.Role getRole() {
        return this.exchange.getRole();
    }

    @Override
    public QName getService() {
        return this.exchange.getService();
    }

    @Override
    public ExchangeStatus getStatus() {
        return this.exchange.getStatus();
    }

    @Override
    public boolean isActiveStatus() {
        return ExchangeStatus.ACTIVE.equals(this.getStatus());
    }

    @Override
    public boolean isConsumerRole() {
        return MessageExchange.Role.CONSUMER.equals(this.getRole());
    }

    @Override
    public boolean isDoneStatus() {
        return ExchangeStatus.DONE.equals(this.getStatus());
    }

    @Override
    public boolean isErrorStatus() {
        return ExchangeStatus.ERROR.equals(this.getStatus());
    }

    @Override
    public boolean isInMessage() {
        return this.exchange.getMessage(IN_MESSAGE_NAME) != null
                && this.exchange.getMessage(IN_MESSAGE_NAME).getContent() != null;
    }

    @Override
    public boolean isInOnlyPattern() {
        return MEPConstants.IN_ONLY_PATTERN.equals(this.exchange.getPattern());
    }

    @Override
    public boolean isInOptionalOutPattern() {
        return MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(this.exchange.getPattern());
    }

    @Override
    public boolean isInOutPattern() {
        return MEPConstants.IN_OUT_PATTERN.equals(this.exchange.getPattern());
    }

    @Override
    public boolean isOutMessage() {
        return this.exchange.getMessage(OUT_MESSAGE_NAME) != null
                && this.exchange.getMessage(OUT_MESSAGE_NAME).getContent() != null;
    }

    @Override
    public boolean isFaultMessage() {
        return this.exchange.getFault() != null && this.exchange.getFault().getContent() != null;
    }

    @Override
    public boolean isProviderRole() {
        return MessageExchange.Role.PROVIDER.equals(this.getRole());
    }

    @Override
    public boolean isRobustInOnlyPattern() {
        return MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(this.exchange.getPattern());
    }

    public boolean isNewExchange() {
        return isActiveStatus() && !isOutMessage() && !isFaultMessage();
    }

    @Override
    public boolean isTransacted() {
        return this.exchange.isTransacted();
    }

    @Override
    public void setActiveStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.ACTIVE);
    }

    public void setCheckRole(final boolean checkRole) {
        // we need this hack because MessageExchangeImpl is not visible from this project nor the component classloader
        if (checkRole) {
            setProperty("org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl.dontDoChecks", null);
        } else {
            setProperty("org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl.dontDoChecks", true);
        }
    }

    @Override
    public void setDoneStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.DONE);
    }

    @Override
    public void setEndpoint(final ServiceEndpoint endpoint) {
        this.exchange.setEndpoint(endpoint);
    }

    @Override
    public void setError(final Exception exception) {
        final String className = exception.getClass().getName();
        if (exception instanceof MessagingException || className.startsWith("java.")
                || className.startsWith("org.w3c.") || className.startsWith("org.xml.")) {
            this.exchange.setError(exception);
        } else {
            final MessagingException msgException = new MessagingException(ExceptionHelper.getStackTrace(exception));
            this.exchange.setError(msgException);
        }
    }

    @Override
    public void setErrorStatus() throws MessagingException {
        this.setStatus(ExchangeStatus.ERROR);
    }

    @Override
    public void setFault(final Fault fault) throws MessagingException {
        this.exchange.setFault(fault);
    }

    @Override
    public void setInMessage(final NormalizedMessage msg) throws MessagingException {
        this.exchange.setMessage(msg, IN_MESSAGE_NAME);
    }

    @Override
    public void setInMessageAttachment(String name, DataHandler attachment)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }

        inMess.addAttachment(name, attachment);
        this.setInMessage(inMess);
    }

    @Override
    public void setInMessageAttachments(final Map<String, DataHandler> attachments)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        addAttachments(attachments, inMess);
        this.setInMessage(inMess);
    }

    @Override
    public void setInMessageContent(final Document inContent) throws MessagingException {
        this.setInMessageContent(SourceUtil.createSource(inContent));
    }

    @Override
    public void setInMessageContent(final InputStream inContent) throws MessagingException {
        this.setInMessageContent(new StreamSource(inContent));
    }

    @Override
    public void setInMessageContent(final Source inContent) throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        inMess.setContent(inContent);
        this.setInMessage(inMess);
    }

    @Override
    public void setInMessageProperties(final Map<String, Object> properties)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        addMessageProperties(properties, inMess);
        this.setInMessage(inMess);
    }

    @Override
    public void setInMessageProperty(final String name, final Object value)
            throws MessagingException {
        NormalizedMessage inMess = this.getInMessage();
        if (inMess == null) {
            inMess = this.exchange.createMessage();
        }
        inMess.setProperty(name, value);
        this.setInMessage(inMess);
    }

    @Override
    public void setInSubject(final Subject subject) {
        this.getInMessage().setSecuritySubject(subject);
    }

    @Override
    public void setInterfaceName(final QName interfaceName) {
        this.exchange.setInterfaceName(interfaceName);
    }

    @Override
    public void setOperation(final QName name) {
        this.exchange.setOperation(name);
    }

    @Override
    public void setOutMessage(final NormalizedMessage msg) throws MessagingException {
        this.exchange.setMessage(msg, OUT_MESSAGE_NAME);
    }

    @Override
    public void setOutMessageAttachment(String name, DataHandler attachment)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }

        outMess.addAttachment(name, attachment);
        this.setOutMessage(outMess);
    }

    @Override
    public void setOutMessageAttachments(final Map<String, DataHandler> attachments)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        addAttachments(attachments, outMess);
        this.setOutMessage(outMess);
    }

    @Override
    public void setOutMessageContent(final Document outContent) throws MessagingException {
        this.setOutMessageContent(SourceUtil.createSource(outContent));
    }

    @Override
    public void setOutMessageContent(final InputStream outContent) throws MessagingException {
        this.setOutMessageContent(new StreamSource(outContent));
    }

    @Override
    public void setOutMessageContent(final Source outContent) throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        outMess.setContent(outContent);
        this.setOutMessage(outMess);
    }

    @Override
    public void setOutMessageProperties(final Map<String, Object> properties)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        addMessageProperties(properties, outMess);
        this.setOutMessage(outMess);
    }

    @Override
    public void setOutMessageProperty(final String name, final Object value)
            throws MessagingException {
        NormalizedMessage outMess = this.getOutMessage();
        if (outMess == null) {
            outMess = this.exchange.createMessage();
        }
        outMess.setProperty(name, value);
        this.setOutMessage(outMess);
    }

    @Override
    public void setOutSubject(final Subject subject) {
        this.getOutMessage().setSecuritySubject(subject);
    }

    @Override
    public void setProperty(final String name, final Object obj) {
        this.exchange.setProperty(name, obj);
    }

    @Override
    public void setService(final QName service) {
        this.exchange.setService(service);
    }

    @Override
    public void setStatus(final ExchangeStatus status) throws MessagingException {
        this.exchange.setStatus(status);
    }

    /**
     * Check that the exchange MEP allows the specified reference. This only
     * works for the JBI predefined MEPs. It check that reference is IN or OUT.
     * Other reference or MEP will cause an exception
     * 
     * @param name
     * @throws MessagingException
     *             the MEP does not allow the reference
     */
    protected void checkPatternMatching(final String name) throws MessagingException {
        if (name == null) {
            throw new MessagingException("The reference name is not defined.");
        }

        if (OUT_MESSAGE_NAME.equalsIgnoreCase(name)
                && (MEPConstants.IN_ONLY_PATTERN.equals(this.exchange.getPattern()) || MEPConstants.ROBUST_IN_ONLY_PATTERN
                        .equals(this.exchange.getPattern()))) {
            throw new MessagingException("the MessageExchange state does not allow this operation.");
        }
    }

    public void setCurrentProcessFlowAttributes(final FlowAttributes fa) {
        this.currentProcessFA = fa;
    }

    /**
     * Contains the flow attributes associated to the processing of the current exchange.
     * 
     * When acting as a provider, it is set when the exchange is first received to a new step following the one of the
     * received exchange and is kept in it until the end of the processing.
     * 
     * When acting as a consumer, it is set when the exchange is sent and received.
     */
    public FlowAttributes getCurrentProcessFlowAttributes() {
        return currentProcessFA;
    }
}
