/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.monitoring;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicLong;

import javax.management.MBeanNotificationInfo;
import javax.management.NotificationBroadcasterSupport;

import org.ow2.petals.component.framework.clientserver.api.monitoring.exception.MonitoringProbeNotInitializedException;
import org.ow2.petals.component.framework.clientserver.api.monitoring.exception.MonitoringProbeNotStartedException;
import org.ow2.petals.component.framework.clientserver.api.monitoring.exception.MonitoringServiceException;
import org.ow2.petals.component.framework.monitoring.defect.JmxDefectCreator;
import org.ow2.petals.component.framework.monitoring.defect.MessageExchangeProcessorThreadPoolDefectCreator;
import org.ow2.petals.component.framework.monitoring.key.ServiceProviderKey;
import org.ow2.petals.component.framework.monitoring.probe.MessageExchangeAcceptorPoolProbe;
import org.ow2.petals.component.framework.monitoring.probe.impl.MessageExchangeAcceptorPoolProbeImpl;
import org.ow2.petals.probes.api.KeyedProbesFactory;
import org.ow2.petals.probes.api.KeyedProbesFactoryBuilder;
import org.ow2.petals.probes.api.MacroProbesFactory;
import org.ow2.petals.probes.api.MacroProbesFactoryBuilder;
import org.ow2.petals.probes.api.ProbesFactory;
import org.ow2.petals.probes.api.ProbesFactoryBuilder;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.key.ExecutionStatusProbeKey;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.CounterGaugeProbe;
import org.ow2.petals.probes.api.probes.DurationProbe;
import org.ow2.petals.probes.api.probes.KeyedCounterProbeWithExecutionStatus;
import org.ow2.petals.probes.api.probes.KeyedResponseTimeProbe;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;
import org.ow2.petals.probes.api.probes.ResponseTimeRelativeValues;
import org.ow2.petals.probes.api.probes.macro.ThreadPoolProbe;

/**
 * The monitoring MBean of the component
 * 
 * @author Christophe DENEUX - Linagora
 */
public class Monitoring extends NotificationBroadcasterSupport implements MonitoringMBean {

    /**
     * Defect notification sequence number
     */
    private final AtomicLong notificationSeqNum = new AtomicLong(0);

    // --- Probes --- //

    /**
     * The probe gauging the message exchange acceptors.
     */
    private final MessageExchangeAcceptorPoolProbe probeMessageExchangeAcceptorPool;

    /**
     * The probe gauging the working message exchange acceptors.
     */
    private final CounterGaugeProbe probeWorkingMessageExchangeAcceptors;

    /**
     * The probe about the processing duration, in nanoseconds, of message exchange acceptors
     */
    private final DurationProbe probeMessageExchangeAcceptorDuration;

    /**
     * The macro probe about the message exchange processor thread pool.
     */
    private final ThreadPoolProbe probeMessageExchangeProcessorThreadPool;

    /**
     * The probe about the processing duration of message exchange processors
     */
    private final DurationProbe probeMessageExchangeProcessorDuration;

    /**
     * The probe about response times of service providers
     */
    private final KeyedResponseTimeProbe<ServiceProviderKey> probeServiceProviderResponseTimes;

    /**
     * The pending start date items of the probe about response times of service providers
     */
    private final ConcurrentMap<String, KeyedStartDateItem<ServiceProviderKey>> serviceProviderStartDateItems = new ConcurrentHashMap<String, KeyedStartDateItem<ServiceProviderKey>>();

    /**
     * The probe counting invocations of service providers
     */
    private final KeyedCounterProbeWithExecutionStatus<ProbeKey> probeServiceProviderInvocations;

    /**
     * The pending invocation keys of the probe counting service provider invocations
     */
    private final ConcurrentMap<String, ServiceProviderKey> serviceProviderInvocationKeys = new ConcurrentHashMap<String, ServiceProviderKey>();

    // --- Defect creators --- //

    /**
     * The defect creator associated to the defect 'exhaustion of the message exchange processors thread pool'.
     */
    private JmxDefectCreator messageExchangeProcessorThreadPoolDefectCreator;

    /**
     * Creates the monitoring MBean
     * 
     * @param responseTimeProbesTimer
     *            The timer used as sampler by response time probes. <b>The caller is responsible to cancel the timer to
     *            free resources attached</b>.
     * @param samplePeriod
     *            The duration of a sample, in milliseconds.
     * @throws NoProbesFactoryFoundException
     *             No probe implementation found in the classloader
     * @throws MultipleProbesFactoriesFoundException
     *             Several probe implementations found in the classloader
     */
    public Monitoring(final Timer responseTimeProbesTimer, final long samplePeriod)
            throws MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final MacroProbesFactoryBuilder macroProbesFactoryBuilder = new MacroProbesFactoryBuilder();
        final MacroProbesFactory macroProbesFactory = macroProbesFactoryBuilder.getMacroProbesFactory();

        this.messageExchangeProcessorThreadPoolDefectCreator = new MessageExchangeProcessorThreadPoolDefectCreator(this);
        this.probeMessageExchangeProcessorThreadPool = macroProbesFactory
                .createThreadPoolProbe(this.messageExchangeProcessorThreadPoolDefectCreator);

        final ProbesFactoryBuilder<Long, Long> longProbesFactoryBuilder = new ProbesFactoryBuilder<Long, Long>();
        final ProbesFactory<Long, Long> longProbesFactory = longProbesFactoryBuilder.getProbesFactory();

        this.probeMessageExchangeProcessorDuration = longProbesFactory.createDurationProbe(responseTimeProbesTimer,
                samplePeriod);

        this.probeMessageExchangeAcceptorPool = new MessageExchangeAcceptorPoolProbeImpl(this);
        this.probeWorkingMessageExchangeAcceptors = longProbesFactory.createCounterGaugeProbe(0);
        this.probeMessageExchangeAcceptorDuration = longProbesFactory.createDurationNanoProbe(responseTimeProbesTimer,
                samplePeriod);
        // Keyed probe for service providers
        final KeyedProbesFactoryBuilder<ServiceProviderKey, Long, Long> serviceProvidersProbesFactoryBuilder = new KeyedProbesFactoryBuilder<ServiceProviderKey, Long, Long>();
        final KeyedProbesFactory<ServiceProviderKey, Long, Long> serviceProvidersProbesFactory = serviceProvidersProbesFactoryBuilder
                .getKeyedProbesFactory();
        this.probeServiceProviderResponseTimes = serviceProvidersProbesFactory.createKeyedResponseTimeProbe(
                responseTimeProbesTimer, samplePeriod);
        this.probeServiceProviderInvocations = serviceProvidersProbesFactory
                .createKeyedCounterProbeWithExecutionStatus();

    }

    /**
     * @return A new sequence number, used to send JMX notification
     */
    public long getNotificationSeqNum() {
        return this.notificationSeqNum.getAndIncrement();
    }

    /**
     * @return The probe of the message exchange acceptor pool
     */
    public MessageExchangeAcceptorPoolProbe getMessageExchangeAcceptorPoolProbe() {
        return this.probeMessageExchangeAcceptorPool;
    }

    /**
     * @return The probe of the working message exchange acceptors
     */
    public CounterGaugeProbe getWorkingMessageExchangeAcceptorProbe() {
        return this.probeWorkingMessageExchangeAcceptors;
    }

    /**
     * @return The probe about the processing duration, in nanoseconds, of message exchange acceptors
     */
    public DurationProbe getMessageExchangeAcceptorDurationProbe() {
        return this.probeMessageExchangeAcceptorDuration;
    }

    /**
     * @return The probe of the message exchange processor thread pool
     */
    public ThreadPoolProbe getMessageExchangeProcessorThreadPoolProbe() {
        return this.probeMessageExchangeProcessorThreadPool;
    }

    /**
     * @return The probe about the processing duration of message exchange processors
     */
    public DurationProbe getMessageExchangeProcessorDurationProbe() {
        return this.probeMessageExchangeProcessorDuration;
    }

    /**
     * @return The probe about the response times of service providers
     */
    public KeyedResponseTimeProbe<ServiceProviderKey> getServiceProviderResponseTimesProbe() {
        return this.probeServiceProviderResponseTimes;
    }

    /**
     * @return The pending start date items of the probe about response times of service providers
     */
    public ConcurrentMap<String, KeyedStartDateItem<ServiceProviderKey>> getServiceProviderStartDateItems() {
        return this.serviceProviderStartDateItems;
    }

    /**
     * @return The probe counting the invocations of service providers
     */
    public KeyedCounterProbeWithExecutionStatus<ProbeKey> getServiceProviderInvocationsProbe() {
        return this.probeServiceProviderInvocations;
    }

    /**
     * @return The pending invocation keys of the probe counting service provider invocations
     */
    public ConcurrentMap<String, ServiceProviderKey> getServiceProviderInvocationKeys() {
        return this.serviceProviderInvocationKeys;
    }

    /**
     * Initialize all probes of the monitoring MBean
     * 
     * @throws ProbeInitializedException
     *             A probe is already initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeInitializationException
     *             An error occurs during a probe initialization
     */
    public void init() throws ProbeInitializedException, ProbeStartedException, ProbeInitializationException {

        this.probeMessageExchangeAcceptorPool.init();
        this.probeWorkingMessageExchangeAcceptors.init();
        this.probeMessageExchangeAcceptorDuration.init();
        this.probeMessageExchangeProcessorThreadPool.init();
        this.probeMessageExchangeProcessorDuration.init();
        this.probeServiceProviderResponseTimes.init();
        this.probeServiceProviderInvocations.init();

        this.doInit();
    }

    /**
     * Add here the initialization of all your probes
     * 
     * @throws ProbeInitializedException
     *             A probe is already initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeInitializationException
     *             An error occurs during a probe initialization
     * 
     */
    public void doInit() throws ProbeInitializedException, ProbeStartedException, ProbeInitializationException {
        // NOP
    }

    /**
     * Start all probes of the monitoring MBean
     * 
     * @throws ProbeStartupException
     *             An error occurs starting a probe
     * @throws ProbeStartedException
     *             A probe is already started
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     */
    public void start() throws ProbeNotInitializedException, ProbeStartedException, ProbeStartupException {
        this.probeMessageExchangeAcceptorPool.start();
        this.probeWorkingMessageExchangeAcceptors.start();
        this.probeMessageExchangeAcceptorDuration.start();
        this.probeMessageExchangeProcessorThreadPool.start();
        this.probeMessageExchangeProcessorDuration.start();
        this.probeServiceProviderResponseTimes.start();
        this.probeServiceProviderInvocations.start();

        this.doStart();
    }

    /**
     * Add here the start of your own probes
     * 
     * @throws ProbeNotInitializedException
     *             An error occurs starting a probe
     * @throws ProbeStartedException
     *             A probe is already started
     * @throws ProbeStartupException
     *             A probe is not initialized
     */
    protected void doStart() throws ProbeNotInitializedException, ProbeStartedException, ProbeStartupException {
        // NOP
    }

    /**
     * Stop all probes of the monitoring MBean
     * 
     * @throws ProbeStopException
     *             An error occurs during the stop of a probe
     * @throws ProbeNotStartedException
     *             A probe is not started
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * 
     */
    public void stop() throws ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException {
        this.probeMessageExchangeAcceptorPool.stop();
        this.probeWorkingMessageExchangeAcceptors.stop();
        this.probeMessageExchangeAcceptorDuration.stop();
        this.probeMessageExchangeProcessorThreadPool.stop();
        this.probeMessageExchangeProcessorDuration.stop();
        this.probeServiceProviderResponseTimes.stop();
        this.probeServiceProviderInvocations.stop();

        this.doStop();
    }

    /**
     * Add here the stop of all your probes
     * 
     * @throws ProbeStopException
     *             An error occurs during the stop of a probe
     * @throws ProbeNotStartedException
     *             A probe is not started
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * 
     */
    public void doStop() throws ProbeNotInitializedException, ProbeNotStartedException, ProbeStopException {
        // NOP
    }

    /**
     * Shutdown all probes of the monitoring MBean
     * 
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeShutdownException
     *             An error occurs during the shutdown of a probe
     * 
     */
    public void shutdown() throws ProbeShutdownException, ProbeStartedException, ProbeNotInitializedException {
        this.probeMessageExchangeAcceptorPool.shutdown();
        this.probeWorkingMessageExchangeAcceptors.shutdown();
        this.probeMessageExchangeAcceptorDuration.shutdown();
        this.probeMessageExchangeProcessorThreadPool.shutdown();
        this.probeMessageExchangeProcessorDuration.shutdown();
        this.probeServiceProviderResponseTimes.shutdown();
        this.probeServiceProviderInvocations.shutdown();

        this.doShutdown();
    }

    /**
     * Add here the shutdown of all you probes
     * 
     * @throws ProbeNotInitializedException
     *             A probe is not initialized
     * @throws ProbeStartedException
     *             A probe is started
     * @throws ProbeShutdownException
     *             An error occurs during the shutdown of a probe
     * 
     */
    public void doShutdown() throws ProbeShutdownException, ProbeStartedException, ProbeNotInitializedException {
        // NOP
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        final List<MBeanNotificationInfo> notifInfos = new ArrayList<MBeanNotificationInfo>();
        notifInfos.addAll(Arrays.asList(this.probeMessageExchangeAcceptorPool.getNotificationInfo()));
        notifInfos.add(this.messageExchangeProcessorThreadPoolDefectCreator.getNotificationInfo());

        notifInfos.addAll(this.addNotificationInfo());

        return notifInfos.toArray(new MBeanNotificationInfo[notifInfos.size()]);
    }

    /**
     * Add here all information about dedicated notification of the component
     */
    public List<MBeanNotificationInfo> addNotificationInfo() {
        return Collections.emptyList();
    }

    // Probe accessors //

    //
    // --- Message exchange acceptor thread pool metrics
    //

    @Override
    public int getMessageExchangeAcceptorThreadPoolMaxSize() throws MonitoringProbeNotInitializedException,
            MonitoringServiceException {
        try {
            return this.probeMessageExchangeAcceptorPool.getMaxValue();
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public int getMessageExchangeAcceptorThreadPoolCurrentSize() throws MonitoringProbeNotStartedException,
            MonitoringServiceException {
        try {
            return this.probeMessageExchangeAcceptorPool.getInstantValue();
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    @Override
    public long getMessageExchangeAcceptorCurrentWorking() throws MonitoringProbeNotStartedException,
            MonitoringServiceException {
        try {
            return this.probeWorkingMessageExchangeAcceptors.getCurrentValue();
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    @Override
    public long getMessageExchangeAcceptorMaxWorking() throws MonitoringProbeNotInitializedException,
            MonitoringServiceException {
        try {
            return this.probeWorkingMessageExchangeAcceptors.getMaxValue();
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public Long[] getMessageExchangeAcceptorAbsoluteDurations() throws MonitoringProbeNotInitializedException,
            MonitoringServiceException {
        try {
            final Long[] values = this.probeMessageExchangeAcceptorDuration.getConvertedAbsoluteResponseTimeValues()
                    .get(ExecutionStatus.SUCCEEDED.name());
            if (values == null) {
                return new Long[] { 0L, 0L, 0L };
            } else {
                return values;
            }
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public Long[] getMessageExchangeAcceptorRelativeDurations() throws MonitoringProbeNotStartedException,
            MonitoringServiceException {
        try {
            final Long[] values = this.probeMessageExchangeAcceptorDuration.getConvertedRelativeResponseTimeValues()
                    .get(ExecutionStatus.SUCCEEDED.name());
            if (values == null) {
                return new Long[] { 0L, 0L, 0L, 0L, 0L, 0L };
            } else {
                return values;
            }
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    //
    // --- Message exchange processor thread pool metrics
    //

    @Override
    public long getMessageExchangeProcessorThreadPoolMaxSize() throws MonitoringServiceException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolMaxSize();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolMinSize() throws MonitoringServiceException {
        return this.probeMessageExchangeProcessorThreadPool.getThreadPoolMinSize();
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolActiveThreadsMax()
            throws MonitoringProbeNotInitializedException, MonitoringServiceException {
        try {
            return this.probeMessageExchangeProcessorThreadPool.getThreadPoolActiveThreadsMax();
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolActiveThreadsCurrent()
            throws MonitoringProbeNotStartedException, MonitoringServiceException {
        try {
            return this.probeMessageExchangeProcessorThreadPool.getThreadPoolActiveThreadsCurrent();
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolIdleThreadsMax() throws MonitoringProbeNotInitializedException,
            MonitoringServiceException {
        try {
            return this.probeMessageExchangeProcessorThreadPool.getThreadPoolIdleThreadsMax();
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolIdleThreadsCurrent() throws MonitoringProbeNotStartedException,
            MonitoringServiceException {
        try {
            return this.probeMessageExchangeProcessorThreadPool.getThreadPoolIdleThreadsCurrent();
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolQueuedRequestsMax() throws MonitoringProbeNotInitializedException,
            MonitoringServiceException {
        try {
            return this.probeMessageExchangeProcessorThreadPool.getThreadPoolQueuedRequestsMax();
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public long getMessageExchangeProcessorThreadPoolQueuedRequestsCurrent() throws MonitoringProbeNotStartedException,
            MonitoringServiceException {
        try {
            return this.probeMessageExchangeProcessorThreadPool.getThreadPoolQueuedRequestsCurrent();
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    //
    // --- Message exchange processors (processing duration)
    //

    @Override
    public Long[] getMessageExchangeProcessorAbsoluteDurations() throws MonitoringProbeNotInitializedException,
            MonitoringServiceException {
        try {
            final Long[] values = this.probeMessageExchangeProcessorDuration.getConvertedAbsoluteResponseTimeValues()
                    .get(ExecutionStatus.SUCCEEDED.name());
            if (values == null) {
                return new Long[] { 0L, 0L, 0L };
            } else {
                return values;
            }
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

    @Override
    public Long[] getMessageExchangeProcessorRelativeDurations() throws MonitoringProbeNotStartedException,
            MonitoringServiceException {
        try {
            final Long[] values = this.probeMessageExchangeProcessorDuration.getConvertedRelativeResponseTimeValues()
                    .get(ExecutionStatus.SUCCEEDED.name());
            if (values == null) {
                return new Long[] { 0L, 0L, 0L, 0L, 0L, 0L };
            } else {
                return values;
            }
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    //
    // --- Response times of service providers
    //

    @Override
    public Map<String[], Long[]> getServiceProviderInvocationsResponseTimeAbs()
            throws MonitoringProbeNotInitializedException, MonitoringServiceException {
        try {
            return this.probeServiceProviderResponseTimes.getConvertedAbsoluteResponseTimeValues();
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);

        }
    }

    @Override
    public Map<String[], Long[]> getServiceProviderInvocationsResponseTimeRel()
            throws MonitoringProbeNotStartedException, MonitoringServiceException {
        try {
            return this.probeServiceProviderResponseTimes.getConvertedRelativeResponseTimeValues();
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    @Override
    public Map<String[], Long[]> getServiceProviderInvocationsResponseTimeRelGroupedByItfSvcOpAndExecStatus()
            throws MonitoringProbeNotStartedException, MonitoringServiceException {
        try {
            return aggregateValues(this.probeServiceProviderResponseTimes.getRelativeResponseTimeValues());
        } catch (final ProbeNotStartedException e) {
            throw new MonitoringProbeNotStartedException(e);
        }
    }

    protected static <PK extends ProbeKey> Map<String[], Long[]> aggregateValues(
            final Map<ExecutionStatusProbeKey<PK>, ResponseTimeRelativeValues> responseTimes) {

        final Map<String[], Long[]> agregatedResults = new HashMap<String[], Long[]>();

        for (final Map.Entry<ExecutionStatusProbeKey<PK>, ResponseTimeRelativeValues> e : responseTimes.entrySet()) {
            final String[] newKey = e.getKey().getOriginalKey().toStringArray();
            if (agregatedResults.containsKey(newKey)) {
                // A new key found --> Aggregation
                final Long[] newValues = agregatedResults.get(newKey);
                final ResponseTimeRelativeValues values = e.getValue();
                newValues[0] = Math.max(values.getMax(), newValues[0]);
                newValues[1] = Math.min(values.getMin(), newValues[1]);
                newValues[2] += values.getResponseTimesSum();
                newValues[3] += values.getResponseTimesNb();
                // TODO: How to aggregate 2 values of percentiles ?
                // VNO: it's impossible without the original data, so this information is bogus!
                newValues[3] = values.getPercent10();
                newValues[4] = values.getPercent50();
                newValues[5] = values.getPercent90();
                agregatedResults.put(newKey, newValues);

            } else {
                final ResponseTimeRelativeValues values = e.getValue();
                final Long[] newValues = new Long[7];
                newValues[0] = values.getMax();
                newValues[1] = values.getMin();
                newValues[2] = values.getResponseTimesSum();
                newValues[3] = values.getResponseTimesNb();
                newValues[4] = values.getPercent10();
                newValues[5] = values.getPercent50();
                newValues[6] = values.getPercent90();
                agregatedResults.put(newKey, newValues);
            }
        }

        final Map<String[], Long[]> results = new HashMap<String[], Long[]>();
        for (final Map.Entry<String[], Long[]> entry : agregatedResults.entrySet()) {
            final Long[] values = entry.getValue();
            final Long[] newValues = new Long[6];
            newValues[0] = values[0];
            newValues[1] = values[2] / values[3];
            newValues[2] = values[1];
            newValues[3] = values[4];
            newValues[4] = values[5];
            newValues[5] = values[6];

            results.put(entry.getKey(), newValues);
        }
        return results;
    }

    //
    // --- Counters of service providers
    //
    @Override
    public Map<String[], Long> getServiceProviderInvocations() throws MonitoringProbeNotInitializedException,
            MonitoringServiceException {
        try {
            return this.probeServiceProviderInvocations.getConvertedValues();
        } catch (final ProbeNotInitializedException e) {
            throw new MonitoringProbeNotInitializedException(e);
        }
    }

}
