/**
 * Copyright (c) 2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.monitoring;

import java.util.logging.Logger;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessagingException;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.monitoring.key.ServiceProviderKey;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeKeyMissingException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ResponseTimeCollectionStoppedException;
import org.ow2.petals.probes.api.exceptions.StartDateItemLostException;
import org.ow2.petals.probes.api.exceptions.StartDateItemUnknownException;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;

public class MonitoringHelper {

    private MonitoringHelper() {
        // utility class
    }

    /**
     * Update monitoring probes when the service provider invocation ends, using the given execution status.
     * 
     * @param exchange
     *            The message exchange associated to the service provider invocation.
     * @param executionStatus
     *            The execution status to use to update stats
     * @param monitoring
     *            The Monitoring MBean containing all the monitoring probes
     * @param logger
     */
    public static void updateStatsOnEndOfServiceProviderInvocation(final Exchange exchange,
            final ExecutionStatus executionStatus, final Monitoring monitoring, final Logger logger) {

        final String exchangeId = exchange.getExchangeId();

        try {
            final KeyedStartDateItem<ServiceProviderKey> item = monitoring.getServiceProviderStartDateItems()
                    .remove(exchangeId);
            if (item != null) {
                monitoring.getServiceProviderResponseTimesProbe().endsExecution(item, executionStatus);
            } else {
                logger.warning("Start date item not found into the monitoring map. Metric values can be uncorrect.");
            }
        } catch (final StartDateItemUnknownException e2) {
            logger.warning(
                    "Start date item not found into the probe about response times of service providers. Metric values can be uncorrect.");
        } catch (final ProbeNotStartedException e2) {
            logger.warning(
                    "The probe about response times of service providers is not started. Metric values can be uncorrect.");
        } catch (final ResponseTimeCollectionStoppedException e2) {
            logger.warning(
                    "The sample collecting of the probe about response times of service providers is stopped. Metric values can be uncorrect.");
        }

        try {
            final ServiceProviderKey item = monitoring.getServiceProviderInvocationKeys().remove(exchangeId);
            if (item != null) {
                monitoring.getServiceProviderInvocationsProbe().move(item, executionStatus);
            } else {
                logger.warning("Service provider key not found in the monitoring map. Metric values can be uncorrect.");
            }
        } catch (final ProbeKeyMissingException e) {
            logger.warning(
                    "The probe counting service provider invocations couldn't retrieve the key. Metric values can be uncorrect.");
        } catch (final ProbeNotStartedException e) {
            logger.warning(
                    "The probe counting service provider invocations is not started. Metric values can be uncorrect.");
        }
    }

    /**
     * Update monitoring probes when the service provider invocation starts.
     * 
     * @param exchange
     *            The message exchange associated to the service provider invocation.
     * @param monitoring
     *            The Monitoring MBean containing all the monitoring probes
     * @param logger
     * @throws MessagingException
     */
    public static void updateStatsOnStartOfServiceProviderInvocation(final Exchange exchange,
            final Monitoring monitoring, final Logger logger) throws MessagingException {

        final String exchangeId = exchange.getExchangeId();

        final ServiceProviderKey serviceProviderKey = new ServiceProviderKey(exchange.getInterfaceName(),
                exchange.getService(), exchange.getOperation(), exchange.getPattern());
        try {
            monitoring.getServiceProviderInvocationKeys().put(exchangeId, serviceProviderKey);
            monitoring.getServiceProviderInvocationsProbe().incPending(serviceProviderKey);
        } catch (final ProbeNotStartedException e1) {
            logger.warning(
                    "The probe counting service provider invocations is not started. Metric values can be uncorrect.");
        }
        try {
            monitoring.getServiceProviderStartDateItems().put(exchangeId,
                    monitoring.getServiceProviderResponseTimesProbe().newExecution(serviceProviderKey));
        } catch (final StartDateItemLostException e2) {
            logger.warning(
                    "Unable to add a new start date item into the probe about response times of service providers. Metric values can be uncorrect.");
        } catch (final ProbeNotStartedException e2) {
            logger.warning(
                    "The probe about response times of service providers is not started. Metric values can be uncorrect.");
        } catch (final ResponseTimeCollectionStoppedException e2) {
            logger.warning(
                    "The sample collecting of the probe about response times of service providers is stopped. Metric values can be uncorrect.");
        }
    }

    /**
     * Update monitoring probes when the service provider invocation ends.
     * 
     * @param exchange
     *            The message exchange associated to the service provider invocation.
     * @param monitoring
     *            The Monitoring MBean containing all the monitoring probes
     * @param logger
     */
    public static void updateStatsOnEndOfServiceProviderInvocation(final Exchange exchange, final Monitoring monitoring,
            final Logger logger) {

        final ExecutionStatus executionStatus;
        if (exchange.getStatus() == ExchangeStatus.ERROR) {
            logger.fine("Ends a start date item in ERROR");
            executionStatus = ExecutionStatus.ERROR;
        } else if (exchange.getFault() != null) {
            executionStatus = ExecutionStatus.FAULT;
            logger.fine("Ends a start date item in FAULT");
        } else {
            executionStatus = ExecutionStatus.SUCCEEDED;
            logger.fine("Ends a start date item in SUCCESS");
        }

        updateStatsOnEndOfServiceProviderInvocation(exchange, executionStatus, monitoring, logger);

    }
}
