/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.monitoring.probe.impl;

import java.util.Collection;

import javax.management.MBeanNotificationInfo;

import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.component.framework.monitoring.defect.JmxDefectCreator;
import org.ow2.petals.component.framework.monitoring.defect.MessageExchangeAcceptorPoolAcceptorDeathDefectCreator;
import org.ow2.petals.component.framework.monitoring.defect.MessageExchangeAcceptorPoolDefectDetector;
import org.ow2.petals.component.framework.monitoring.defect.MessageExchangeAcceptorPoolExhaustedDefectCreator;
import org.ow2.petals.component.framework.monitoring.probe.MessageExchangeAcceptorPoolProbe;
import org.ow2.petals.component.framework.monitoring.sensor.MessageExchangeAcceptorPoolSensor;
import org.ow2.petals.component.framework.process.MessageExchangeAcceptor;
import org.ow2.petals.probes.api.ProbesFactory;
import org.ow2.petals.probes.api.ProbesFactoryBuilder;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.GaugeProbe;

/**
 * The implementation of probe {@link MessageExchangeAcceptorPoolProbe}
 * 
 * @author Christophe DENEUX - Lianagora
 * 
 */
public class MessageExchangeAcceptorPoolProbeImpl implements MessageExchangeAcceptorPoolProbe {

    /**
     * The defect detector attached to the message exchange acceptor thread
     * pool.
     */
    private final MessageExchangeAcceptorPoolDefectDetector messageExchangeAcceptorPoolDefectDetector;

    /**
     * The defect creator associated to the defect 'exhaustion of the message
     * exchange acceptor pool'.
     */
    private final JmxDefectCreator messageExchangeAcceptorPoolExhaustedDefectCreator;

    /**
     * The defect creator associated to the defect 'an acceptor thread is dead unexpectedly'.
     */
    private final JmxDefectCreator messageExchangeAcceptorPoolAcceptorDeathDefectCreator;

    /**
     * The sensor attached to the message exchange acceptor thread pool.
     */
    private final MessageExchangeAcceptorPoolSensor messageExchangeAcceptorPoolSensor;

    /**
     * The real probe gauging the message exchange acceptors.
     */
    private final GaugeProbe<Integer, Integer> probeMessageExchangeAcceptorPool;

    /**
     * 
     * @param monitoringMbean
     * @throws MultipleProbesFactoriesFoundException
     *             Several probes factories are found.
     * @throws NoProbesFactoryFoundException
     *             No probes factory is found.
     */
    public MessageExchangeAcceptorPoolProbeImpl(final Monitoring monitoringMbean)
            throws MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final ProbesFactoryBuilder<Integer, Integer> intProbesFactoryBuilder = new ProbesFactoryBuilder<Integer, Integer>();
        final ProbesFactory<Integer, Integer> intProbesFactory = intProbesFactoryBuilder.getProbesFactory();
        this.messageExchangeAcceptorPoolSensor = new MessageExchangeAcceptorPoolSensor();
        this.messageExchangeAcceptorPoolExhaustedDefectCreator = new MessageExchangeAcceptorPoolExhaustedDefectCreator(
                monitoringMbean);
        this.messageExchangeAcceptorPoolAcceptorDeathDefectCreator = new MessageExchangeAcceptorPoolAcceptorDeathDefectCreator(
                monitoringMbean);
        this.messageExchangeAcceptorPoolDefectDetector = new MessageExchangeAcceptorPoolDefectDetector(
                this.messageExchangeAcceptorPoolExhaustedDefectCreator,
                this.messageExchangeAcceptorPoolAcceptorDeathDefectCreator);
        this.probeMessageExchangeAcceptorPool = intProbesFactory.createGaugeProbe(
                this.messageExchangeAcceptorPoolSensor,
                this.messageExchangeAcceptorPoolDefectDetector);
    }

    @Override
    public void setAcceptorPool(final Collection<MessageExchangeAcceptor> acceptorPool) {
        this.messageExchangeAcceptorPoolSensor.setAcceptorPool(acceptorPool);
        this.messageExchangeAcceptorPoolDefectDetector.setAcceptorPool(acceptorPool);
    }

    @Override
    public void setAcceptorPoolSize(final long acceptorPoolSize) {
        this.messageExchangeAcceptorPoolDefectDetector.setAcceptorPoolSize(acceptorPoolSize);
    }

    @Override
    public void init() throws ProbeInitializedException, ProbeStartedException,
            ProbeInitializationException {
        this.probeMessageExchangeAcceptorPool.init();
    }

    @Override
    public void start() throws ProbeNotInitializedException, ProbeStartedException,
            ProbeStartupException {
        this.probeMessageExchangeAcceptorPool.start();
    }

    @Override
    public void stop() throws ProbeNotInitializedException, ProbeNotStartedException,
            ProbeStopException {
        this.probeMessageExchangeAcceptorPool.stop();

    }

    @Override
    public void shutdown() throws ProbeShutdownException, ProbeStartedException,
            ProbeNotInitializedException {
        this.probeMessageExchangeAcceptorPool.shutdown();
    }

    @Override
    public void pick() throws ProbeNotStartedException {
        this.probeMessageExchangeAcceptorPool.pick();
    }

    @Override
    public Integer getInstantValue() throws ProbeNotStartedException {
        return this.probeMessageExchangeAcceptorPool.getInstantValue();
    }

    @Override
    public Integer getMaxValue() throws ProbeNotInitializedException {
        return this.probeMessageExchangeAcceptorPool.getMaxValue();
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] {
                this.messageExchangeAcceptorPoolExhaustedDefectCreator.getNotificationInfo(),
                this.messageExchangeAcceptorPoolAcceptorDeathDefectCreator.getNotificationInfo() };
    }

}
