/**
 * Copyright (c) 2006-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.lang.Thread.UncaughtExceptionHandler;
import java.util.Collections;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.monitoring.probe.MessageExchangeAcceptorPoolProbe;

/**
 * This class helps a Component to listen to {@link MessageExchange}s on the
 * DeliveryChannel (accept() method).
 * <p>
 * It starts or stops a pool of thread that listen to new incoming JBI from the
 * bus. {@link MessageExchange}s.
 * </p>
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe DENEUX - EBM WebSourcing
 * @since 3.0
 */
public class JBIAcceptorManager implements NotificationListener {

    static final String GROUP_SUFFIX = " - JBI Acceptors";

    private final Set<MessageExchangeAcceptor> acceptorPool = Collections
            .newSetFromMap(new ConcurrentHashMap<MessageExchangeAcceptor, Boolean>());

    private final AtomicInteger threadNumber = new AtomicInteger(1);

    private final JBIProcessorManager jbiProcessorManager;

    private final RuntimeConfigurationNotifier runtimeConfiguration;
    
    /**
     * The message exchange acceptor pool probe
     */
    private final MessageExchangeAcceptorPoolProbe messageExchangeAcceptorPoolProbe;

    private final AbstractComponent component;

    private final Logger logger;

    private long acceptorStopMaxWait;

    private ThreadGroup group;

    public JBIAcceptorManager(final AbstractComponent component, final JBIProcessorManager jbiProcessorManager,
            final RuntimeConfigurationNotifier runtimeConfiguration, final Logger logger) {
        this.component = component;
        this.jbiProcessorManager = jbiProcessorManager;
        this.runtimeConfiguration = runtimeConfiguration;
        this.logger = logger;
        this.messageExchangeAcceptorPoolProbe = component.getMonitoringBean()
                .getMessageExchangeAcceptorPoolProbe();
        this.messageExchangeAcceptorPoolProbe.setAcceptorPool(this.acceptorPool);

        this.runtimeConfiguration.addNotificationListener(this, null, null);
    }

    /**
     * Initialize and start threads which listen to {@link MessageExchange}s
     * 
     * @throws JBIException
     * 
     */
    public void start() throws JBIException {
        this.logger.fine("Creating JBI Acceptor pool:");

        this.group = new ThreadGroup(component.getContext().getComponentName() + GROUP_SUFFIX);

        final int corePoolSize = this.component.getJbiComponentDescriptor().getComponent().getAcceptorPoolSize()
                .getValue();

        this.acceptorStopMaxWait = this.component.getJbiComponentDescriptor().getComponent().getAcceptorStopMaxWait()
                .getValue();

        this.logger.fine("\t- core pool size: " + corePoolSize);
        this.logger.fine("\t- stop max wait: " + this.acceptorStopMaxWait);

        for (int i = 0; i < corePoolSize; i++) {
            createAndStartAcceptor();
        }

        this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool.size());
    }

    private void createAndStartAcceptor() {
        final String name = this.component.getContext().getComponentName() + " - JBI Acceptor Thread #"
                + this.threadNumber.getAndIncrement();

        this.logger.log(Level.FINE, "Creating a new thread: " + name);

        final MessageExchangeAcceptor acceptorThread = new MessageExchangeAcceptor(this.component,
                this.jbiProcessorManager, this.logger, this.group, name);

        acceptorThread.setDaemon(true);

        acceptorThread.setUncaughtExceptionHandler(new UncaughtExceptionHandler() {
            @Override
            public void uncaughtException(final Thread t, final Throwable e) {
                JBIAcceptorManager.this.logger.log(Level.SEVERE,
                        t.getName() + " threw an uncaught exception. Will start a new thread to replace it.", e);
                JBIAcceptorManager.this.acceptorPool.remove(t);
                createAndStartAcceptor();
            }
        });
        acceptorThread.start();

        this.acceptorPool.add(acceptorThread);
    }

    /**
     * Stop threads that listen to {@link MessageExchange}s
     * 
     */
    public void stop() {

        // so that it destroy itself when there is no thread left
        this.group.setDaemon(true);

        // let's first end them
        for (final MessageExchangeAcceptor acceptor : this.acceptorPool) {
            acceptor.ends();
        }

        // and then wait for them to stop
        final Iterator<MessageExchangeAcceptor> itAcceptorPool = this.acceptorPool.iterator();
        while (itAcceptorPool.hasNext()) {
            final MessageExchangeAcceptor acceptorThread = itAcceptorPool.next();
            itAcceptorPool.remove();
            this.stopAcceptorThread(acceptorThread);
        }

        if (this.group.activeCount() > 0) {
            this.logger.warning("All thread in the group '" + this.group.getName()
                    + "' should have ended after interrupt but have not yet... " + "Maybe there is a bug in the CDK.");
        }

        this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool.size());
    }

    public void shutdown() throws JBIException {
        try {
            this.runtimeConfiguration.removeNotificationListener(this, null, null);
        } catch (final ListenerNotFoundException e) {
            throw new JBIException("Error unregistering the runtime configuration listener", e);
        }
    }

    private void stopAcceptorThread(final MessageExchangeAcceptor acceptorThread) {
        try {
            // Wait the end of the pending task
            acceptorThread.join(this.acceptorStopMaxWait);

            // we need to interrupt it in case it is waiting on the delivery channel
            acceptorThread.interrupt();

            // now we are sure it should be ending very soon, but let's use a timeout in case
            // there is a problem in order not to block all and to be able to log it
            acceptorThread.join(100);
            if (acceptorThread.isAlive()) {
                this.logger
                        .warning(acceptorThread.getName() + " should have ended after interrupt but has not (yet?)...");
            }
        } catch (InterruptedException e) {
            this.logger.log(Level.WARNING, "The stop of one acceptor of the acceptor thread pool was interrupted", e);
        }
    }

    int getPoolSize() {
        return this.acceptorPool.size();
    }

    long getStopMaxWait() {
        return this.acceptorStopMaxWait;
    }

    /**
     * Reset the core Pool size.
     * 
     * @param size
     */
    private void setCorePoolSize(final int size) {
        if (this.acceptorPool != null) {
            this.logger.info("Reset the acceptor thread pool size to " + size);
            if (size > this.acceptorPool.size()) {
                final int nbThreadToCreate = size - this.acceptorPool.size();
                for (int i = 0; i < nbThreadToCreate; i++) {
                    createAndStartAcceptor();
                    this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool
                            .size());
                }
            } else if (size < this.acceptorPool.size()) {
                int nbThreadToStop = this.acceptorPool.size() - size;
                final Iterator<MessageExchangeAcceptor> itAcceptorPool = this.acceptorPool
                        .iterator();
                while (itAcceptorPool.hasNext() && nbThreadToStop > 0) {
                    final MessageExchangeAcceptor acceptorThread = itAcceptorPool.next();
                    itAcceptorPool.remove();
                    this.messageExchangeAcceptorPoolProbe.setAcceptorPoolSize(this.acceptorPool
                            .size());
                    nbThreadToStop--;
                    this.stopAcceptorThread(acceptorThread);
                }
            } else {
                this.logger.fine("Unchange value, so no acceptor thread created or stopped !");
            }
        }
    }

    /**
     * Reset the stop max wait time of the acceptor pool
     */
    private void setStopMaxWait(final long time) {
        this.logger.info("Reset the acceptor stop max wait time to " + time);
        if (time <= 0) {
            final String warningMessage = String.format(
                    "Invalid value used (less than or equals to zero) to configure the acceptor stop max wait time: %d",
                    time);
            this.logger.warning(warningMessage);
        } else {
            this.acceptorStopMaxWait = time;
        }
    }

    ///////////////////////////////////////////////////////
    // Implementation of methods of NotificationListener //
    ///////////////////////////////////////////////////////

    @Override
    public void handleNotification(final Notification notification, final Object obj) {
        if (notification instanceof AttributeChangeNotification) {
            final AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
            final String attrName = attrNotif.getAttributeName();
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.log(Level.FINE,
                        "Notification received by " + this.getClass().getName() + " for attribute : " + attrName);
            }
            switch (attrName) {
                case "acceptorPoolSize": {
                    final Integer newVal = (Integer) attrNotif.getNewValue();
                    this.setCorePoolSize(newVal);
                    break;
                }
                case "acceptorStopMaxWait": {
                    final Long newVal = (Long) attrNotif.getNewValue();
                    this.setStopMaxWait(newVal);
                    break;
                }
                default: {
                    this.logger.log(Level.FINE,
                            "Notification ignored by " + this.getClass().getName() + " for attribute : " + attrName);
                }
            }
        }
    }
}
