/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.component.framework.process.JBIProcessorThreadPoolExecutor.ExchangeRunnable;
import org.ow2.petals.component.framework.process.JBIProcessorThreadPoolExecutor.RuntimeInterruptedException;
import org.ow2.petals.component.framework.process.async.AsyncContextExchange;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.probes.macro.ThreadPoolProbe;


/**
 * The JBI processor manager is in charge of JBI Message processor threads and instances.
 * @author Christophe Hamerling - EBM WebSourcing
 * @since 3.0
 */
public class JBIProcessorManager implements NotificationListener {

    static final String GROUP_SUFFIX = " - JBI Processors";

    /**
     * The message exchange processor thread pool
     */
    private ThreadPoolExecutor processorThreadpool;

    private ThreadGroup group;

    private int retryMaxTry;

    private long retryBaseDuration;

    private long processorStopMaxWait;

    /**
     * The probe attached to the message exchange processor thread pool
     */
    private final ThreadPoolProbe processorThreadPoolProbe;

    private final AbstractComponent component;

    private final RuntimeConfigurationNotifier runtimeConfiguration;

    private final Logger logger;

    private final List<JBIProcessorManagerObserver> jbiProccesorManagerObservers = new ArrayList<>();

    public JBIProcessorManager(final AbstractComponent component,
            final RuntimeConfigurationNotifier runtimeConfiguration, final Logger logger) {
        this.component = component;
        this.runtimeConfiguration = runtimeConfiguration;

        final Monitoring monitoringBean = component.getMonitoringBean();
        this.processorThreadPoolProbe = monitoringBean.getMessageExchangeProcessorThreadPoolProbe();
        this.logger = logger;

        this.runtimeConfiguration.addNotificationListener(this, null, null);
    }

    public void shutdown() throws JBIException {
        try {
            this.runtimeConfiguration.removeNotificationListener(this, null, null);
        } catch (final ListenerNotFoundException e) {
            throw new JBIException("Error unregistering the runtime configuration listener", e);
        }
    }

    /**
     * Start the Processor Thread Pool.
     */
    public void start() throws JBIException {
        this.logger.fine("Creating JBI Processor pool:");
        
        // Core pool size
        final int coreProcessorPoolSize = this.component.getJbiComponentDescriptor().getComponent()
                .getProcessorPoolSize().getValue();

        // Max pool size
        int maxProcessorPoolSize = this.component.getJbiComponentDescriptor().getComponent()
                .getProcessorMaxPoolSize().getValue();

        // Compare core pool size to the max pool size. The max pool size MUST
        // be upper or equals to the core pool size.
        if (coreProcessorPoolSize > maxProcessorPoolSize) {
            // Incoherent values -> max pool size is increased to match the
            // value of core pool size
            final int previousMax = maxProcessorPoolSize;
            maxProcessorPoolSize = coreProcessorPoolSize;
            this.logger.warning("The value of 'processor-max-pool-size' (" + previousMax
                    + ") is invalid, smaller than 'processor-pool-size' (" + coreProcessorPoolSize
                            + "). The value of 'processor-max-pool-size' is set to the value of 'processor-pool-size'");
        }

        // Keep alive time
        final int processorPoolKeepAliveTime = this.component.getJbiComponentDescriptor().getComponent()
                .getProcessorKeepAliveTime().getValue();

        this.retryMaxTry = this.component.getJbiComponentDescriptor().getComponent().getAcceptorRetryNumber()
                .getValue();
        this.retryBaseDuration = this.component.getJbiComponentDescriptor().getComponent().getAcceptorRetryWait()
                .getValue();

        this.processorStopMaxWait = this.component.getJbiComponentDescriptor().getComponent().getProcessorStopMaxWait()
                .getValue();

        this.logger.fine("\t- core pool size: " + coreProcessorPoolSize);
        this.logger.fine("\t- max core pool size: " + maxProcessorPoolSize);
        this.logger.fine("\t- keep alive time: " + processorPoolKeepAliveTime);
        this.logger.fine("\t- retry policy : max try number: " + this.retryMaxTry);
        this.logger.fine("\t- retry policy : base duration: " + this.retryBaseDuration);
        this.logger.fine("\t- stop max wait: " + this.processorStopMaxWait);

        this.group = new ThreadGroup(component.getContext().getComponentName() + GROUP_SUFFIX);
        this.processorThreadpool = new JBIProcessorThreadPoolExecutor(coreProcessorPoolSize, maxProcessorPoolSize,
                processorPoolKeepAliveTime, TimeUnit.SECONDS, this.jbiProccesorManagerObservers, this.group,
                this.component, this.logger, this.retryBaseDuration);

        this.processorThreadpool.prestartAllCoreThreads();

        this.processorThreadPoolProbe.setThreadPool(this.processorThreadpool);
    }

    /**
     * Stop the Processor Thread Pool.
     * 
     */
    public void stop() {

        // so that it destroy itself when there is no thread left
        this.group.setDaemon(true);

        // try to let the executing tasks finish
        this.processorThreadpool.shutdown();

        try {
            this.processorThreadpool.awaitTermination(this.processorStopMaxWait, TimeUnit.MILLISECONDS);
        } catch (final InterruptedException e) {
            this.logger.log(Level.SEVERE, "Workers pool termination interrupted", e);
        }

        // interrupt the remaining executing tasks
        this.processorThreadpool.shutdownNow();

        // let's just run a thread that will live very shortly in order to log potential problems with the component
        // implementation
        new CleaningMonitorThread(this.logger, this.group).start();
}

    private static final class CleaningMonitorThread extends Thread {
        private final ThreadGroup group;

        private final Logger logger;

        public CleaningMonitorThread(final Logger logger, final ThreadGroup group) {
            super("JBI Processor Cleaning Monitor");
            this.logger = logger;
            this.group = group;
        }

        @Override
        public void run() {
            try {
                // let's wait a bit for the threads to actually end after the interrupt
                // but we are never sure it will happen, because there is component code executed there...
                Thread.sleep(2000);
            } catch (final InterruptedException e) {
                logger.log(Level.WARNING, "Sleep interrupted...", e);
            }

            if (group.activeCount() > 0) {
                logger.warning("All thread in the group '" + group.getName()
                        + "' should have ended 2000ms after interrupt but have not yet... "
                        + "Maybe there is a bug in the component implementation!");
            }
        }
    }

    long getProcessorStopMaxWait() {
        return this.processorStopMaxWait;
    }

    int getRetryMaxTry() {
        return this.retryMaxTry;
    }

    long getRetryBaseDuration() {
        return this.retryBaseDuration;
    }

    /**
     * Get the message exchange processor thread pool.
     * 
     * Note: only there for tests
     */
    ThreadPoolExecutor getProcessorThreadPool() {
        return this.processorThreadpool;
    }

    /**
     * Reset the core Pool size.
     * 
     * @param size
     */
    void setCorePoolSize(final int size) {
        this.logger.info("Reset the processor thread pool size to " + size);
        try {
            this.processorThreadpool.setCorePoolSize(size);
            this.processorThreadpool.prestartAllCoreThreads();
        } catch (final IllegalArgumentException e) {
            final String warningMessage = String.format(
                    "Invalid value used (less than zero) to configure the processor thread pool core size: %d", size);
            this.logger.warning(warningMessage);
            this.logger.log(Level.FINE, warningMessage, e);
        }
    }

    /**
     * Reset the max core Pool size.
     * 
     * @param size
     */
    void setMaxCorePoolSize(final int size) {
        this.logger.info("Reset the maximum processor thread pool size to " + size);
        try {
            this.processorThreadpool.setMaximumPoolSize(size);
        } catch (final IllegalArgumentException e) {
            final String warningMessage = String.format(
                    "Invalid value used (less than or equal to zero, or less than the core pool size) to configure the processor thread pool max size: %d",
                    size);
            this.logger.warning(warningMessage);
            this.logger.log(Level.FINE, warningMessage, e);
        }
    }

    /**
     * Reset the keep alive time of the processor pool
     * 
     * @param size
     */
    void setPoolKeepAliveTime(final int time) {
        this.logger.info("Reset the processor thread pool keep alive time to " + time);
        try {
            this.processorThreadpool.setKeepAliveTime(time, TimeUnit.SECONDS);
        } catch (final IllegalArgumentException e) {
            final String warningMessage = String.format(
                    "Invalid value used (less than zero) to configure the processor thread pool keep alive time: %d",
                    time);
            this.logger.warning(warningMessage);
            this.logger.log(Level.FINE, warningMessage, e);
        }
    }

    /**
     * Reset the stop max wait time of the processor pool
     */
    void setStopMaxWait(final long time) {
        this.logger.info("Reset the processor stop max wait time to " + time);
        if (time <= 0) {
            final String warningMessage = String.format(
                    "Invalid value used (less than or equals to zero) to configure the processor stop max wait time: %d",
                    time);
            this.logger.warning(warningMessage);
        } else {
            this.processorStopMaxWait = time;
        }
    }

    public void process(final ExchangeImpl exchange, final AsyncContextExchange expiredContext) {
        // we retry at most retryMaxTry times
        for (int nbTry = 1; nbTry <= retryMaxTry; nbTry++) {
            // each time, we do it several times in order to increase the time
            for (int nb = 1; nb <= nbTry; nb++) {
                try {
                    // this will wait at most retryBaseDuration (see JBIProcessorThreadPoolExecutor)
                    this.processorThreadpool.execute(new ExchangeRunnable(exchange, expiredContext));
                    threadPoolPick();
                    return;
                } catch (final Exception e) {
                    final String message;
                    final Exception toLog;
                    final boolean stop;
                    if (e instanceof RejectedExecutionException) {
                        message = "No JBI Processor thread available to handle exchange";
                        // no need to log the exception
                        toLog = null;
                        stop = false;
                    } else if (e instanceof RuntimeInterruptedException) {
                        message = "Thread was interrupted, aborting handling of exchange";
                        stop = true;
                        toLog = null;
                    } else {
                        message = "An unknown error occured while submitting an exchange to the message exchange processor pool";
                        toLog = e;
                        stop = false;
                    }
                    if (stop || (nbTry == retryMaxTry && nb == nbTry)) {
                        handleNoMoreTry(exchange, nbTry, message);
                    } else if (nb == nbTry) {
                        // there we need to log the exception
                        this.logger.log(Level.WARNING, "Try #" + nbTry + " of " + retryMaxTry + " for "
                                + exchange.getExchangeId() + ". " + message + ".", toLog);
                    }
                }
            }
        }
    }
    
    private void handleNoMoreTry(final ExchangeImpl exchange, int nbTry, final String message) {
        // if it's not active, we can't send back an error
        final boolean isActive = exchange.isActiveStatus();
        this.logger.warning("Try #" + nbTry + " of " + retryMaxTry + " for " + exchange.getExchangeId() + ". " + message
                + "." + (isActive ? "Exchange is active, sending back an error."
                        : "Exchange is not active, discarding it..."));
        if (isActive) {
            exchange.setError(new MessagingException(message));
            try {
                this.component.getChannel().send(exchange.getMessageExchange());
            } catch (final MessagingException e) {
                this.logger.log(Level.WARNING, "Can't send back an error.", e);
            }
        }
    }

    private void threadPoolPick() {
        try {
            this.processorThreadPoolProbe.pick();
        } catch (final ProbeNotStartedException e) {
            this.logger.warning(
                    "The message exchange processor thread pool probe is not started. Metric values will not be valid.");
        }
    }

    // /////////////////////////////////////////////////////
    // Implementation of methods of NotificationListener //
    // /////////////////////////////////////////////////////

    @Override
    public void handleNotification(final Notification notification, final Object obj) {
        if (notification instanceof AttributeChangeNotification) {
            final AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
            final String attrName = attrNotif.getAttributeName();
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.log(Level.FINE, "Notification received by " + this.getClass().getName()
                        + " for attribute : " + attrName);
            }
            switch (attrName) {
                case "processorPoolSize": {
                    final Integer newVal = (Integer) attrNotif.getNewValue();
                    this.setCorePoolSize(newVal);
                    break;
                }
                case "processorMaxPoolSize": {
                    final Integer newVal = (Integer) attrNotif.getNewValue();
                    this.setMaxCorePoolSize(newVal);
                    break;
                }
                case "processorKeepAliveTime": {
                    final Integer newVal = (Integer) attrNotif.getNewValue();
                    this.setPoolKeepAliveTime(newVal);
                    break;
                }
                case "processorStopMaxWait": {
                    final Long newVal = (Long) attrNotif.getNewValue();
                    this.setStopMaxWait(newVal);
                    break;
                }
                default: {
                    this.logger.log(Level.FINE,
                            "Notification ignored by " + this.getClass().getName() + " for attribute : " + attrName);
                }
            }
        }
    }

    public final void addObserver(JBIProcessorManagerObserver jbiProccesorManagerObserver) {
        synchronized (this.jbiProccesorManagerObservers) {
            this.jbiProccesorManagerObservers.add(jbiProccesorManagerObserver);
        }
    }

    public final void deleteObserver(JBIProcessorManagerObserver jbiProccesorManagerObserver) {
        synchronized (this.jbiProccesorManagerObservers) {
            this.jbiProccesorManagerObservers.remove(jbiProccesorManagerObserver);
        }
    }
}
