/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.lang.Thread.UncaughtExceptionHandler;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;

/**
 * JBI message processor thread factory
 * 
 * In particular, this class is responsible of attributing a {@link MessageExchangeProcessor} (that will be used in
 * {@link JBIProcessorThreadPoolExecutor}) to a thread of the pool.
 * 
 * @author Christophe DENEUX - EBM Websourcing
 * @author vnoel
 */
class JBIProcessorThreadFactory implements ThreadFactory {

    private final AtomicInteger threadNumber = new AtomicInteger(1);

    private final AbstractComponent component;

    private final Logger logger;

    private final ThreadGroup group;

    public JBIProcessorThreadFactory(final ThreadGroup group, final AbstractComponent component, final Logger logger) {
        this.group = group;
        this.component = component;
        this.logger = logger;
    }

    @Override
    public Thread newThread(final Runnable r) {
        final String name = this.component.getContext().getComponentName() + " - JBI Processor Thread #"
                + this.threadNumber.getAndIncrement();

        this.logger.fine("Creating a new thread: " + name);

        final Thread thread;
        try {
            thread = new JBIProcessorThread(this.getGroup(), r, name, this.component, this.logger);
        } catch (final PEtALSCDKException e) {
            this.logger.log(Level.SEVERE, "Can't create thread '" + name + "'", e);
            return null;
        }

        thread.setUncaughtExceptionHandler(new UncaughtExceptionHandler() {
            @Override
            public void uncaughtException(final Thread t, final Throwable e) {
                JBIProcessorThreadFactory.this.logger.log(Level.SEVERE, t.getName() + " threw an uncaught exception.",
                        e);
            }
        });
        return thread;
    }

    public ThreadGroup getGroup() {
        return group;
    }

    public static class JBIProcessorThread extends Thread {

        private final MessageExchangeProcessor messageExchangeProcessor;

        public JBIProcessorThread(final ThreadGroup group, final Runnable worker, final String name,
                final AbstractComponent component, final Logger logger) throws PEtALSCDKException {
            super(group, worker, name);
            messageExchangeProcessor = new MessageExchangeProcessor(component, logger);
        }

        public MessageExchangeProcessor getMessageExchangeProcessor() {
            return messageExchangeProcessor;
        }

        @Override
        public void run() {
            super.run();

            this.messageExchangeProcessor.destroy();
        }
    }
}