/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.process.JBIProcessorThreadFactory.JBIProcessorThread;
import org.ow2.petals.component.framework.process.async.AsyncContextCleaner;
import org.ow2.petals.component.framework.process.async.AsyncContextExchange;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.probes.DurationProbe;
import org.ow2.petals.probes.api.probes.StartDateItem;

/**
 * JBI Message processor pool
 * 
 * @author Christophe DENEUX (EBMWebsourcing)
 * @author vnoel
 * 
 */
class JBIProcessorThreadPoolExecutor extends ThreadPoolExecutor {

    private final Logger logger;

    private final List<JBIProcessorManagerObserver> jbiProccesorManagerObservers;

    /**
     * The probe about the processing duration
     */
    private final DurationProbe mexProcessingDurationProbe;

    public JBIProcessorThreadPoolExecutor(final int corePoolSize, final int maximumPoolSize, final long keepAliveTime,
            final TimeUnit unit, final List<JBIProcessorManagerObserver> jbiProccesorManagerObservers,
            final ThreadGroup threadGroup, final AbstractComponent component, final Logger logger, final long timeout) {
        super(corePoolSize, maximumPoolSize, keepAliveTime, unit, new IncreasingTimeoutSynchronousQueue(timeout),
                new JBIProcessorThreadFactory(threadGroup, component, logger), new ThreadPoolExecutor.AbortPolicy());
        this.jbiProccesorManagerObservers = jbiProccesorManagerObservers;
        this.logger = logger;
        this.mexProcessingDurationProbe = component.getMonitoringBean().getMessageExchangeProcessorDurationProbe();
    }

    public static class ExchangeRunnable implements Runnable {

        private final ExchangeImpl exchange;

        private final AsyncContextExchange expiredContext;

        /**
         * Used in {@link JBIProcessorThreadPoolExecutor#afterExecute(Runnable, Throwable)}
         */
        private MessageExchangeProcessor messageExchangeProcessor = null;

        /**
         * The current duration item used by {@link JBIProcessorThreadPoolExecutor#mexProcessingDurationProbe}.
         */
        private StartDateItem mexProcessingDurationItem = null;

        public ExchangeRunnable(final ExchangeImpl exchange, final AsyncContextExchange expiredContext) {
            this.exchange = exchange;
            this.expiredContext = expiredContext;
        }

        @Override
        public void run() {
            // this should never be null
            this.messageExchangeProcessor.run();
        }
    }

    @Override
    protected void beforeExecute(Thread t, Runnable r) {

        super.beforeExecute(t, r);

        if (t instanceof JBIProcessorThread && r instanceof ExchangeRunnable) {
            final ExchangeRunnable er = (ExchangeRunnable) r;
            final JBIProcessorThread jpt = (JBIProcessorThread) t;

            final MessageExchangeProcessor mexProcessor = jpt.getMessageExchangeProcessor();

            er.messageExchangeProcessor = mexProcessor;

            mexProcessor.initCurrentProcessing(er.exchange, er.expiredContext);

            try {
                er.mexProcessingDurationItem = this.mexProcessingDurationProbe.newExecution();
            } catch (final ProbeException e) {
                this.logger
                        .warning("An error occurs in the the processing duration probe of message exchange processors: "
                                + e.getMessage() + ". Metric values will not be valid.");
            }

            this.notifyObserversBeforeProcessing(er.exchange);
        } else {
            this.logger.warning(
                    "Message exchange processor thread pool is executing a thread of type " + t.getClass().getName()
                            + " and a runnable of type " + r.getClass().getName() + ". Not what I expected.");
        }
    }

    @Override
    protected void afterExecute(final Runnable r, final Throwable t) {

        if (r instanceof ExchangeRunnable) {
            final ExchangeRunnable er = (ExchangeRunnable) r;
            final MessageExchangeProcessor mexProcessor = er.messageExchangeProcessor;

            try {
                this.notifyObserversAfterProcessing(er.exchange);

                try {
                    this.mexProcessingDurationProbe.endsExecution(er.mexProcessingDurationItem,
                            ExecutionStatus.SUCCEEDED);
                } catch (final ProbeException e) {
                    this.logger.warning(
                            "An error occurs in the the processing duration probe of message exchange processors: "
                                    + e.getMessage() + ". Metric values will not be valid.");
                }

            } finally {
                // Clean the Message Exchange Processor
                mexProcessor.clean();
            }
        }

        super.afterExecute(r, t);
    }

    private final void notifyObserversBeforeProcessing(ExchangeImpl exchange) {
        synchronized (this.jbiProccesorManagerObservers) {
            for (JBIProcessorManagerObserver jbiProccesorManagerObserver : jbiProccesorManagerObservers) {
                jbiProccesorManagerObserver.beforeProcessingMessage(exchange);
            }
        }
    }

    private final void notifyObserversAfterProcessing(ExchangeImpl exchange) {
        synchronized (this.jbiProccesorManagerObservers) {
            for (JBIProcessorManagerObserver jbiProccesorManagerObserver : jbiProccesorManagerObservers) {
                jbiProccesorManagerObserver.afterProcessingMessage(exchange);
            }
        }
    }
    
    public static class RuntimeInterruptedException extends RuntimeException {

        public RuntimeInterruptedException(final InterruptedException e) {
            super(e);
        }

        private static final long serialVersionUID = -6570130537977648017L;
        
    }

    /**
     * This implementation is needed to have a timeout before {@link IncreasingTimeoutSynchronousQueue#offer(Runnable)}
     * fails: it is called by {@link ThreadPoolExecutor#execute(Runnable)} and it is the only way we have to control the
     * timeout.
     * 
     * In particular, since this queue is used globally by the {@link ThreadPoolExecutor}, the timeout will impact all
     * {@link MessageExchangeAcceptor}s as well as the {@link AsyncContextCleaner}, so it must be a fixed value.
     *
     * We would not need all of that if we could simply call {@link ThreadPoolExecutor#execute(Runnable)} with a
     * timeout, but it is not present in the API.
     * 
     * TODO see if that can be replaced by a {@link RejectedExecutionHandler} that simply do the
     * {@link BlockingQueue#offer(Object, long, TimeUnit)}... in that case we could exploits the
     * {@link ExchangeRunnable} to store information about retry and keep everything located in there instead of the
     * {@link JBIProcessorManager}...
     */
    private static class IncreasingTimeoutSynchronousQueue extends SynchronousQueue<Runnable> {

        private static final long serialVersionUID = 3461572255001836062L;

        private final long timeout;

        public IncreasingTimeoutSynchronousQueue(final long timeout) {
            this.timeout = timeout;
        }

        @Override
        public boolean offer(final Runnable e) {
            try {
                return super.offer(e, timeout, TimeUnit.MILLISECONDS);
            } catch (InterruptedException e1) {
                throw new RuntimeInterruptedException(e1);
            }
        }

    }
}