/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;

import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeException;
import org.ow2.petals.probes.api.probes.CounterGaugeProbe;
import org.ow2.petals.probes.api.probes.DurationProbe;
import org.ow2.petals.probes.api.probes.StartDateItem;


/**
 * The JBI Message Exchange Acceptor. 
 * <p>
 * Dispatch the message to a JBI worker which
 * will use the {@link AbstractJBIListener} component implementation.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * @since 3.0
 */
public class MessageExchangeAcceptor extends Thread {

    /**
     * The component's logger
     */
    private final Logger logger;

    private final DeliveryChannel deliveryChannel;

    private volatile boolean running = false;

    private final Object runningLock = new Object();

    private final JBIProcessorManager jbiProcessorManager;

    /**
     * The probe about the acceptance duration
     */
    private final DurationProbe mexAcceptanceDurationProbe;

    /**
     * The probe counting working message exchange acceptors
     */
    private final CounterGaugeProbe mexWorkingAcceptorCounter;

    public MessageExchangeAcceptor(final AbstractComponent component, final JBIProcessorManager jbiProcessorManager,
            final Logger logger, final ThreadGroup threadGroup, final String threadName) {
        super(threadGroup, threadName);

        this.jbiProcessorManager = jbiProcessorManager;
        this.deliveryChannel = component.getChannel();
        this.logger = logger;

        final Monitoring monitoringMbean = component.getMonitoringBean();
        this.mexAcceptanceDurationProbe = monitoringMbean.getMessageExchangeAcceptorDurationProbe();
        this.mexWorkingAcceptorCounter = monitoringMbean.getWorkingMessageExchangeAcceptorProbe();
    }

    @Override
    public void run() {
        this.logger.log(Level.FINE, "Starting MessageExchangeAcceptor Thread " + this.getName());
        this.running = true;
        StartDateItem currentStartDateItem = null;
        while (this.running) {
            // let's clean context in case somewhere we initialised it
            PetalsExecutionContext.clear();
            try {
                // wait a message
                final MessageExchange mexReceived = this.deliveryChannel.accept();

                this.mexWorkingAcceptorCounter.inc();
                currentStartDateItem = this.mexAcceptanceDurationProbe.newExecution();

                // accept the message exchange
                final ExchangeImpl exchange = new ExchangeImpl(mexReceived);

                this.logger.fine("Accepting a JBI message with Id : " + exchange.getExchangeId());
                
                this.jbiProcessorManager.process(exchange, null);

                this.mexWorkingAcceptorCounter.dec();
                this.mexAcceptanceDurationProbe.endsExecution(currentStartDateItem, ExecutionStatus.SUCCEEDED);
                currentStartDateItem = null;

            } catch (final Exception e) {
                synchronized (this.runningLock) {
                    // normally, only JBIAcceptorManager will interrupt us, and it is for stopping us.
                    // it is always wrapped in a MessagingException because accept() only throws those
                    if (e instanceof MessagingException && e.getCause() instanceof InterruptedException) {
                        this.logger.fine("Thread '" + this.getName() + "' interrupted, stop accepting JBI messages");
                        break;
                    }
                    if (this.running) {
                        this.logger.log(Level.WARNING,
                                "An unexpected exception occurs during the processing of the JBI acceptor '"
                                        + this.getName() + "'", e);

                        if (currentStartDateItem != null) {
                            try {
                                this.mexAcceptanceDurationProbe.endsExecution(currentStartDateItem,
                                        ExecutionStatus.ERROR);
                            } catch (final ProbeException e1) {
                                this.logger
                                        .warning("An error occurs in the the processing duration probe of message exchange acceptor: "
                                                + e1.getMessage() + ". Metric values will not be valid.");
                            }
                        }
                    }
                }
            }
        }
        this.logger.fine("Thread '" + this.getName() + "' ended.");
    }

    public void ends() {
        synchronized (this.runningLock) {
            if (this.running) {
                this.running = false;
                this.logger.fine("Thread '" + this.getName() + "' will stop to accept JBI messages.");
            }
        }
    }
}
