/**
 * Copyright (c) 2006-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.FlowAttributesExchangeHelper;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.ExchangeProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageParam;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.Param;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.logger.StepLogHelper;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.monitoring.MonitoringHelper;
import org.ow2.petals.component.framework.process.async.AsyncContextExchange;
import org.ow2.petals.component.framework.process.async.AsyncMessageManager;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;

import com.ebmwebsourcing.easycommons.lang.StringHelper;

/**
 * Processor for received MessageExchanges.
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 */
public class MessageExchangeProcessor {

    public static final String PROVIDER_FLOWATTRIBUTES_PREFIX = "org.ow2.petals.component.framework.process.MessageExchangeProcessor.provider";

    /**
     * The component logger;
     */
    private final AbstractComponent component;

    /**
     * The component logger
     */
    private final Logger logger;

    /**
     * The component implementation of the message processor
     */
    private final AbstractJBIListener jbiListener;

    /**
     * The component service unit manager
     */
    private final AbstractServiceUnitManager suManager;

    /**
     * The component asynchronous message manager
     */
    private final AsyncMessageManager asyncManager;

    /**
     * The component name
     */
    private final String componentName;

    /**
     * The exchange, during one message processing
     */
    private ExchangeImpl exchange = null;

    /**
     * not <code>null</code> if the current exchange is an expired exchange.
     * 
     * This is needed because we don't have ownership of expired messages, so we can't check their properties!
     */
    private AsyncContextExchange expiredContext = null;

    /**
     * Creates a processor that listen to new {@link MessageExchange} on the specified channel. Exchanges are ignored
     * depending on the specified exchange-status to ignore.
     * 
     * @param component
     * @param jbiListener
     * @param logger
     * @throws PEtALSCDKException
     */
    public MessageExchangeProcessor(final AbstractComponent component, final Logger logger) throws PEtALSCDKException {
        this.component = component;
        this.jbiListener = createAndInitJBIListener();

        this.logger = logger;

        // get a direct reference to the required component attributes
        this.suManager = component.getServiceUnitManager();
        this.asyncManager = component.getAsyncManager();
        this.componentName = component.getContext().getComponentName();
    }

    /**
     * Instantiate and initialize a JBI listener. This JBI Listener is the one which is defined by the component
     * developer.
     * 
     * @return
     * @throws PEtALSCDKException
     */
    private AbstractJBIListener createAndInitJBIListener() throws PEtALSCDKException {

        final String jbiListenerClassNameValue = this.component.getJbiComponentDescriptor().getComponent()
                .getJbiListenerClassName();
        final String jbiListenerClassName = jbiListenerClassNameValue.replaceAll("\\s", "");

        final AbstractJBIListener jbiListener;
        if (!StringHelper.isNullOrEmpty(jbiListenerClassName)) {
            try {
                final ClassLoader componentClassLoader = this.getClass().getClassLoader();
                final Class<?> clazz = componentClassLoader.loadClass(jbiListenerClassName);
                jbiListener = (AbstractJBIListener) clazz.newInstance();
            } catch (final InstantiationException e) {
                throw new PEtALSCDKException("Can't instanciate JBIListener : " + jbiListenerClassName, e);
            } catch (final IllegalAccessException e) {
                throw new PEtALSCDKException("Can't instanciate JBIListener : " + jbiListenerClassName, e);
            } catch (final ClassNotFoundException e) {
                throw new PEtALSCDKException("Can't find JBIListener class : " + jbiListenerClassName, e);
            }
            jbiListener.init(this.component);
            // component implementor specific init
            jbiListener.init();
        } else {
            throw new PEtALSCDKException("JBIListener class name can not be empty");
        }
        return jbiListener;
    }

    /**
     * Must be called before submitting this processor to the executor.
     * 
     * @param exchange
     *            can't be <code>null</code>
     * @param expiredContext
     *            if not <code>null</code>, then it means a send async of this exchange expired.
     */
    public void initCurrentProcessing(final ExchangeImpl exchange, final AsyncContextExchange expiredContext) {
        this.exchange = exchange;
        this.expiredContext = expiredContext;
    }

    public void run() {
        final boolean wasActive = exchange.isActiveStatus();
        try {
            if (expiredContext != null) {
                processExpired();
            } else {
                process();
            }
        } catch (final Exception e) {
            // Return an error acknowledgment only if the exchange was active before processing
            if (wasActive) {
                this.logger.log(Level.WARNING,
                        "An error occured during message processing, let's send it back since the exchange was active before.",
                        e);
                try {
                    this.exchange.setError(e);
                    sendNoException();
                } catch (final IllegalStateException me) {
                    this.logger.log(Level.SEVERE,
                            "It seems I'm not the owner of the exchange anymore... Maybe there is a bug somewhere! There is nothing I can do now.",
                            me);
                }
            } else {
                this.logger.log(Level.WARNING,
                        "An error occured during message processing, but the exchange wasn't active before, there is nothing to do now.",
                        e);
            }
        } finally {
            PetalsExecutionContext.clear();
        }
    }

    /**
     * Clean the Message Exchange Processor. It cleans all the resources used
     * for one Message Processing.
     */
    public void clean() {
        // reset the JBI Listener instance
        this.jbiListener.setConsumes(null);
        this.jbiListener.setProvides(null);
        this.exchange = null;
        this.expiredContext = null;
    }

    public void destroy() {
        this.jbiListener.shutdown();
    }

    /**
     * Important: these methods call send() and then, NO code should try to access the exchange's content (except id and
     * status) because we loose ownership after that.
     */
    private void process() throws MessagingException {

        // 1) first let's setup the flow attributes before doing anything else
        PetalsExecutionContext.clear();

        // here either the exchange is not associated to a context (null),
        // or there is none (length == 0), or there is one (length == 1)
        final AsyncContextExchange[] asyncContext = this.asyncManager.unreferenceAsyncContext(exchange);

        // 1a) let's see what is present in the exchanges and/or our async context
        final FlowAttributes retrieved = getFlowFromExchangeOrContext(exchange, asyncContext, logger);

        // 1b) then let's compute a flow attribute in case it was missing
        // and let's store it in the execution context
        final FlowAttributes actual = setOrInitFlowForThread(exchange, asyncContext, retrieved, logger);

        // 1c) let's compute the flow attribute to use during the rest of the processing
        final FlowAttributes current;
        if (exchange.isProviderRole() && exchange.isNewExchange()) {
            // only increment step id when a new processing as provider is requested
            // we should *NOT* increment stepid on acknowledgements or errors
            current = PetalsExecutionContext.nextFlowStepId();

            StepLogHelper.addMonitStartTrace(logger, exchange.getMessageExchange(), current, actual);

            MonitoringHelper.updateStatsOnStartOfServiceProviderInvocation(exchange, this.component.getMonitoringBean(),
                    this.logger);
        } else {
            current = actual;
        }

        // 1d) We take the flow attributes and we store them
        // we will need them when send back an answer to log the end of the processing
        // Note: theoretically, we could always use the context one, but it is safer to rely on the exchange
        // itself since the component implementation is not meant to have access to this method!
        exchange.setCurrentProcessFlowAttributes(current);

        // 2) and now we can send errors or do processing
        if (asyncContext != null && asyncContext.length == 0) {
            if (this.logger.isLoggable(Level.INFO)) {
                this.logger.info("The asynchronous response does not match any AsyncContext for "
                        + exchange.getExchangeId() + ". It must have expired before: "
                        + (exchange.isActiveStatus() ? "let's notify the sender with an error."
                                : "we ignore the exchange since it was not active."));
            }
            if (exchange.isActiveStatus()) {
                exchange.setError(new MessagingException("Exchange was ignored: it had already timed out."));
                sendNoException();
            }
        } else {
            final AsyncContextExchange acontext = asyncContext != null ? asyncContext[0] : null;
            if (exchange.isProviderRole()) {
                this.processAsProvider(acontext);
            } else {
                this.processAsConsumer(acontext);
            }
        }
    }

    private static FlowAttributes setOrInitFlowForThread(final ExchangeImpl exchange,
            final AsyncContextExchange[] asyncContext, final FlowAttributes retrieved, final Logger logger) {
        final FlowAttributes actual;
        if (retrieved != null) {
            PetalsExecutionContext.putFlowAttributes(retrieved);
            actual = retrieved;
        } else {
            // if asyncContext is not null, then current must be non-null since we set it before sending!
            // (or it had expired and then it's ok)
            assert asyncContext == null || asyncContext.length == 0;
            // in case we received incorrect information, let's initialise the flows instead
            actual = PetalsExecutionContext.initFlowAttributes();
            if (exchange.isProviderRole()) {
                if (exchange.isNewExchange()) {
                    logger.warning(
                            "Missing consumer flow attributes in Exchange (the consumer should have put them in before sending, "
                                    + "it is not normal...), initialising them with: " + actual);
                } else {
                    logger.warning(
                            "Missing provider flow attributes in Exchange (I should have put them in before sending, "
                                    + "the consumer must have removed them...), initialising them with: " + actual);
                }
            } else {
                logger.warning(
                        "Missing consumer flow attributes in Exchange (I should have put them in before sending, "
                                + "the provider must have removed them...), initialising them with: " + actual);
            }
        }

        if (logger.isLoggable(Level.FINE)) {
            logger.fine(
                    String.format("Flow attributes for exchange (id %s) set to %s", exchange.getExchangeId(), actual));
        }

        assert actual != null;

        return actual;
    }

    private static FlowAttributes getFlowFromExchangeOrContext(final ExchangeImpl exchange,
            final AsyncContextExchange[] asyncContext, final Logger logger) {
        final FlowAttributes retrieved;
        final Boolean fromConsumer;
        if (exchange.isProviderRole()) {
            if (exchange.isNewExchange()) {
                // this should never be non-null because it is a new exchange, there can't be an async context!
                assert asyncContext == null;

                retrieved = FlowAttributesExchangeHelper.getFlowAttributes(exchange.getMessageExchange());
                fromConsumer = true;
            } else {
                if (asyncContext != null && asyncContext.length == 1) {
                    retrieved = asyncContext[0].getExchange().getCurrentProcessFlowAttributes();
                    fromConsumer = null;
                } else {
                    retrieved = FlowAttributesExchangeHelper.getFlowAttributes(exchange.getMessageExchange(),
                            PROVIDER_FLOWATTRIBUTES_PREFIX);
                    fromConsumer = false;
                }
            }
        } else {
            if (asyncContext != null && asyncContext.length == 1) {
                retrieved = asyncContext[0].getExchange().getCurrentProcessFlowAttributes();
                fromConsumer = null;
            } else {
                retrieved = FlowAttributesExchangeHelper.getFlowAttributes(exchange.getMessageExchange());
                fromConsumer = true;
            }
        }

        if (logger.isLoggable(Level.FINE)) {
            logger.fine(String.format("Retrieved%s flow attributes (%s) from %s (%s)",
                    fromConsumer == null ? "" : (fromConsumer ? " consumer" : " provider"), retrieved,
                    fromConsumer == null ? "async context" : "exchange",
                    exchange.getExchangeId()));
        }

        return retrieved;
    }

    /**
     * We need this specific processing that takes care of NOT accessing things that needs ownership of the exchange
     * since we don't have it anymore after a sendAsync expired.
     */
    private void processExpired() throws MessagingException {

        final FlowAttributes current = expiredContext.getExchange().getCurrentProcessFlowAttributes();
        // it can't be null because we set it before sending!
        assert current != null;

        PetalsExecutionContext.putFlowAttributes(current);

        final Object service = expiredContext.getService();

        final String type;
        if (expiredContext.isAsProvider()) {
            type = "PROVIDES";
            this.jbiListener.setProvides((Provides) service);
        } else {
            type = "CONSUMES";
            this.jbiListener.setConsumes((Consumes) service);
        }

        this.logger.fine(
                "Process an expired (sendAsync) exchange as " + type + " with id : " + exchange.getExchangeId());

        // note that we can't reply in case of expired exchange
        this.jbiListener.onExpiredAsyncJBIMessage(exchange, expiredContext.getAsyncContext());

    }

    /**
     * See {@link #process(Exchange)} documentation!
     */
    private void processAsConsumer(final AsyncContextExchange asyncContext) throws MessagingException {

        this.logger.fine("Process an exchange as CONSUMER with id : " + exchange.getExchangeId());

        final Consumes consumes = this.suManager != null ? this.suManager.getConsumesFromExchange(exchange) : null;
        if (consumes != null) {
            this.jbiListener.setConsumes(consumes);
            try {
                this.jbiListener.getMessageInterceptorProcessor().interceptSendResponse(exchange, consumes);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        } else {
            this.logger
                    .fine("The exchange '"
                            + exchange.getExchangeId()
                            + "' refers to an unknown consumer on the component : "
                            + this.componentName
                            + ". The listener will not have the consumes and the extensions for this exchange");
        }

        // store it before calling the component's code
        final boolean wasActive = exchange.isActiveStatus();

        final boolean shouldReply = invokeJBIListener(asyncContext);

        // if it was active, it means we need to answer in any case
        if (shouldReply && wasActive) {
            // let's set it to Done if it is needed
            if (exchange.isActiveStatus() && (exchange.isRobustInOnlyPattern() || exchange.isInOutPattern()
                    || (exchange.isInOptionalOutPattern() && exchange.getFault() == null))) {
                exchange.setDoneStatus();
            }
            sendNoException();
        }
    }

    private void sendNoException() {
        try {
            this.jbiListener.send(exchange);
        } catch (final MessagingException e) {
            // in case there is an exception here, we don't want it to be handled by the top-most handler in run()!
            // but if it is an IllegalStateException, we want, because it's not normal that it happens!
            this.logger.log(Level.WARNING,
                    "The CDK couldn't send an answer to the exchange on behalf of the component. There is nothing we can do now.",
                    e);
        }
    }

    /**
     * See {@link #process(Exchange)} documentation!
     */
    private void processAsProvider(final AsyncContextExchange asyncContext) throws MessagingException {

        logger.fine("Process an exchange as PROVIDER with id : " + exchange.getExchangeId());

        final Provides provides = this.suManager != null ? this.suManager.getProvidesFromExchange(exchange) : null;
        
        if (provides == null) {
            // if it's not new, we don't want to check for service description (because we consider that the service
            // description should be present only if the endpoint is active and it is possible to get answer after
            // deactivation)
            if (exchange.isNewExchange() && this.component.getServiceDescription(exchange.getEndpoint()) == null) {
                throw new MessagingException("The endpoint '" + exchange.getEndpoint()
                        + "' is unknown on the component : " + this.componentName);
            }

            // TODO normally we should check if the operation is defined in the native service as we do for the SU
            // services...

            this.logger.fine("Process an exchange managed directly by the component");
        } else {
            if (!isOperationAvailableForExchangeEndpoint()) {
                throw new MessagingException("Operation '" + exchange.getOperation().toString()
                        + "' not found in endpoint '" + exchange.getEndpointName() + "' description");
            }
            
            this.jbiListener.setProvides(provides);
            try {
                this.jbiListener.getMessageInterceptorProcessor().interceptAccept(exchange, provides);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        }

        final boolean hadOut = exchange.isOutMessage();
        final boolean wasActive = exchange.isActiveStatus();

        final boolean shouldReply = invokeJBIListener(asyncContext);

        if (shouldReply && wasActive) {

            if (exchange.isActiveStatus()) {

                final boolean hasOut = exchange.isOutMessage();

                // only if the out was just added!
                if (hasOut && !hadOut) {
                    // TODO move that to jbiListener.send ???
                    forwardMessageData(exchange, provides);
                }

                final boolean hasFault = exchange.getFault() != null;

                // this is true when there is both out and fault, or none of them
                if (hasOut ^ !hasFault) {
                    // these are the cases when it makes sense to set the exchange to done
                    exchange.setDoneStatus();
                }
            }

            // TODO move that to jbiListener.send ???
            setProperties(exchange, provides);

            sendNoException();
        }
    }

    private boolean invokeJBIListener(final AsyncContextExchange asyncContext) {
        if (asyncContext == null) {
            return this.jbiListener.onJBIMessage(exchange);
        } else {
            return this.jbiListener.onAsyncJBIMessage(exchange, asyncContext.getAsyncContext());
        }
    }

    private boolean isOperationAvailableForExchangeEndpoint() {
        // this is called only if we are manipulating a Provides, so there should be a SU Manager
        assert suManager != null;
        final ServiceEndpoint endpoint = exchange.getEndpoint();
        final ServiceUnitDataHandler suDataHandler = suManager.getSUDataHandler(endpoint);
        if (suDataHandler != null) {
            final List<QName> epOperations = suDataHandler.getEndpointOperationsList(endpoint);
            // if operation list is empty this means that endpoint is described
            // by a fake generated wsdl, so skip operation check phase
            if (epOperations != null && !epOperations.isEmpty()) {
                return epOperations.contains(exchange.getOperation());
            } else {
                return true;
            }
        }
        return false;
    }

    /**
     * Set the configured Exchange and Message Properties.
     * 
     * @param exchange2
     * 
     * @param exchangeProperties
     * @throws MessagingException
     */
    private static void setProperties(Exchange exchange, final Provides provides) throws MessagingException {

        if (provides == null) {
            return;
        }

        final ExchangeProperties exchangeProperties = provides.getExchangeProperties();
        if (exchangeProperties != null) {
            for (final Param exchangeProperty : exchangeProperties.getExchangeProperty()) {
                exchange.setProperty(exchangeProperty.getName(), exchangeProperty.getValue());
            }
        }

        final MessageProperties messageProperties = provides.getMessageProperties();
        if (messageProperties != null) {
            for (final MessageParam messageProperty : messageProperties.getMessageProperty()) {
                final NormalizedMessage message = ((ExchangeImpl) exchange).getMessageExchange()
                        .getMessage(messageProperty.getMessage().name());
                message.setProperty(messageProperty.getName(), messageProperty.getValue());
            }
        }
    }

    /**
     * Forward the data from the In to the Out message.
     * 
     * @param exchange
     * @param provides
     * @throws MessagingException
     */
    private static final void forwardMessageData(final Exchange exchange, final Provides provides)
            throws MessagingException {

        if (provides == null) {
            return;
        }

        // forward security subject
        if (provides.isForwardSecuritySubject() != null
                && provides.isForwardSecuritySubject().booleanValue()) {
            if (exchange.getInSubject() != null && exchange.getOutSubject() == null) {
                exchange.setOutSubject(exchange.getInSubject());
            }
        }

        // forward message properties
        if (provides.isForwardMessageProperties() != null
                && provides.isForwardMessageProperties().booleanValue()) {
            for (final String propertyName : exchange.getInMessagePropertyNames()) {
                if (exchange.getOutMessageProperty(propertyName) == null) {
                    exchange.setOutMessageProperty(propertyName,
                            exchange.getInMessageProperty(propertyName));
                }
            }
        }

        // forward message attachment
        if (provides.isForwardAttachments() != null
                && provides.isForwardAttachments().booleanValue()) {
            final Set<String> attachmentNames = exchange.getInMessageAttachmentNames();
            for (final String attachmentName : attachmentNames) {
                if (exchange.getOutMessageAttachment(attachmentName) == null) {
                    exchange.setOutMessageAttachment(attachmentName,
                            exchange.getInMessageAttachment(attachmentName));
                }
            }
        }
    }
}
