/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Interceptor;
import org.ow2.petals.component.framework.api.Message.InterceptorPhase;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;

/**
 * The message interceptor processor. 
 * <p>
 * Used by the component to intercept messages.
 * </p>
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * @since 2.0
 */
public class MessageInterceptorProcessor {

    /**
     * The JBI component instance
     */
    protected AbstractComponent component;

    /**
     * The logger
     */
    protected Logger logger;

    /**
     * Creates a new instance of {@link MessageInterceptorProcessor}
     * 
     */
    public MessageInterceptorProcessor(final AbstractComponent component, final Logger logger) {
        this.component = component;
        this.logger = logger;
    }

    public void interceptSend(final Exchange exchange, final Consumes consumes) throws PEtALSCDKException {
        intercept(exchange, consumes, InterceptorPhase.SEND);
    }

    public void interceptSendResponse(final Exchange exchange, final Consumes consumes) throws PEtALSCDKException {
        intercept(exchange, consumes, InterceptorPhase.SEND_RESPONSE);
    }

    public void interceptAccept(final Exchange exchange, final Provides provides) throws PEtALSCDKException {
        intercept(exchange, provides, InterceptorPhase.ACCEPT);
    }

    public void interceptAcceptResponse(final Exchange exchange, final Provides provides) throws PEtALSCDKException {
        intercept(exchange, provides, InterceptorPhase.ACCEPT_RESPONSE);
    }

    /**
     * Intercepts the message.
     * 
     * @param exchange
     * @param extensible
     */
    public void intercept(final Exchange exchange, final Object service,
            final InterceptorPhase interceptorPhase) throws PEtALSCDKException {
        logger.fine("Intercepting message processing on " + interceptorPhase);

        this.checkInterceptorPhase(interceptorPhase, service);

        final AbstractServiceUnitManager suManager = component.getServiceUnitManager();

        // we are manipulating a Consumes or a Provides, the SU Manager can't be null
        assert suManager != null;

        final ServiceUnitDataHandler dataHandler = suManager.getSUDataHandler(service);

        if (dataHandler != null) { // no DH when no deployed SU
            MessageInterceptorManager messageInterceptorManager = dataHandler.getMessageInterceptorManager(service);

            // unactivate the check on the Role when intercepting at
            // reception,
            // to be able to reset the normalized message.
            if (InterceptorPhase.ACCEPT.equals(interceptorPhase)) {
                ((ExchangeImpl) exchange).setCheckRole(false);
            }
            this.intercept(exchange, service, messageInterceptorManager, interceptorPhase);
            if (InterceptorPhase.ACCEPT.equals(interceptorPhase)) {
                ((ExchangeImpl) exchange).setCheckRole(true);
            }
        }
    }

    private void checkInterceptorPhase(final InterceptorPhase interceptorPhase, final Object service)
            throws PEtALSCDKException {
        if (InterceptorPhase.ACCEPT.equals(interceptorPhase)
                || InterceptorPhase.ACCEPT_RESPONSE.equals(interceptorPhase)) {
            if (!(service instanceof Provides)) {
                throw new PEtALSCDKException(
                        "The InterceptorPhase is not a Provides InterceptorPhase");
            }
        } else {
            if (!(service instanceof Consumes)) {
                throw new PEtALSCDKException(
                        "The InterceptorPhase is not a Consumes InterceptorPhase");
            }
        }
    }

    /**
     * @param exchange
     * @param service
     * @param messageInterceptorManager
     * @param interceptorPhase
     * @throws IllegalArgumentException
     *             The value of <code>interceptorPhase</code> is invalid.
     */
    private void intercept(Exchange exchange, Object service,
            MessageInterceptorManager messageInterceptorManager, InterceptorPhase interceptorPhase) {
        Map<String, Interceptor> interceptors = null;
        Map<String, ConfigurationExtensions> additionalConfiguration = null;

        // getting good interceptors
        if (interceptorPhase == InterceptorPhase.ACCEPT) {
            interceptors = messageInterceptorManager.getInputOrderedInterceptors();
            additionalConfiguration = messageInterceptorManager.getInputAdditionalConfiguration();
        } else if (interceptorPhase == InterceptorPhase.SEND) {
            interceptors = messageInterceptorManager.getOutputOrderedInterceptors();
            additionalConfiguration = messageInterceptorManager.getOutputAdditionalConfiguration();
        } else if (interceptorPhase == InterceptorPhase.ACCEPT_RESPONSE) {
            interceptors = messageInterceptorManager.getInputResponseOrderedInterceptors();
            additionalConfiguration = messageInterceptorManager
                    .getInputResponseAdditionalConfiguration();
        } else if (interceptorPhase == InterceptorPhase.SEND_RESPONSE) {
            interceptors = messageInterceptorManager.getOutputResponseOrderedInterceptors();
            additionalConfiguration = messageInterceptorManager
                    .getOutputResponseAdditionalConfiguration();
        } else {
            // This exception can occur only when adding new item to the enum
            // InterceptorPhase.
            throw new IllegalArgumentException("Unknown InterceptorPhase: "
                    + interceptorPhase.toString());
        }

        for (final Map.Entry<String, Interceptor> entry : interceptors.entrySet()) {
            // get interceptor
            final Interceptor interceptor = entry.getValue();
            final String interceptorName = entry.getKey();

            logger.fine("Calling message interceptor " + interceptorName);

            // override component parameters with SU ones
            Map<String, String> finalParameters = new HashMap<String, String>();
            finalParameters.putAll(interceptor.getComponentParameters());
            if (additionalConfiguration.get(interceptorName) != null) {
                finalParameters.putAll(additionalConfiguration.get(interceptorName));
            }

            // intercept...
            if (interceptorPhase == InterceptorPhase.ACCEPT) {
                logger.fine("Intercepting message on accept InterceptorPhase");
                interceptor.handleMessageAccept(exchange, (Provides) service, finalParameters);
            } else if (interceptorPhase == InterceptorPhase.SEND) {
                logger.fine("Intercepting message on send InterceptorPhase");
                interceptor.handleMessageSent(exchange, (Consumes) service, finalParameters);
            } else if (interceptorPhase == InterceptorPhase.ACCEPT_RESPONSE) {
                logger.fine("Intercepting message on accept response InterceptorPhase");
                interceptor.handleMessageAcceptResponse(exchange, (Provides) service,
                        finalParameters);
            } else if (interceptorPhase == InterceptorPhase.SEND_RESPONSE) {
                logger.fine("Intercepting message on send response InterceptorPhase");
                interceptor
                        .handleMessageSentResponse(exchange, (Consumes) service, finalParameters);
            }
        }
    }
}
