/**
 * Copyright (c) 2009-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process.async;

import org.ow2.petals.component.framework.api.Constants.Component;

/**
 * The Asynchronous context.
 * <p>
 * The Asynchronous context holds the required data to handle asynchronous message responses in the JBI Listener:
 * <ul>
 * <li>the original message exchange is needed:
 * <ul>
 * <li>to be able to send reply to the service consumer,</li>
 * <li>to be able to process expired message sent asynchronously reinjecting this original message with the id of the
 * expired message as property,</li>
 * </ul>
 * </li>
 * <li>a TTL (Time-To-Live) to be able to manage invocation timeout.</li>
 * </ul>
 * The component developer can extends this class to provide its content of this context.
 * </p>
 * 
 * @author Roland Naudin - EBM WebSourcing
 */
public class AsyncContext {
    
    /**
     * TTL of the asynchronous sent. If <code>0</code>, no expiration will
     * occur. If less than <code>0</code>, it means it will be set by the CDK to the
     * default value from the JBI descriptor or {@value Component#DEFAULT_SEND_TIMEOUT}.
     */
    private long timeToLive = -1;
    
    /**
     * Date/Time when the asynchronous message will expired. The expiration
     * date/time is defined using {@link #setExpirationTime()}. No expiration is
     * defined by the value <code>0</code>.
     */
    private long expirationTime = -1;
    
    /**
     * Create an asynchronous context from an original message exchange with a
     * TTL set to its default value set in the extension of the JBI if it applies
     * or to ({@value Component#DEFAULT_SEND_TIMEOUT}). 
     * 
     * @param originalExchange
     *            Message which starts the processing needed asynchronous sent.
     *            It will be used to send the reply of the orchestration.
     */
    public AsyncContext() {
    }
    
    /**
     * Create an asynchronous context from an original message exchange with a TTL set to the defined value.
     * 
     * @param originalExchange
     *            Message which starts the processing needed asynchronous sent. It will be used to send the reply of the
     *            orchestration.
     * @param timeToLive
     *            Time to live of the asynchronous message. See {@link #timeToLive}.
     */
    public AsyncContext(final long timeToLive) {
        this.timeToLive = timeToLive;
    }

    /**
     * Change time to live (this has impacts only if the expiration hasn't been set yet)
     * 
     * @param timeToLive
     *            the new time to live. See {@link #timeToLive}.
     */
    public void setTimeToLive(long timeToLive) {
        this.timeToLive = timeToLive;
    }
    
    /**
     * @return the timeToLive (or null if none was set)
     */
    public long getTimeToLive() {
        return this.timeToLive;
    }
    
    /**
     * Set the expiration date/time from the time-to-live previously set.
     * 
     * @see {@link #expirationTime}
     */
    void setExpirationTime() {
        // it should have been set by the CDK by calling setTimeToLive()
        assert this.timeToLive >= 0;
        if (this.timeToLive != 0) {
            this.expirationTime = System.currentTimeMillis() + this.timeToLive;
        } else {
            this.expirationTime = 0;
        }
    }
    
    /**
     * @return the expirationTime
     */
    long getExpirationTime() {
        // it should have been set by the AsyncMessageManager by calling setExpirationTime()!
        assert this.expirationTime >= 0;
        return this.expirationTime;
    }
}
