/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process.async;

import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.process.JBIProcessorManager;



/**
 * An asynchronous message Manager.
 * @author Christophe Hamerling - EBM WebSourcing
 * @since 3.0
 */
public class AsyncMessageManager {

    static final String THREAD_SUFFIX = " - Asynchonous message exchange cleaner";

    /**
     * Prefix of the properties set on a message exchange to define it as sent asynchronously.
     * 
     * Note: they are not meant to be understood by the other side!!! But only by us when the exchange comes back!
     */
    public static final String ASYNC_MESSAGE_PROPERTY_PREFIX = "org.ow2.petals.component.framework.asynchronous.";

    private static final String ASYNC_MESSAGE_PROVIDER_PREFIX = "P";

    private static final String ASYNC_MESSAGE_CONSUMER_PREFIX = "C";

    private final JBIProcessorManager jbiProcessorManager;

    private final AbstractComponent component;

    private final Logger logger;

    /**
     * Map of contexts and message exchanges sent asynchronously, indexed by
     * identifier of message exchange sent.
     */
    private final ConcurrentMap<String, AsyncContextExchange> asynchronousContexts;
    
    /**
     * The timer needed to run cleaning task.
     */
    private final Timer timer;

    /**
     * {@link TimerTask} used to detect expired messages.
     */
    private TimerTask cleanerTask;

    private long timeBeetweenAsyncCleanerRuns;

    public AsyncMessageManager(final AbstractComponent component,
            final JBIProcessorManager jbiProcessorManager, final Logger logger) {
        this.component = component;
        this.logger = logger;
        this.asynchronousContexts = new ConcurrentHashMap<>();
        this.jbiProcessorManager = jbiProcessorManager;
        this.timer = new Timer(component.getContext().getComponentName() + THREAD_SUFFIX);
    }

    /**
     * Start the Async Message Manager.
     */
    public void start() {
        this.cleanerTask = new AsyncContextCleaner(this.asynchronousContexts, this.jbiProcessorManager, this.logger);
        this.timeBeetweenAsyncCleanerRuns = this.component.getJbiComponentDescriptor().getComponent()
                .getTimeBeetweenAsyncCleanerRuns().getValue();
        this.timer.schedule(this.cleanerTask, 0, this.timeBeetweenAsyncCleanerRuns);
    }

    /**
     * Stop the Async Message Manager.
     */
    public void stop() {
        this.cleanerTask.cancel();
    }

    /**
     * Shutdown the Async Message Manager.
     */
    public void shutdown() {
        this.timer.cancel();
    }

    long getTimeBeetweenAsyncCleanerRuns() {
        return this.timeBeetweenAsyncCleanerRuns;
    }

    /**
     * <p>
     * Reference an message as an asynchronous message:
     * <ul>
     * <li>adding the context to the asynchronous contexts map,</li>
     * <li>setting the expiration time of the message.</li>
     * </ul>
     * </p>
     * 
     * @param exchange
     *            The exchange sent asynchronously
     * @param asyncContext
     *            The asynchronous context associated to the exchange.
     * @return the key used to reference the context: in case the caller wants to revert this, it can use it directly!
     */
    public String referenceAsyncContext(final Exchange exchange, final AsyncContext asyncContext,
            final Object service) {

        final String type;
        final String prefix;
        if (exchange.isProviderRole()) {
            prefix = ASYNC_MESSAGE_PROVIDER_PREFIX;
            type = "as provider";
        } else {
            prefix = ASYNC_MESSAGE_CONSUMER_PREFIX;
            type = "as consumer";
        }

        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine("Add AsyncContext (" + type + ") for exchange " + exchange.getExchangeId());
        }

        final String key = prefix + exchange.getExchangeId();

        // Note we store also the exchangeId to prevent conflict when exchanges
        // are copied with their properties to other exchanges.
        // See PETALSDISTRIB-282 for details.
        exchange.setProperty(ASYNC_MESSAGE_PROPERTY_PREFIX + key, true);

        asyncContext.setExpirationTime();

        // Note: normally there can't be a conflict between provider and consumer's async contexts in one given
        // component for one given exchange...
        this.asynchronousContexts.put(key, new AsyncContextExchange(asyncContext, (ExchangeImpl) exchange, service));

        return key;
    }

    /**
     * Used to revert a reference after a failed send
     */
    public void revertReferenceAsyncContext(final String key) {
        this.asynchronousContexts.remove(key);
    }

    /**
     * <p>
     * Unreference an asynchronous context:
     * <ul>
     * <li>removing the context from the asynchronous messages map.</li>
     * </ul>
     * </p>
     * 
     * @param exchange
     *            answer to the asynchronous message to unreference
     * @return null if the asynchronous context is not referenced (not found in the map)
     */
    public AsyncContextExchange[] unreferenceAsyncContext(final Exchange exchange) {

        final String type;
        final String prefix;
        if (exchange.isProviderRole()) {
            prefix = ASYNC_MESSAGE_PROVIDER_PREFIX;
            type = "as provider";
        } else {
            prefix = ASYNC_MESSAGE_CONSUMER_PREFIX;
            type = "as consumer";
        }

        final String key = prefix + exchange.getExchangeId();
        final String property = ASYNC_MESSAGE_PROPERTY_PREFIX + key;

        final AsyncContextExchange context;
        if (exchange.getProperty(property) != null) {
            // A message previously sent asynchronously is returned

            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Remove AsyncContext (" + type + ") for exchange " + exchange.getExchangeId());
            }

            // let's clean the property
            exchange.setProperty(property, null);

            context = this.asynchronousContexts.remove(key);
            if (context == null) {
                return new AsyncContextExchange[] {};
            } else {
                return new AsyncContextExchange[] { context };
            }
        } else {
            return null;
        }
        // NOTE: the returned array is a sad hack to differentiate the context
        // not being there and the exchange not being an async exchange
    }
}
