/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.su;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.jbi.management.DeploymentServiceMBean;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.jbidescriptor.generated.SUInterceptor;
import org.ow2.petals.component.framework.jbidescriptor.generated.SUInterceptors;
import org.ow2.petals.component.framework.util.EndpointUtil;
import org.ow2.petals.component.framework.util.ManagementMessageUtil;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.ow2.petals.component.framework.util.ServiceUnitUtil;
import org.ow2.petals.component.framework.util.WSDLUtilImpl;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.lang.ExceptionHelper;
import com.ebmwebsourcing.easycommons.lang.StringHelper;

/**
 * The common part of the ServiceUnitManager. 
 * <p>
 * It handles service units JBI descriptor during their deployment, life-cycle and undeployment.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 * @author Adrien Louis - EBM WebSourcing
 */
public abstract class AbstractServiceUnitManager implements javax.jbi.component.ServiceUnitManager {

    /**
     * The reference to the component
     */
    private final AbstractComponent component;

    /**
     * The component's Logger
     */
    protected Logger logger;

    /**
     * The service unit data handlers. Each service unit is linked to a
     * datahandler. Key is the service unit name.
     * 
     * @see {@link getServiceUnitDataHandlers()}
     * 
     */
    private final ConcurrentMap<String, ServiceUnitDataHandler> serviceUnitDataHandlers;

    private final boolean activateEndpoints;

    public AbstractServiceUnitManager(final AbstractComponent component) {
        this(component, true);
    }

    public AbstractServiceUnitManager(final AbstractComponent component, final boolean activateEndpoints) {
        this.component = component;
        this.activateEndpoints = activateEndpoints;
        this.logger = component.getLogger();
        this.serviceUnitDataHandlers = new ConcurrentHashMap<>();
    }

    @Override
    public final String deploy(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {

        if (this.serviceUnitDataHandlers.containsKey(serviceUnitName)) {
            throw new DeploymentException(
                    "Failed to deploy Service Unit '" + serviceUnitName
                            + "', another SU with the same name already exists");
        }

        this.logger.log(Level.INFO, "Deploying Service Unit '" + serviceUnitName + "'");

        Exception exception = null;
        Jbi jbiDescriptor = null;
        ServiceUnitDataHandler suDH = null;
        boolean internal = false;
        boolean implementation = false;
        try {
            // let's do all the pre-processing of the descriptor
            // before using the values for the su handler and other things
            jbiDescriptor = getFinalJbiDescriptor(suRootPath);

            suDH = createSUDH(serviceUnitName, suRootPath, jbiDescriptor);

            // register the datahandler here, because of the call of the method
            // 'getServiceDescription' when activating an endpoint
            this.logger.log(Level.FINE, "Register a data handler for SU '" + serviceUnitName + "'");
            this.serviceUnitDataHandlers.put(serviceUnitName, suDH);

            internal = true;
            deployInternal(suDH);

            implementation = true;
            doDeploy(suDH);

            suDH.setState(DeploymentServiceMBean.SHUTDOWN);
        } catch (final Exception e) {
            exception = e;
            if (jbiDescriptor == null) {
                this.logger.log(Level.SEVERE,
                        "Failed to parse the JBI descriptor for Service Unit '" + serviceUnitName + "'", e);
            } else if (suDH == null) {
                this.logger.log(Level.SEVERE,
                        "Failed to create a Data Handler for Service Unit '" + serviceUnitName + "', reverting", e);
            } else if (internal && !implementation) {
                this.logger.log(Level.SEVERE, "Failed to deploy Service Unit '" + serviceUnitName + "', reverting", e);
            } else {
                this.logger.log(Level.SEVERE,
                        "Failed to deploy Service Unit '" + serviceUnitName + "', reverting and calling undeploy", e);
            }

            final String revertError = "Failed to undeploy the Service Unit '" + serviceUnitName
                    + "' during revert from failed deploy";

            if (implementation) {
                try {
                    doUndeploy(suDH);
                } catch (final Exception ex) {
                    this.logger.log(Level.WARNING, revertError, ex);
                }
            }
            
            if (internal) {
                try {
                    undeployInternal(suDH);
                } catch (final Exception ex) {
                    this.logger.log(Level.WARNING, revertError, ex);
                }
            }
        } finally {
            if (exception != null) {
                this.serviceUnitDataHandlers.remove(serviceUnitName);
            }
        }

        final String result;
        if (exception == null) {
            result = ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "deploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
            this.logger.log(Level.INFO, "Service Unit '" + serviceUnitName + "' deployed");
        } else {
            // instead of throwing an exception, return an error message
            result = ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "deploy", ManagementMessageUtil.TASK_RESULT_FAILED,
                    ManagementMessageUtil.MESSAGE_TYPE_ERROR, "1",
                    "Failed to deploy Service Unit : {1}", new String[] { serviceUnitName },
                    ExceptionHelper.getStackTrace(exception));
        }

        return result;
    }

    private ServiceUnitDataHandler createSUDH(final String serviceUnitName, final String suRootPath,
            final Jbi jbiDescriptor) throws Exception {
        final ServiceUnitDataHandler suDH = new ServiceUnitDataHandler(serviceUnitName, suRootPath, jbiDescriptor,
                this.component.getPlaceHolders());

        // these two methods only modify the suDH
        processConsumesNodes(suDH);
        processProvidesNodes(suDH);
        return suDH;
    }

    /**
     * Meant to be used for CDK, please do not override and use {@link #doDeploy(ServiceUnitDataHandler)}!
     */
    protected void deployInternal(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    private Jbi getFinalJbiDescriptor(final String suRootPath) throws Exception {
        final File jbiXmlFile = new File(suRootPath, Constants.Component.META_INF + File.separator
                + Constants.Component.JBI_XML);
        try (final FileInputStream fisJbi = new FileInputStream(jbiXmlFile)) {
            final Jbi jbiDescriptor = CDKJBIDescriptorBuilder.getInstance().buildJavaJBIDescriptor(fisJbi);

            boolean bSaveJBI = false;

            for (Provides provides : jbiDescriptor.getServices().getProvides()) {

                final Description wsdlDescription = ServiceUnitUtil.getWsdlDescription(suRootPath, provides);

                String endpointName = provides.getEndpointName();

                boolean bSaveWsdl = false;

                if (Constants.Component.AUTOGENERATED_ENDPOINT_NAME.equals(endpointName)) {
                    endpointName = EndpointUtil.generateEndpointName();
                    // update the provides object to set the real endpoint
                    provides.setEndpointName(endpointName);
                    bSaveJBI = true;

                    // change arbitrary all the endpoint named 'autogenerate' found to
                    // the given auto-generated value
                    final Service service = wsdlDescription.getService(provides.getServiceName());
                    if (service != null) {
                        final Endpoint autogeneratedEndpoint = service
                                .getEndpoint(Constants.Component.AUTOGENERATED_ENDPOINT_NAME);
                        if (autogeneratedEndpoint != null) {
                            autogeneratedEndpoint.setName(endpointName);
                        }
                        bSaveWsdl = true;
                    } else {
                        throw new JBIException("Failed to find provided service '" + provides.getServiceName()
                                + "' in the WSDL description");
                    }
                }

                // Update dynamically the service name of the WSDL
                // if the implementation of the SU Manager wants (see getServiceName)
                final QName originalServiceName = provides.getServiceName();
                final QName newServiceName = this.getServiceName(originalServiceName);
                if (newServiceName != null) {
                    // Update the JBI descriptor with the new service name
                    provides.setServiceName(newServiceName);
                    bSaveJBI = true;

                    // Update the WSDL with the new service name
                    wsdlDescription.getService(originalServiceName).setQName(newServiceName);
                    bSaveWsdl = true;
                }

                // Save changes of the WSDL
                if (bSaveWsdl) {
                    AbstractServiceUnitManager.saveWsdl(suRootPath, provides, wsdlDescription);
                }
            }

            if (bSaveJBI) {
                AbstractServiceUnitManager.saveJBI(suRootPath, jbiDescriptor);
            }

            return jbiDescriptor;
        }
    }

    /**
     * Get the consumes matching the given endpoint name.
     * 
     * @param serviceName
     *            The endpoint name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    private Consumes getConsumesForEndpoint(final String endpointName) {
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (endpointName.equals(consumes.getEndpointName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the consumes matching the given service name.
     * 
     * @param serviceName
     *            The service qualified name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    private Consumes getConsumesForInterface(final QName interfaceName) {
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (interfaceName.equals(consumes.getInterfaceName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the consumes matching the given service name.
     * 
     * @param serviceName
     *            The service qualified name to match
     * @return The {@code Consumes} instance if matched, @ null} otherwise
     */
    private Consumes getConsumesForService(final QName serviceName) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            for (final Consumes consumes : dataHandler.getDescriptor().getServices().getConsumes()) {
                if (serviceName.equals(consumes.getServiceName())) {
                    return consumes;
                }
            }
        }
        return null;
    }

    /**
     * Get the Consumes element for the given exchange, see {@link #getConsumesFromDestination(String, QName, QName)}.
     */
    public Consumes getConsumesFromExchange(final Exchange exchange) {
        final ServiceEndpoint endpoint = exchange.getEndpoint();
        final String endpointName;
        final QName serviceName;
        if (endpoint != null) {
            endpointName = endpoint.getEndpointName();
            serviceName = endpoint.getServiceName();
        } else {
            endpointName = null;
            serviceName = exchange.getService();
        }
        return getConsumesFromDestination(endpointName, serviceName, exchange.getInterfaceName());
    }

    /**
     * Retrieve the Consumes node matching the given destination information
     * (endpoint name, service name, interface name). Search order is by
     * ascending priority : endpointName, serviceName, interfaceName. If there
     * are many consumes matching information, return the first one found.
     * 
     * @param endpointName
     *            an endpoint name, can be null
     * @param serviceName
     *            a service {@link QName}, can be null
     * @param interfaceName
     *            an interface {@link QName}, can be null
     * @return the {@link Consumes} element matching information, can be null
     */
    public Consumes getConsumesFromDestination(final String endpointName, final QName serviceName,
            final QName interfaceName) {
        Consumes result = null;
        if (endpointName != null) {
            result = this.getConsumesForEndpoint(endpointName);
        }
        if (result == null) {
            if (serviceName != null) {
                result = this.getConsumesForService(serviceName);
            }
            if (result == null) {
                if (interfaceName != null) {
                    result = this.getConsumesForInterface(interfaceName);
                }
            }
        }
        return result;
    }

    /**
     * Get the Provides element for the given exchange, see {@link #getProvidesFromEndpoint(ServiceEndpoint)}.
     * 
     */
    public Provides getProvidesFromExchange(final Exchange exchange) {
        return getProvidesFromEndpoint(exchange.getEndpoint());
    }

    /**
     * Retrieve the Provides node that has activated the given jbi Endpoint
     */
    public Provides getProvidesFromEndpoint(final ServiceEndpoint endpoint) {
        if (endpoint != null) {
            final ServiceUnitDataHandler suDH = getSUDataHandler(endpoint);
            if (suDH != null) {
                return suDH.getProvides(endpoint);
            }
        }
        return null;
    }

    /**
     * Retrieve the Service description (as DOM Document) of the given Provides
     */
    public Document getServiceDescription(final ServiceEndpointKey key) {
        if (key != null) {
            for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
                final Document desc = dataHandler.getEndpointDescription(key);
                if (desc != null) {
                    return desc;
                }
            }
        }
        return null;
    }

    /**
     * Return all {@link ServiceUnitDataHandler}s of the component
     */
    public Collection<ServiceUnitDataHandler> getServiceUnitDataHandlers() {
        return Collections.unmodifiableCollection(this.serviceUnitDataHandlers.values());
    }

    /**
     * Get the SU Data Handler for the given SU name.
     * 
     */
    public ServiceUnitDataHandler getSUDataHandler(final String serviceUnitName) {
        return this.serviceUnitDataHandlers.get(serviceUnitName);
    }

    private ServiceUnitDataHandler getSUDataHandler(final String serviceUnitName, final String expectedState)
            throws DeploymentException {
        final ServiceUnitDataHandler suDH = getSUDataHandler(serviceUnitName);
        if (suDH == null) {
            throw new DeploymentException(serviceUnitName + " is not deployed on this component");
        } else if (!suDH.getState().equals(expectedState)) {
            throw new DeploymentException(serviceUnitName + " is not in the " + expectedState + " state");
        }
        return suDH;
    }

    /**
     * Get the SU Data Handler matching the given Consumes.
     */
    public ServiceUnitDataHandler getSUDataHandler(final Consumes consumes) {
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            if (dataHandler.hasConsumes(consumes)) {
                return dataHandler;
            }
        }
        return null;
    }

    /**
     * Retrieve the data handler of the Service Unit that has activated the given endpoint
     */
    public ServiceUnitDataHandler getSUDataHandler(final ServiceEndpoint endpoint) {
        // it is important to look for endpoint by their key because we are never sure the object is not a different one
        // (because of the way JBI works)
        return getSUDataHandler(new ServiceEndpointKey(endpoint));
    }

    /**
     * Retrieve the data handler of the Service Unit for the given Provides's key
     */
    public ServiceUnitDataHandler getSUDataHandler(final ServiceEndpointKey key) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            if (dataHandler.hasProvides(key)) {
                return dataHandler;
            }
        }

        return null;
    }

    /**
     * Get the SU Data Handler matching the given Consumes.
     * 
     * @param provides
     *            The @ code Provides} instance
     * @return
     */
    public ServiceUnitDataHandler getSUDataHandler(final Provides provides) {
        // No synchronization on serviceUnitDataHandlers is needed because
        // it's a ConcurrentMap
        for (final ServiceUnitDataHandler dataHandler : this.serviceUnitDataHandlers.values()) {
            if (dataHandler.hasProvides(provides)) {
                return dataHandler;
            }
        }
        return null;
    }

    /**
     * Get the SU Data Handler matching the given Consumes/Provides service.
     */
    public ServiceUnitDataHandler getSUDataHandler(final Object service) {
        ServiceUnitDataHandler suDataHandler = null;

        if (service instanceof Consumes) {
            suDataHandler = this.getSUDataHandler((Consumes) service);
        } else if (service instanceof Provides) {
            suDataHandler = this.getSUDataHandler((Provides) service);
        }

        return suDataHandler;
    }

    @Override
    public final void init(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        this.logger.log(Level.INFO, "Initializing Service Unit '" + serviceUnitName + "'");

        final ServiceUnitDataHandler suDH = getSUDataHandler(serviceUnitName, DeploymentServiceMBean.SHUTDOWN);

        boolean implementation = false;
        try {
            initInternal(suDH);

            implementation = true;
            // Call specific processing on init phase
            doInit(suDH);

            if (activateEndpoints) {
                // and then finally activate the endpoints
                activateEndpoints(suDH);
            }

            suDH.setState(DeploymentServiceMBean.STOPPED);
        } catch (final Exception e) {
            this.logger.log(Level.SEVERE, "Failed to init Service Unit '" + serviceUnitName + "', reverting"
                    + (implementation ? " and calling shutdown" : ""));

            if (activateEndpoints) {
                // unregister if an error happens (this doesn't throw any exception)
                deactivateEndpoints(suDH);
            }

            final String revertError = "Failed to shutdown the Service Unit '" + serviceUnitName
                    + "' during revert from failed init";

            if (implementation) {
                try {
                    doShutdown(suDH);
                } catch (final PEtALSCDKException ex) {
                    this.logger.log(Level.WARNING, revertError, ex);
                }
            }

            try {
                shutdownInternal(suDH);
            } catch (final PEtALSCDKException ex) {
                this.logger.log(Level.WARNING, revertError, ex);
            }

            throw new DeploymentException("Failed to initialize SU", e);
        }

        this.logger.log(Level.INFO, "Service unit '" + serviceUnitName + "' initialized");
    }

    /**
     * Meant to be used for CDK, please do not override and use {@link #doInit(ServiceUnitDataHandler)}!
     */
    protected void initInternal(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    @Override
    public final void shutDown(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.INFO, "Shutting down Service Unit '" + serviceUnitName + "'");

        final ServiceUnitDataHandler suDH = getSUDataHandler(serviceUnitName, DeploymentServiceMBean.STOPPED);

        Exception exception = null;
        try {
            if (activateEndpoints) {
                // first lets deactivate endpoints (this doesn't throw any exception!)
                deactivateEndpoints(suDH);
            }

            // Call specific processing on shutting down phase
            try {
                doShutdown(suDH);
            } catch (final Exception e) {
                exception = e;
            }

            try {
                shutdownInternal(suDH);
            } catch (final Exception e) {
                if (exception != null) {
                    exception.addSuppressed(e);
                } else {
                    exception = e;
                }
            }
        } finally {
            suDH.setState(DeploymentServiceMBean.SHUTDOWN);
        }

        if (exception != null) {
            this.logger.log(Level.SEVERE, "Failed to shutdown the Service Unit '" + serviceUnitName + "'", exception);
        } else {
            this.logger.log(Level.INFO, "Service unit '" + serviceUnitName + "' shut down");
        }
    }

    /**
     * Meant to be used for CDK, please do not override and use {@link #doShutdown(ServiceUnitDataHandler)}!
     */
    protected void shutdownInternal(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    @Override
    public final void start(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.INFO, "Starting Service Unit '" + serviceUnitName + "'");

        final ServiceUnitDataHandler suDH = getSUDataHandler(serviceUnitName, DeploymentServiceMBean.STOPPED);

        boolean implementation = false;
        try {
            this.startInternal(suDH);

            // Call specific processing on starting phase
            implementation = true;
            this.doStart(suDH);

            suDH.setState(DeploymentServiceMBean.STARTED);

            this.logger.log(Level.INFO, "Service unit '" + serviceUnitName + "' started");
        } catch (final Exception e) {
            this.logger.log(Level.SEVERE, String.format(
                    "An error occurs starting the service unit '%s', reverting%s. See next traces to know the error cause.",
                    serviceUnitName, implementation ? " and calling stop" : ""));

            final String revertError = String
                    .format("Failed to stop the service unit '%s' during revert from failed start", serviceUnitName);

            if (implementation) {
                try {
                    this.doStop(suDH);
                } catch (final Exception ex) {
                    this.logger.log(Level.WARNING, revertError, ex);
                }
            }

            try {
                this.stopInternal(suDH);
            } catch (final Exception ex) {
                this.logger.log(Level.WARNING, revertError, ex);
            }

            throw new DeploymentException("Failed to start SU", e);
        }
    }

    /**
     * Meant to be used for CDK, please do not override and use {@link #doStart(ServiceUnitDataHandler)}!
     */
    protected void startInternal(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    private void activateEndpoints(final ServiceUnitDataHandler handler) throws JBIException {
        for (final Provides provides : handler.getDescriptor().getServices().getProvides()) {
            // register and get the new endpoint instance
            final ServiceEndpoint ep = this.component.getContext().activateEndpoint(provides.getServiceName(),
                    provides.getEndpointName());

            this.logger.log(Level.INFO, "New Service Endpoint deployed: " + ep);

            handler.addServiceEndpoint(ep);
        }
    }

    private void deactivateEndpoints(final ServiceUnitDataHandler handler) {
        // we need a copy of the collection because we call
        // a method that modify it in the loop
        final Collection<ServiceEndpoint> values = new ArrayList<>(handler.getServiceEndpoints());
        for (ServiceEndpoint endpoint : values) {
            try {
                this.component.getContext().deactivateEndpoint(endpoint);
                this.logger.log(Level.INFO, "Service Endpoint undeployed: " + endpoint);
            } catch (final JBIException e) {
                this.logger.log(Level.WARNING,
                        "Failed to deactivate endpoint '" + endpoint + " ' of Service unit '" + handler.getName() + "'",
                        e);
            } finally {
                handler.removeEndpoint(endpoint);
            }
        }
    }

    public Collection<ServiceEndpoint> getEndpointsForConsumes(final Consumes consumes) {

        final String endpointName = consumes.getEndpointName();
        final QName serviceName = consumes.getServiceName();
        final QName interfaceName = consumes.getInterfaceName();

        final Collection<ServiceEndpoint> endpoints;
        if (endpointName != null && serviceName != null) {
            final ServiceEndpoint endpoint = component.getContext().getEndpoint(serviceName, endpointName);
            if (endpoint != null) {
                if (interfaceName == null || EndpointUtil.matches(endpoint, interfaceName)) {
                    endpoints = Collections.singletonList(endpoint);
                } else {
                    logger.warning(String.format(
                            "Endpoint found for Consumes %s/%s/%s but interface does not match (was %s)", endpointName,
                            serviceName, interfaceName, Arrays.deepToString(endpoint.getInterfaces())));
                    endpoints = Collections.emptyList();
                }
            } else {
                endpoints = Collections.emptyList();
            }
        } else if (serviceName != null) {
            final ServiceEndpoint[] preMatch = component.getContext().getEndpointsForService(serviceName);
            if (interfaceName != null) {
                final List<ServiceEndpoint> matched = new ArrayList<>();
                for (final ServiceEndpoint endpoint : preMatch) {
                    assert endpoint != null;
                    if (EndpointUtil.matches(endpoint, interfaceName)) {
                        matched.add(endpoint);
                    } else {
                        logger.warning(
                                String.format("Endpoint found for Consumes %s/%s but interface does not match (was %s)",
                                        serviceName, interfaceName, Arrays.deepToString(endpoint.getInterfaces())));
                    }
                }
                endpoints = matched;
            } else {
                endpoints = Arrays.asList(preMatch);
            }
        } else {
            endpoints = Arrays.asList(component.getContext().getEndpoints(interfaceName));
        }

        return endpoints;
    }

    @Override
    public final void stop(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.INFO, "Stopping Service Unit '" + serviceUnitName + "'");

        final ServiceUnitDataHandler suDH = getSUDataHandler(serviceUnitName, DeploymentServiceMBean.STARTED);

        Exception exception = null;
        try {
            try {
                // Call specific processing on stop phase
                doStop(suDH);
            } catch (final Exception e) {
                exception = e;
            }

            try {
                stopInternal(suDH);
            } catch (final Exception e) {
                if (exception != null) {
                    exception.addSuppressed(e);
                } else {
                    exception = e;
                }
            }
        } finally {
            suDH.setState(DeploymentServiceMBean.STOPPED);
        }

        if (exception != null) {
            this.logger.log(Level.SEVERE, "Failed to stop the Service Unit '" + serviceUnitName + "'", exception);
        } else {
            this.logger.log(Level.INFO, "Service unit '" + serviceUnitName + "' stopped");
        }
    }

    /**
     * Meant to be used for CDK, please do not override and use {@link #doStop(ServiceUnitDataHandler)}!
     */
    protected void stopInternal(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    @Override
    public final String undeploy(final String serviceUnitName, final String suRootPath) throws DeploymentException {
        this.logger.log(Level.INFO, "Undeploying Service Unit " + serviceUnitName);

        final ServiceUnitDataHandler suDH = getSUDataHandler(serviceUnitName, DeploymentServiceMBean.SHUTDOWN);

        Exception exception = null;
        try {
            try {
                doUndeploy(suDH);
            } catch (final Exception e) {
                exception = e;
            }

            try {
                undeployInternal(suDH);
            } catch (final Exception e) {
                if (exception != null) {
                    exception.addSuppressed(e);
                } else {
                    exception = e;
                }
            }
        } finally {
            this.serviceUnitDataHandlers.remove(serviceUnitName);
        }

        if (exception != null) {
            this.logger.log(Level.SEVERE, "Failed to undeploy the Service Unit '" + serviceUnitName + "'", exception);
            // instead of throwing an exception, return an error message
            return ManagementMessageUtil.getComponentTaskResult(this.component.getContext().getComponentName(),
                    "undeploy", ManagementMessageUtil.TASK_RESULT_FAILED, ManagementMessageUtil.MESSAGE_TYPE_ERROR, "1",
                    "Failed to undeploy Service Unit {1}", new String[] { serviceUnitName },
                    ExceptionHelper.getStackTrace(exception));
        } else {
            this.logger.log(Level.INFO, "Service unit '" + serviceUnitName + "' undeployed");
            return ManagementMessageUtil.getComponentTaskResult(this.component.getContext().getComponentName(),
                    "undeploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
        }
    }

    /**
     * Meant to be used for CDK, please do not override and use {@link #doUndeploy(ServiceUnitDataHandler)}!
     */
    protected void undeployInternal(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the deploy step.
     * 
     * After deploy, resources must be allocated, but the component should not have to process requests for the SU
     * Provides nor send requests from the SU Consumes. The endpoints are not yet activated.
     * 
     * In case of error happening, the container will execute {@link #undeploy(String, String)} and
     * {@link #doUndeploy(ServiceUnitDataHandler)} will thus be called!
     * 
     */
    protected void doDeploy(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the init step.
     * 
     * After init, the component must be able to answer requests for the SU Provides and the endpoints will be activated
     * just after this method is invoked.
     * 
     * In case of error happening, the container will execute {@link #shutDown(String)} and
     * {@link #doShutdown(ServiceUnitDataHandler)} will thus be called!
     */
    protected void doInit(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the start step.
     * 
     * Requests can be sent safely now.
     * 
     * In case of error happening, the container will execute {@link #stop(String)} and
     * {@link #doStop(ServiceUnitDataHandler)} will thus be called!
     */
    protected void doStart(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the stop step.
     * 
     * No new requests should be sent anymore after stop.
     * 
     * Care must be taken to clean everything even in case of errors during stop.
     */
    protected void doStop(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the shut down step.
     * 
     * Before this method is executed, endpoints are deactivated. After, the component won't answer new requests for the
     * SU Provides nor send requests from the SU Consumes.
     * 
     * Notice that unfinished requests can potentially still arrive to end their business, hence after shutdown the
     * component should still be able to take care of them!
     * 
     * Care must be taken to clean everything even in case of errors during shutdown.
     */
    protected void doShutdown(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Override this method to implement your component logic on the undeploy step.
     * 
     * After undeploy, the component should have released all the resources associated to the SU.
     * 
     * Care must be taken to clean everything even in case of errors during undeploy.
     */
    protected void doUndeploy(final ServiceUnitDataHandler suDH) throws PEtALSCDKException {
        // do nothing
    }

    /**
     * Process the consumes elements.
     * 
     * @param handler
     * @throws JBIException
     */
    private void processConsumesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        final Jbi descriptor = handler.getDescriptor();
        final List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {
            this.processInterceptors(handler, consumes);
        }
    }

    /**
     * Process the interceptors for a consumes/provides service.
     * 
     * @param handler
     * @param service
     *            the {@code Provides} or {@code Consumes} service
     * @throws JBIException
     */
    private void processInterceptors(final ServiceUnitDataHandler handler, final Object service)
            throws JBIException {

        final SUInterceptors suInterceptors;

        if (service instanceof Consumes) {
            suInterceptors = ((Consumes) service).getSuInterceptors();
        } else if (service instanceof Provides) {
            suInterceptors = ((Provides) service).getSuInterceptors();
        } else {
            throw new PEtALSCDKException(
                    "The service parameter must either a Consumes class or a Provides class");
        }

        // create the manager for this extension element. Each element has a
        // linked manager with preloaded interceptors.

        final MessageInterceptorManager interceptorManager;
        final String propertiesFile = this.component.getJbiComponentDescriptor().getComponent().getPropertiesFile()
                .getValue();
        if (!StringHelper.isNullOrEmpty(propertiesFile)) {
            interceptorManager = new MessageInterceptorManager(propertiesFile, this.logger);
        } else {
            interceptorManager = new MessageInterceptorManager(this.logger);
        }

        List<SUInterceptor> sendSUInterceptors = null;
        List<SUInterceptor> sendResponseSUInterceptors = null;
        List<SUInterceptor> acceptSUInterceptors = null;
        List<SUInterceptor> acceptResponseSUInterceptors = null;
        if (suInterceptors != null) {
            if (suInterceptors.getSend() != null) {
                sendSUInterceptors = suInterceptors.getSend().getInterceptor();
            }
            if (suInterceptors.getSendResponse() != null) {
                sendResponseSUInterceptors = suInterceptors.getSendResponse().getInterceptor();
            }
            if (suInterceptors.getAccept() != null) {
                acceptSUInterceptors = suInterceptors.getAccept().getInterceptor();
            }
            if (suInterceptors.getAcceptResponse() != null) {
                acceptResponseSUInterceptors = suInterceptors.getAcceptResponse().getInterceptor();
            }
        }

        interceptorManager.init(this.component.getInterceptors(), acceptSUInterceptors,
                sendSUInterceptors, acceptResponseSUInterceptors, sendResponseSUInterceptors);

        handler.addMessageInterceptorManager(service, interceptorManager);
    }

    /**
     * Process the given provides node. Search the valid location of the wsdl
     * document, or generate one new if needed generate the endpoints and then
     * update the provides, the description and the wsdl file if needed Activate
     * the new endpoint
     * 
     * @param handler
     *            the service unit handler
     * @param provides
     *            the provides of the su
     * @return The resulted Service Endpoint
     * @throws JBIException
     * @throws WSDLException
     * @throws IOException
     */
    private void processProvidesNode(final ServiceUnitDataHandler handler, final Provides provides)
            throws WSDLException, IOException, JBIException {

        final Description wsdlDescription = ServiceUnitUtil.getWsdlDescription(handler.getInstallRoot(), provides);
        final String endpointName = provides.getEndpointName();
        final QName serviceName = provides.getServiceName();
        final Document wsdlDocument = WSDLUtilImpl.convertDescriptionToDocument(wsdlDescription);

        if (provides.getValidateWsdl() == null || provides.getValidateWsdl().isValue()) {
            this.logger.fine("Validating the provided WSDL description");
            ServiceUnitUtil.validateDescription(wsdlDescription, provides);
        }

        this.processInterceptors(handler, provides);

        // note this will be used when the endpoint is activated by the container to get the description
        handler.addProvidesData(provides, wsdlDocument,
                WSDLUtilImpl.getEndpointOperations(wsdlDescription, serviceName, endpointName));
    }

    private static void saveJBI(final String suRootPath, final Jbi jbiDecriptor)
            throws JBIDescriptorException, IOException {
        final File jbiXmlFile = new File(suRootPath, Constants.Component.META_INF
                + File.separator + Constants.Component.JBI_XML);

        try (final OutputStream jbiOutputStream = new FileOutputStream(jbiXmlFile)) {
            CDKJBIDescriptorBuilder.getInstance().writeXMLJBIdescriptor(jbiDecriptor, jbiOutputStream);
        }
    }

    private static void saveWsdl(final String suRootPath, final Provides provides, final Description wsdlDescription)
            throws Exception {

        if (provides.getWsdl() != null) {
            WSDLUtilImpl.convertDescriptionToFile(wsdlDescription, new File(suRootPath, provides.getWsdl()), false);
        }
    }

    /**
     * Extension point to update the service name according to the component
     * needs.
     * 
     * @param originalServiceName
     *            The service name read from the SU archive
     * @return The service name to use, or <code>null</code> if the original
     *         service name must be kept.
     */
    protected QName getServiceName(final QName originalServiceName) {
        return null;
    }

    /**
     * Process the Provides elements.
     * 
     * @param handler
     * @throws IOException
     * @throws WSDLException
     * @throws JBIException
     */
    private void processProvidesNodes(final ServiceUnitDataHandler handler)
            throws WSDLException, IOException, JBIException {
        for (final Provides provides : handler.getDescriptor().getServices().getProvides()) {
            this.processProvidesNode(handler, provides);
        }
    }

    /**
     * Place holder values was reload, we propagate this to resolve the new values of service unit config params
     */
    public final void reloadPlaceHolders() {
        for (final ServiceUnitDataHandler serviceUnitDataHandler : this.serviceUnitDataHandlers.values()) {
            serviceUnitDataHandler.onPlaceHolderValuesReloaded();
        }

        onPlaceHolderValuesReloaded();
    }

    /**
     * Override to add component-specific placeholder reloading logic
     */
    protected void onPlaceHolderValuesReloaded() {
        // do nothing
    }

    protected AbstractComponent getComponent() {
        return this.component;
    }
}
