/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.su;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.jbi.management.DeploymentServiceMBean;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.api.configuration.SuConfigurationParameters;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.w3c.dom.Document;

/**
 * Handles useful information about a Service Unit.
 * <ul>
 * <li>The service unit description (jbi descriptor)</li>
 * <li>Links between Provides nodes and activated endpoints.</li>
 * <li>Links between activated endpoints and corresponding service description.</li>
 * <li>Links between Consumes nodes and corresponding addresses.</li>
 * <li>The installation root.</li>
 * <li>The service unit name - the message interceptor manager.</li>
 * </ul>
 * <p>
 * A SUDataHandler instance is available for each SU of the component.
 * </p>
 * 
 * @author Olivier Fabre - EBM WebSourcing
 */
public class ServiceUnitDataHandler {

    private Jbi descriptor;

    private Map<ServiceEndpointKey, ProvidesData> epProvides;

    /**
     * Map of service/MessageInterceptorManager
     */
    private final Map<Object, MessageInterceptorManager> messageInterceptorManagersMap;

    /**
     * Map of service/ConfigurationExtensions
     */
    private final Map<Object, SuConfigurationParameters> configurationExtensionsMap;

    /**
     * Map of Consumes/ExchangeProperties
     */
    private final Map<Consumes, SuConfigurationParameters> exchangePropertiesMap;

    private final String installRoot;

    private final String name;

    private String state = DeploymentServiceMBean.SHUTDOWN;

    /**
     * Creates a new instance of {@link ServiceUnitDataHandler}
     * 
     * @param placeHolders
     *            The place holder values used to configure configuration parameters of the service-unit (providing from
     *            the property file set at component level). Not <code>null</code>
     * @throws PEtALSCDKException
     * 
     */
    public ServiceUnitDataHandler(final String suName, final String installRoot, final Jbi jbiDescriptor,
            final Properties placeHolders) {
        this.epProvides = new HashMap<>();
        this.messageInterceptorManagersMap = new HashMap<>();
        this.configurationExtensionsMap = new HashMap<>();
        this.exchangePropertiesMap = new HashMap<>();

        this.name = suName;
        this.installRoot = installRoot;
        this.descriptor = jbiDescriptor;

        for (Consumes consumes : this.descriptor.getServices().getConsumes()) {
            this.configurationExtensionsMap.put(consumes,
                    new SuConfigurationParameters(consumes.getAny(), placeHolders));
            if (consumes.getExchangeProperties() != null) {
                this.exchangePropertiesMap.put(consumes, new SuConfigurationParameters(
                        consumes.getExchangeProperties().getExchangeProperty(), placeHolders));
            }
        }

        for (Provides provides : this.descriptor.getServices().getProvides()) {
            this.configurationExtensionsMap.put(provides,
                    new SuConfigurationParameters(provides.getAny(), placeHolders));
        }
    }

    /**
     * Place holder values was reload, we propagate this to resolve the new values of service unit config params
     */
    void onPlaceHolderValuesReloaded() {

        for (final SuConfigurationParameters suConfParams : this.configurationExtensionsMap.values()) {
            suConfParams.reloadPlaceholderValues();
        }

        for (final SuConfigurationParameters suConfParams : this.exchangePropertiesMap.values()) {
            suConfParams.reloadPlaceholderValues();
        }
    }
    
    public Document getEndpointDescription(final ServiceEndpoint endpoint) {
        return getEndpointDescription(new ServiceEndpointKey(endpoint));
    }

    public Document getEndpointDescription(final Provides provides) {
        return getEndpointDescription(new ServiceEndpointKey(provides));
    }

    public Document getEndpointDescription(final ServiceEndpointKey key) {
        final ProvidesData data = this.epProvides.get(key);
        return data != null ? data.description : null;
    }

    public ServiceEndpoint getEndpoint(final Provides provides) {
        return getEndpoint(new ServiceEndpointKey(provides));
    }

    public ServiceEndpoint getEndpoint(final ServiceEndpointKey key) {
        final ProvidesData data = this.epProvides.get(key);
        return data != null ? data.endpoint : null;
    }

    void addProvidesData(final Provides provides, final Document description, final List<QName> operations) {
        assert provides != null;
        // the CDK creates a default wsdl if there is none
        assert description != null;
        assert operations != null;
        final ServiceEndpointKey key = new ServiceEndpointKey(provides);
        this.epProvides.put(key, new ProvidesData(provides, description, operations));
    }

    public List<QName> getEndpointOperationsList(final ServiceEndpoint endpoint) {
        return getEndpointOperationsList(new ServiceEndpointKey(endpoint));
    }

    public List<QName> getEndpointOperationsList(final Provides provides) {
        return getEndpointOperationsList(new ServiceEndpointKey(provides));
    }

    public List<QName> getEndpointOperationsList(final ServiceEndpointKey key) {
        final ProvidesData data = this.epProvides.get(key);
        return data != null ? data.operations : null;
    }

    public Provides getProvides(final ServiceEndpoint endpoint) {
        return getProvides(new ServiceEndpointKey(endpoint));
    }

    public Provides getProvides(final ServiceEndpointKey key) {
        final ProvidesData data = this.epProvides.get(key);
        return data != null ? data.provides : null;
    }

    public boolean hasProvides(final ServiceEndpointKey key) {
        return this.epProvides.containsKey(key);
    }

    public boolean hasConsumes(final Consumes consumes) {
        for (final Consumes c : descriptor.getServices().getConsumes()) {
            if (c.equals(consumes)) {
                return true;
            }
        }
        return false;
    }

    public boolean hasProvides(final Provides provides) {
        for (final Provides p : descriptor.getServices().getProvides()) {
            if (p.equals(provides)) {
                return true;
            }
        }
        return false;
    }

    void addServiceEndpoint(final ServiceEndpoint endpoint) {
        assert endpoint != null;
        this.epProvides.get(new ServiceEndpointKey(endpoint)).endpoint = endpoint;
    }

    /**
     * The content of this {@link Collection} is only valid (and most certainly non-empty) AFTER the call to
     * {@link AbstractServiceUnitManager#doInit(ServiceUnitDataHandler)} has completed and ONLY BEFORE
     * {@link AbstractServiceUnitManager#doShutdown(ServiceUnitDataHandler)} is called.
     */
    public Collection<ServiceEndpoint> getServiceEndpoints() {
        final Collection<ServiceEndpoint> res = new ArrayList<>();
        for (final ProvidesData data : this.epProvides.values()) {
            if (data.endpoint != null) {
                res.add(data.endpoint);
            }
        }
        return Collections.unmodifiableCollection(res);
    }

    public Jbi getDescriptor() {
        return this.descriptor;
    }

    public String getInstallRoot() {
        return this.installRoot;
    }

    public String getName() {
        return this.name;
    }

    void removeEndpoint(final ServiceEndpoint endpoint) {
        this.epProvides.get(new ServiceEndpointKey(endpoint)).endpoint = null;
    }

    /**
     * Add an interceptor manager.
     */
    void addMessageInterceptorManager(final Object service, final MessageInterceptorManager manager) {
        assert manager != null;
        assert service instanceof Consumes || service instanceof Provides;
        this.messageInterceptorManagersMap.put(service, manager);
    }

    public MessageInterceptorManager getMessageInterceptorManager(final Object service) {
        assert service instanceof Consumes || service instanceof Provides;
        return this.messageInterceptorManagersMap.get(service);
    }

    /**
     * Return the map of Interceptor managers.
     * 
     * @return the messageInterceptorManager map
     */
    public Collection<MessageInterceptorManager> getMessageInterceptorManagers() {
        return this.messageInterceptorManagersMap.values();
    }

    /**
     * Return the extensions for the given Consumes service.
     */
    public SuConfigurationParameters getConfigurationExtensions(final Object service) {
        assert service instanceof Consumes || service instanceof Provides;
        return this.configurationExtensionsMap.get(service);
    }

    private static class ProvidesData {

        private final Provides provides;

        private final Document description;

        private final List<QName> operations;

        private ServiceEndpoint endpoint = null;

        public ProvidesData(Provides provides, Document description, List<QName> operations) {
            this.provides = provides;
            this.description = description;
            this.operations = operations;
        }
    }

    /**
     * @return one of the states from {@link DeploymentServiceMBean}.
     */
    public String getState() {
        return this.state;
    }

    /**
     * @param state
     *            one of the states from {@link DeploymentServiceMBean}.
     */
    void setState(final String state) {
        this.state = state;
    }
}
