/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import java.io.File;
import java.io.FileFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.List;

/**
 * This class is used to create classloader for JBI components.
 * @author Adrien Louis - EBM WebSourcing
 */
public final class ClassLoaderUtil {

    /**
     * As all methods of this class are static, no constructor is available.
     */
    private ClassLoaderUtil() {
        // NOP
    }

    /**
     * Returns an URL[] table, with the root path directory and 
     * all Jar contained in the given directory path.
     * 
     * @param directoryName
     *      directory containing the Jars
     * @return an URL[] table
     */
    public static URL[] getUrls(final String directoryName) {
        final File path = new File(directoryName);

        final FileFilter jarFileFilter = new FileFilter() {

            public boolean accept(final File pathname) {
                return pathname.getName().toLowerCase().endsWith(".jar");
            }
        };

        final File[] files = path.listFiles(jarFileFilter);

        final List<URL> urls = new ArrayList<URL>();

        if (files != null && files.length > 0) {
            for (final File file : files) {
                try {
                    urls.add(file.toURI().toURL());
                } catch (final MalformedURLException e) {
                    // nothing
                }
            }
        }
        
        // Add the directory path
        try {
            urls.add(path.toURI().toURL());
        } catch (MalformedURLException e) {
            // nothing
        }
        
        return urls.toArray(new URL[urls.size()]);
    }

    /**
     * Return an URLClassLoader object, with the root path directory and
     * all Jar contained in the given directory path.
     * 
     * @param rootPath
     *      directory containing the Jars
     * @param parent
     *      the parent classloader to use
     * @return a classLoader
     */
    public static URLClassLoader createClassLoader(final String rootPath, final ClassLoader parent) {
        return AccessController.doPrivileged(new PrivilegedAction<URLClassLoader>() {
            @Override
            public URLClassLoader run() {
                return new URLClassLoader(ClassLoaderUtil.getUrls(rootPath), parent);
            }
        });
    }
}
