/**
 * Copyright (c) 2007-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import java.io.File;
import java.io.FilenameFilter;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Utilitary class providing java.io FilenameFilters.
 * 
 * @author Adrien Louis - EBM WebSourcing
 */
public class FileNamePatternUtil {

    private FileNamePatternUtil() {
        // No constructor, it's an utility class
    }

    /**
     * Return a FilenameFilter with the given regex filter
     * 
     * @param regex
     * @return
     */
    public static RegexFileNameFilter buildFileNameFilterFromRegex(final String regex) {
        return new RegexFileNameFilter(regex);
    }

    /**
     * Return a FilenameFilter with the given wild card filter (using * and ?)
     * 
     * @param wildCard
     * @return
     */
    public static FilenameFilter buildFileNameFilterFromWildCard(final String wildCard) {
        return new RegexFileNameFilter(wildcardToRegex(wildCard));
    }

    /**
     * @return A {@link FilenameFilter} accepting all file names
     */
    public static FilenameFilter buildFileNameFilterFromWildCard() {
        return new RegexFileNameFilter();
    }

    /**
     * Return a FilenameFilter matching all filters
     * 
     * @param filters
     * @return
     */
    public static FilenameFilter buildFileNameFilterFromFilters(final List<FilenameFilter> filters) {
        return new RegexFileNameFilterList(filters);
    }

    /**
     * convert a regex expression to a standard file wild card filter
     * 
     * @param regex
     * @return
     */
    public static String regexToWildcard(final String regex) {
        String a = regex.replaceAll("\\\\.", "²");
        String b = a.replaceAll("\\.", "?");
        return b.replaceAll("²", ".");
    }

    /**
     * convert a standard file wild card filter to a regex expression
     * 
     * @param wildCard
     * @return
     */
    public static String wildcardToRegex(final String wildCard) {
        return wildCard.replaceAll("\\.", "\\\\.").replace("*", ".*").replaceAll("\\?", ".");
    }

    public static class RegexFileNameFilter implements FilenameFilter {

        private static final String ANY_VALUE = ".*";

        private final String regex;

        public RegexFileNameFilter() {
            this(ANY_VALUE);
        }

        public RegexFileNameFilter(final String regex) {
            this.regex = regex;
        }

        @Override
        public boolean accept(final File dir, final String name) {
            return Pattern.compile(regex, Pattern.CASE_INSENSITIVE).matcher(name).matches();
        }
    }

    public static class RegexFileNameFilterList implements FilenameFilter {

        private final List<FilenameFilter> list;

        public RegexFileNameFilterList(final List<FilenameFilter> list) {
            this.list = list;
        }

        @Override
        public boolean accept(final File dir, final String name) {
            boolean accept = false;
            final Iterator<FilenameFilter> iterator = this.list.iterator();
            while (iterator.hasNext() && !accept) {
                accept = iterator.next().accept(null, name);
            }
            return accept;
        }
    }

}
