/**
 * Copyright (c) 2009-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;

/**
 * 
 * Utility class for referring to a unique endpoint provided by a component.
 * 
 * It is meant to be used for existing service provided (with a {@link ServiceEndpoint}, a received {@link Exchange} or
 * a {@link Provides}) and not for potential called {@link Consumes} or sent {@link Exchange}.
 * 
 * It thus discriminates service based on the pair of service {@link QName} and endpoint name (because this pair is
 * registered in the ESB and is unique in it).
 * 
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Christophe DENEUX - Linagora
 * @author Victor Noël - Linagora
 */
public class ServiceEndpointKey {

    /**
     * Currently in Petals, endpoint names are unique across different service names, but actually, according to JBI,
     * nothing prevent two service names to have endpoints with the same name because normally the pair service/endpoint
     * is used to identify ONE endpoint. Thus we use it here to be coherent with JBI.
     */
    private final QName serviceName;

    private final String endpointName;

    public ServiceEndpointKey(final Provides provides) {
        this(provides.getServiceName(), provides.getEndpointName());
    }

    public ServiceEndpointKey(final ServiceEndpoint serviceEndpoint) {
        this(serviceEndpoint.getServiceName(), serviceEndpoint.getEndpointName());
    }

    public ServiceEndpointKey(final QName serviceName, final String endpointName) {
        this.serviceName = serviceName;
        this.endpointName = endpointName;
    }

    @Override
    public String toString() {
        return "ServiceEndpointKey [endpointName=" + this.endpointName + ", serviceName=" + this.serviceName + "]";
    }

    public String getEndpointName() {
        return this.endpointName;
    }

    public QName getServiceName() {
        return this.serviceName;
    }

    // Generated by Eclipse
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((endpointName == null) ? 0 : endpointName.hashCode());
        result = prime * result + ((serviceName == null) ? 0 : serviceName.hashCode());
        return result;
    }

    // Generated by Eclipse
    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        ServiceEndpointKey other = (ServiceEndpointKey) obj;
        if (endpointName == null) {
            if (other.endpointName != null)
                return false;
        } else if (!endpointName.equals(other.endpointName))
            return false;
        if (serviceName == null) {
            if (other.serviceName != null)
                return false;
        } else if (!serviceName.equals(other.serviceName))
            return false;
        return true;
    }

}
