/**
 * Copyright (c) 2015-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;

/**
 * 
 * Utility class like {@link ServiceEndpointKey} for referring to a unique operation provided by a component.
 * 
 * Not used internally in the CDK but for implementors.
 * 
 * It is meant to be used for existing service provided (with a {@link ServiceEndpoint}, a received {@link Exchange} or
 * a {@link Provides}) and not for potential called {@link Consumes} or sent {@link Exchange}.
 * 
 * It thus discriminates operations based on, on one hand, the operation {@link QName}, and on the other hand the pair
 * of service {@link QName} the endpoint name (this pair is registered in the ESB and is unique).
 * 
 * Note: this could be factorised by extending {@link ServiceEndpointKey}, but it is not to avoid the use of a
 * {@link ServiceEndpointOperationKey} in place of a {@link ServiceEndpointKey} by mistake.
 * 
 * @author Victor Noël - Linagora
 *
 */
public class ServiceEndpointOperationKey {

    private final QName serviceName;

    private final String endpointName;

    private final QName operation;

    public ServiceEndpointOperationKey(final Exchange exchange) {
        this(exchange.getEndpoint().getServiceName(), exchange.getEndpoint().getEndpointName(),
                exchange.getOperation());
    }

    public ServiceEndpointOperationKey(final Provides provides, final QName operation) {
        this(provides.getServiceName(), provides.getEndpointName(), operation);
    }

    public ServiceEndpointOperationKey(final QName serviceName, final String endpoint, final QName operation) {
        assert endpoint != null;
        assert operation != null;
        assert serviceName != null;
        this.serviceName = serviceName;
        this.endpointName = endpoint;
        this.operation = operation;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((endpointName == null) ? 0 : endpointName.hashCode());
        result = prime * result + ((serviceName == null) ? 0 : serviceName.hashCode());
        result = prime * result + ((operation == null) ? 0 : operation.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        ServiceEndpointOperationKey other = (ServiceEndpointOperationKey) obj;
        if (endpointName == null) {
            if (other.endpointName != null)
                return false;
        } else if (!endpointName.equals(other.endpointName))
            return false;
        if (serviceName == null) {
            if (other.serviceName != null)
                return false;
        } else if (!serviceName.equals(other.serviceName))
            return false;
        if (operation == null) {
            if (other.operation != null)
                return false;
        } else if (!operation.equals(other.operation))
            return false;
        return true;
    }

    @Override
    public String toString() {
        return "EndpointOperationKey [serviceName=" + serviceName + ", endpointName=" + endpointName + ", operation="
                + operation + "]";
    }

    public QName getServiceName() {
        return serviceName;
    }

    public String getEndpointName() {
        return endpointName;
    }

    public QName getOperation() {
        return operation;
    }
}
