/**
 * Copyright (c) 2015-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;

import org.ow2.easywsdl.wsdl.api.Binding;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;

import com.ebmwebsourcing.easycommons.lang.StringHelper;

/**
 * Various utilities primarily used by the {@link AbstractServiceUnitManager}.
 * 
 * @author vnoel
 *
 */
public class ServiceUnitUtil {

    private ServiceUnitUtil() {
        // not meant to be instantiated
    }

    /**
     * Either return a file in the SU (i.e., a relative path, not starting with /) or a file in the filesystem (i.e., an
     * absolute path, starting wih /)
     */
    public static File getFile(final String suRootPath, final String path) {
        if (path.startsWith("/")) {
            return new File(path);
        } else {
            return new File(suRootPath, path);
        }
    }

    public static final Description getWsdlDescription(final String suRootPath, final Provides provides)
            throws PEtALSCDKException {

        final URL location;
        if (StringHelper.isNullOrEmpty(provides.getWsdl())) {
            // if the WSDL field is null, generate a basic WSDL
            try {
                return WSDLUtilImpl.createLightWSDL20Description(provides.getInterfaceName(), provides.getServiceName(),
                        provides.getEndpointName());
            } catch (final WSDLException e) {
                throw new PEtALSCDKException("Cannot auto-generate a WSDL document", e);
            }
        } else {
            final File file = getFile(suRootPath, provides.getWsdl());

            if (!file.exists()) {
                throw new PEtALSCDKException("Can't retrieve file, it doesn't exist: " + file);
            }

            try {
                location = file.toURI().toURL();
            } catch (final MalformedURLException e) {
                throw new PEtALSCDKException("Can't retrieve file: " + file, e);
            }

            try {
                return WSDLUtilImpl.createWsdlDescription(location, false);
            } catch (final WSDLException e) {
                throw new PEtALSCDKException(e);
            }
        }

    }

    /**
     * Validate the WSDL description; Check that it match with the endpointName, serviceName, and interfaceName of the
     * {@code Provides} section.
     * 
     * @param wsdlDescription
     * @param provides
     * @throws JBIException
     */
    public static final void validateDescription(final Description wsdlDescription, final Provides provides)
            throws PEtALSCDKException {

        try {
            final Service ss = wsdlDescription.getService(provides.getServiceName());
            if (ss == null) {
                throw new PEtALSCDKException(
                        "Failed to find provided service '" + provides.getServiceName() + "' in the WSDL description");
            }

            final Endpoint e = ss.getEndpoint(provides.getEndpointName());
            if (e == null) {
                throw new PEtALSCDKException("Failed to find provided endpoint '" + provides.getEndpointName()
                        + "' in the WSDL description");
            }

            // in WSDL2, the service has an interface defined, but in WSDL1, we are not totally assured that
            // getting interface on the service will succeed, so it is safer to check directly the binding. We
            // fallback on the service's interface else.
            final InterfaceType interfaceType;
            final Binding b = e.getBinding();
            if (b != null && b.getInterface() != null) {
                interfaceType = b.getInterface();
            } else {
                interfaceType = ss.getInterface();
            }

            if (interfaceType != null) {
                if (!provides.getInterfaceName().equals(interfaceType.getQName())) {
                    throw new PEtALSCDKException("Failed to match provided interface '" + provides.getInterfaceName()
                            + "' in the WSDL description. Found '" + interfaceType.getQName() + "' instead");
                }
            } else {
                throw new PEtALSCDKException(
                        "Failed to find an interface in WSDL description bound to the provided service '"
                                + ss.getQName() + "'");
            }
        } catch (final WSDLException e) {
            throw new PEtALSCDKException(
                    "Failed to parse the WSDL description bound to the provided endpoint/service/interface '"
                            + provides.getEndpointName() + "/" + provides.getServiceName() + "/"
                            + provides.getInterfaceName() + "'",
                    e);
        }
    }

    /**
     * Retrieve the MEP from the consumer node extensions.
     * 
     * @param consumes
     *            The consumer node of the SU JBI descriptor.
     * @return MEP specified in extensions, or <code>null</code> if no MEP pattern specified.
     * @throws MessagingException
     *             MEP pattern specified in consumes extension is not valid.
     */
    public static MEPPatternConstants retrievePattern(final Consumes consumes) throws MessagingException {
        if (consumes.getMep() != null) {
            switch (consumes.getMep()) {
                case IN_ONLY:
                    return MEPPatternConstants.IN_ONLY;
                case ROBUST_IN_ONLY:
                    return MEPPatternConstants.ROBUST_IN_ONLY;
                case IN_OUT:
                    return MEPPatternConstants.IN_OUT;
                case IN_OPTIONAL_OUT:
                    return MEPPatternConstants.IN_OPTIONAL_OUT;
                default:
                    throw new MessagingException(
                            "MEP pattern '" + consumes.getMep().value() + "' is not a valid JBI pattern");
            }
        }
        return null;
    }
}
