/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;

import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;

import org.ow2.petals.commons.xml.BytesSource;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.ebmwebsourcing.easycommons.pooling.PoolException;
import com.ebmwebsourcing.easycommons.xml.DOMHelper;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.SourceHelper;

/**
 * This class provides mapping between {@link Source} and some types such as {@link String} or {@link Document}.
 * 
 * It relies on the default implementation used by the kernel for {@link Source}: {@link BytesSource} and
 * {@link DOMSource}.
 * 
 * It differs from {@link SourceHelper} by the fact that it is made for the way petals work while {@link SourceHelper}
 * is more generic and low-level. Also, exceptions are wrapped in {@link PEtALSCDKException} for easier use.
 * 
 * @author Adrien Louis - EBM WebSourcing
 * @author Frederic Gardes - EBM WebSourcing
 * @author Nicolas Oddoux - EBM WebSourcing
 * @author Victor Noël - Linagora
 */
public class SourceUtil {

    /**
     * Creates a new instance of {@link SourceUtil}
     * 
     */
    private SourceUtil() {
        super();
    }

    /**
     * Create a {@link Document} from a {@link Source}.
     * 
     * If the Source is a {@link DOMSource} and contains a {@link Document}, the returned document is a clone of this
     * document.
     * 
     * @param source
     *            the {@link Source}
     * @return the DOM document built from the Source
     * @throws PEtALSCDKException
     */
    public static final Document createDocument(final Source source) throws PEtALSCDKException {
        return createDocument(source, true);
    }

    /**
     * Create a {@link Document} from a {@link Source}. <br>
     * If the Source is a {@link DOMSource} and contains a {@link Document} and {@code cloneDOM} is set to {@code true},
     * the returned document is a clone of this document. Otherwise the original {@link Document} is returned.
     * 
     * @param source
     *            the {@link Source}
     * @param cloneDOM
     *            {@code true} to clone the {@link Document} if it is a {@link DOMSource}.
     * @return a DOM document
     * @throws PEtALSCDKException
     */
    public static final Document createDocument(final Source source, final boolean cloneDOM)
            throws PEtALSCDKException {
        try {
            if (source instanceof DOMSource) {
                final Node node = ((DOMSource) source).getNode();
                return DOMHelper.toDocument(cloneDOM ? cloneToDocument(node) : node);
            } else {
                return SourceHelper.toDocument(source);
            }
        } catch (final TransformerException e) {
            throw new PEtALSCDKException(e);
        } catch (final PoolException e) {
            throw new PEtALSCDKException(e);
        }
    }

    private static Document cloneToDocument(final Node node) {
        final Document document = DocumentBuilders.newDocument();
        if (node instanceof Document) {
            document.appendChild(document.importNode(((Document) node).getDocumentElement(), true));
        } else {
            document.appendChild(document.importNode(node, true));
        }
        return document;
    }

    /**
     * Create a DOM Source from the DOM document.
     * 
     * @param document
     *            A DOM document
     * @return the DOM source built from the DOM document
     */
    public static final DOMSource createSource(final Document document) {
        return SourceHelper.toDOMSource(document);
    }

    /**
     * Create a Source from the String. The message must be a valid XML String.
     * If an encoding is specified in the XML String header it is used to create
     * the XML bytes stream, otherwise the default JVM encoding is used.
     * 
     * @param message
     *            a valid XML String
     * @return the Source built from the XML String
     * @throws PEtALSCDKException
     *             if an error occurs during Source creation
     * 
     */
    public static final Source createSource(final String message) {
        return createSource(message, Charset.defaultCharset());
    }

    /**
     * Create a Source from the String with a specified encoding format. The message must be a valid XML String. The
     * encoding must match the XML declaration in the message.
     * 
     * @param message
     *            a valid XML String
     * @param charFormat
     *            the encoding format, must be non {@code null}
     * 
     * @return the Source built from the XML String
     * @throws PEtALSCDKException
     *             if an error occurs because of an invalid charset
     */
    public static final Source createSource(final String message, final String charFormat) throws PEtALSCDKException {
        try {
            return new BytesSource(message.getBytes(charFormat));
        } catch (UnsupportedEncodingException e) {
            throw new PEtALSCDKException(e);
        }
    }

    /**
     * Create a Source from the String with a specified encoding format. The message must be a valid XML String. The
     * encoding must match the XML declaration in the message.
     * 
     * @param message
     *            a valid XML String
     * @param charFormat
     *            the encoding format, must be non {@code null}
     * 
     * @return the Source built from the XML String
     */
    public static final Source createSource(final String message, final Charset charset) {
        return new BytesSource(message.getBytes(charset));
    }

    /**
     * Create a {@link String} from a {@link Source}.
     * 
     * @param source
     *            the {@link Source}
     * @return the String built from the Source
     * @throws PEtALSCDKException
     */
    public static final String createString(final Source source) throws PEtALSCDKException {
        try {
            return SourceHelper.toString(source);
        } catch (TransformerException e) {
            throw new PEtALSCDKException(e);
        }
    }

}
