/**
 * Copyright (c) 2005-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.extensions.wsdl4complexwsdl.WSDL4ComplexWsdlFactory;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.WSDL4ComplexWsdlReaders;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.WSDL4ComplexWsdlWriters;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlReader;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlWriter;
import org.ow2.easywsdl.wsdl.WSDLFactory;
import org.ow2.easywsdl.wsdl.api.Binding;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Operation;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfDescription.WSDLVersionConstants;
import org.ow2.petals.component.framework.api.Constants;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.xml.DOMHelper;

/**
 * Util class to manipulate WSDL Document or Definition.
 * @author wjoseph - EBM WebSourcing
 */
public final class WSDLUtilImpl {

	/**
	 * Creates a new instance of {@link WSDLUtilImpl}
	 * 
	 * @throws WSDLException
	 * 
	 */
	public WSDLUtilImpl() throws WSDLException {
	    // nothing
	}

	/**
	 * Create a light WSDL2.0 description from the given ServiceEndpoint. All
	 * attributes are required into the function parameter, otherwise a
	 * NullPointerException will be thrown. Namespaces are managed.
	 * 
	 * @param interfaceQName
	 *            must be non null
	 * @param serviceQName
	 *            must be non null
	 * @param endpointName
	 *            must be non null
	 * @return a WSDL2.0 light description
	 * @throws WSDLException
	 * @see 
	 *      org.ow2.petals.component.framework.util#createLightWSDL20Description(
	 *      final QName[], final QName, final String)
	 */
	public static final Description createLightWSDL20Description(
			final QName interfaceQName, final QName serviceQName,
			final String endpointName) throws WSDLException {

		if (!interfaceQName.getNamespaceURI().equals(
				serviceQName.getNamespaceURI())) {
			throw new WSDLException("The interface namespace '"
					+ interfaceQName.getNamespaceURI()
					+ "' must be the same than the service namespace '"
					+ serviceQName.getNamespaceURI()
					+ "' to build a light WSDL");
		}

        // TODO we don't really need a complex wsdl if we don't have any imported documents...
		final Description wsdlDescription = WSDL4ComplexWsdlFactory
				.newInstance().addComplexWsdlElmt2Description(
						WSDLFactory.newInstance().newDescription(
								WSDLVersionConstants.WSDL20));
		wsdlDescription.setTargetNamespace(interfaceQName.getNamespaceURI());
		wsdlDescription.setQName(new QName("autoGeneratedDescription"));

		final InterfaceType interfazz = wsdlDescription.createInterface();
		interfazz.setQName(interfaceQName);
		wsdlDescription.addInterface(interfazz);

		final Binding binding = wsdlDescription.createBinding();
		binding.setInterface(interfazz);
		binding.setQName(new QName("autoGeneratedBinding"));
		wsdlDescription.addBinding(binding);

		final Service service = wsdlDescription.createService();
		service.setQName(serviceQName);
		service.setInterface(interfazz);

		final Endpoint endpoint = service.createEndpoint();
		endpoint.setName(endpointName);
		endpoint.setBinding(binding);

		service.addEndpoint(endpoint);
		wsdlDescription.addService(service);

		return wsdlDescription;
	}

	/**
	 * Get the endpoints into a wsdl description. Retrieve all of them if no
	 * filter name is provided.<br />
	 * Retrieve only the endpoints matching the service QName filter if
	 * provided.<br />
	 * Retrieve only the endpoints matching the endpoint name filter if
	 * provided.
	 * 
	 * @param wsdlDescription
	 *            the wsdl description generated by easy wsdl
	 * @param serviceQNameFilter
	 *            if not null, only the endpoints with this service QName will
	 *            be return
	 * @param endpointNameFilter
	 *            if not null, only the endpoints with this name will be return
	 * @return The list of endpoints, empty if nothing found
	 */
	public static final List<Endpoint> getEndpointList(
			final Description wsdlDescription, final QName serviceQNameFilter,
			final String endpointNameFilter) {
		final List<Endpoint> endpointList = new ArrayList<Endpoint>();
		Endpoint endpoint = null;
		if (wsdlDescription != null) {
			if (serviceQNameFilter == null) {
				final List<Service> serviceList = wsdlDescription.getServices();
				for (final Service service : serviceList) {
					if (service != null) {
						if (endpointNameFilter == null) {
							endpointList.addAll(service.getEndpoints());
						} else {
							endpoint = service.getEndpoint(endpointNameFilter);
							if (endpoint != null) {
								endpointList.add(service
										.getEndpoint(endpointNameFilter));
							}
						}
					}
				}
			} else {
				final Service service = wsdlDescription
						.getService(serviceQNameFilter);
				if (service != null) {
					if (endpointNameFilter == null) {
						endpointList.addAll(service.getEndpoints());
					} else {
						endpoint = service.getEndpoint(endpointNameFilter);
						if (endpoint != null) {
							endpointList.add(service
									.getEndpoint(endpointNameFilter));
						}
					}
				}
			}
		}
		return endpointList;
	}

	/**
	 * Get the endpoints into a wsdl description.
	 * 
	 * @param wsdlDescription
	 *            the wsdl description generated by easy wsdl
	 * @return The list of Endpoint, or null if the description is null
	 */
	public static final List<Endpoint> getEndpointList(
			final Description wsdlDescription) {
		return getEndpointList(wsdlDescription, null, null);
	}

	/**
	 * Create a wsdl description. <br/>
	 * NOTE : Imported document are written in the description.
	 * 
	 * @param wsdlLocation
	 *            the location of the wsdl file
	 * @return the description of the document
	 * @throws WSDLException
	 */
	public static final Description createWsdlDescription(final URL wsdlLocation)
			throws WSDLException {
		return createWsdlDescription(wsdlLocation, false, true);
	}

	/**
	 * Create a wsdl description. <br/>
	 * NOTE : Imported document are written in the description.
	 * 
	 * @param wsdlLocation
	 *            the location of the wsdl file
	 * @param processAutogenerateEndpoint
	 *            true if you want to replace the endpoints named
	 *            {@link Constants.Component#AUTOGENERATED_ENDPOINT_NAME}<br />
	 *            by a name generated by PEtALS
	 * @return the description of the document
	 * @throws WSDLException
	 */
	public static final Description createWsdlDescription(
			final URL wsdlLocation, final boolean processAutogenerateEndpoint)
			throws WSDLException {
		return createWsdlDescription(wsdlLocation, processAutogenerateEndpoint,
				true);
	}

	/**
	 * Create a wsdl description.
	 * 
	 * @param wsdlLocation
	 *            the location of the wsdl file
	 * @param processAutogenerateEndpoint
	 *            true if you want to replace the endpoints named
	 *            {@link Constants.Component#AUTOGENERATED_ENDPOINT_NAME}<br />
	 *            by a name generated by PEtALS
	 * @param addImportedDocumentsInWsdl
	 *            if true the imported document are written in the description.
	 * @return the description of the document
	 * @throws WSDLException
	 */
    public static final org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description createWsdlDescription(
			final URL wsdlLocation, final boolean processAutogenerateEndpoint,
			final boolean addImportedDocumentsInWsdl) throws WSDLException {
		String autogeneratedEndpointName = null;
		if (wsdlLocation == null) {
			return null;
		}
		try {
			WSDL4ComplexWsdlReader wsdl4ComplexWsdlReader = null;
			try {
				wsdl4ComplexWsdlReader = WSDL4ComplexWsdlReaders
						.takeWSDL4ComplexWsdlReader();
                org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description description = wsdl4ComplexWsdlReader
						.read(wsdlLocation);

				if (addImportedDocumentsInWsdl) {
					description.addImportedDocumentsInWsdl();
				}

				if (processAutogenerateEndpoint) {
					List<Endpoint> endpointList = getEndpointList(description,
							null,
							Constants.Component.AUTOGENERATED_ENDPOINT_NAME);
					for (Endpoint endpoint : endpointList) {
						autogeneratedEndpointName = EndpointUtil
								.generateEndpointName();
						endpoint.setName(autogeneratedEndpointName);
					}
				}

				return description;
			} finally {
				if (wsdl4ComplexWsdlReader != null) {
					WSDL4ComplexWsdlReaders
							.releaseWSDL4ComplexWsdlReader(wsdl4ComplexWsdlReader);
				}
			}
		} catch (final URISyntaxException e) {
			throw new WSDLException("Failed to parse the WSDL", e);
		} catch (final IOException e) {
			throw new WSDLException("Failed to read the WSDL", e);
		}
	}

    /**
     * Convert the description to a self-contained document (note: if it contains imported documents, it MUST be read
     * with a {@link WSDL4ComplexWsdlReader}!)
     * 
     * @see #convertDescriptionToDocument(Description, boolean)
     * @param wsdlDescription
     *            the description
     * @return a self-contained document
     * @throws WSDLException
     */
    public static final Document convertDescriptionToDocument(final Description wsdlDescription) throws WSDLException {
        return convertDescriptionToDocument(wsdlDescription, true);
    }

	/**
     * Convert the description to a document
     * 
     * @param wsdlDescription
     *            the description
     * @param selfContained
     *            if the description is of type {@link org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description} and
     *            there was external documents imported, then if <code>true</code> the written file will include them
     *            instead of imports (note: if it contains imported documents, it MUST be read with a
     *            {@link WSDL4ComplexWsdlReader}!)
     * @return the document
     * @throws WSDLException
     */
    public static final Document convertDescriptionToDocument(final Description wsdlDescription,
            final boolean selfContained) throws WSDLException {

        if (!selfContained
                && wsdlDescription instanceof org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description) {
            ((org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description) wsdlDescription)
                    .deleteImportedDocumentsInWsdl();
        }

        final Document document;
		WSDL4ComplexWsdlWriter wsdl4ComplexWsdlWriter = null;
		try {
			wsdl4ComplexWsdlWriter = WSDL4ComplexWsdlWriters
					.takeWSDL4ComplexWsdlWriter();
            document = wsdl4ComplexWsdlWriter.getDocument(wsdlDescription);
		} finally {
			if (wsdl4ComplexWsdlWriter != null) {
				WSDL4ComplexWsdlWriters
						.releaseWSDL4ComplexWsdlWriter(wsdl4ComplexWsdlWriter);
			}
		}

        if (!selfContained
                && wsdlDescription instanceof org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description) {
            ((org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description) wsdlDescription)
                    .addImportedDocumentsInWsdl();
        }

        return document;
	}

	/**
	 * Check if a description contains a specific endpoint
	 * 
	 * @param wsdlDescription
	 *            the description
	 * @param endpoint
	 *            the endpoint. must be non null, with a service non null and a
	 *            name non null
	 * @return true if the description contains the endpoint, false else
	 */
	public static final boolean isDescriptionContaining(
			final Description wsdlDescription, final ServiceEndpoint endpoint) {
		if (endpoint == null || endpoint.getServiceName() == null
				|| endpoint.getEndpointName() == null) {
			return false;
		}
		return !getEndpointList(wsdlDescription, endpoint.getServiceName(),
				endpoint.getEndpointName()).isEmpty();
	}

	/**
     * Write a description into a file.
     * 
     * @param description
     *            the wsdl description
     * @param wsdlFile
     *            The location of the destination wsdl file.
     * @param selfContained
     *            see {@link #convertDescriptionToDocument(Description, boolean)}
     * @throws WSDLException
     */
	public static final void convertDescriptionToFile(
            final Description description, final File wsdlFile, final boolean selfContained)
			throws WSDLException {

        Document wsdlDocument = convertDescriptionToDocument(description, selfContained);

		try {
			DOMHelper.prettyPrint(wsdlDocument, new FileOutputStream(wsdlFile));
		} catch (FileNotFoundException e) {
			throw new WSDLException("Failed to write the WSDL", e);
		}
	}

	/**
	 * Retrun operations defined in interface implemented by the specified
	 * endpoint. If the wsdl is a fake wsdl (genrated wsdl with no binding or
	 * porttype operation), it returns an empty list
	 * 
	 * @param wsdlDescription
	 *            the endpoint description; must be not null
	 * @param endpointName
	 *            the endpoint name. must be non null and non empty
	 * @param serviceName
	 *            the service name. must be non null
	 * @return operations defined in interface implemented by the specified
	 *         endpoint
	 */
	public static final List<QName> getEndpointOperations(
			final Description wsdlDescription, final QName serviceName,
			final String endpointName) {
		assert wsdlDescription != null;
		assert serviceName != null;
		assert endpointName != null && !endpointName.isEmpty();
		List<QName> operations = new ArrayList<QName>();
		List<Endpoint> endpoints = getEndpointList(wsdlDescription,
				serviceName, endpointName);
		assert endpoints != null && endpoints.size() == 1;
		Binding binding = endpoints.get(0).getBinding();
		if (binding != null) {
			InterfaceType interfaceType = binding.getInterface();
			if (interfaceType != null) {
				if (interfaceType.getOperations() != null) {
					for (Operation operation : interfaceType.getOperations()) {
						operations.add(operation.getQName());
					}
				}
			}
		}
		return operations;
	}
}
