/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.lang.ref.PhantomReference;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.util.ArrayList;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.easymock.classextension.EasyMock;
import org.junit.Assert;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.jbidescriptor.generated.Component;
import org.ow2.petals.component.framework.jbidescriptor.generated.Identification;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivestrictint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Runtimepositivestrictlong;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositivestrictint;
import org.ow2.petals.component.framework.jbidescriptor.generated.Settablepositivestrictlong;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.monitoring.Monitoring;
import org.ow2.petals.component.framework.monitoring.key.ServiceProviderKey;
import org.ow2.petals.component.framework.monitoring.probe.MessageExchangeAcceptorPoolProbe;
import org.ow2.petals.probes.api.exceptions.MultipleProbesFactoriesFoundException;
import org.ow2.petals.probes.api.exceptions.NoProbesFactoryFoundException;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.DurationProbe;
import org.ow2.petals.probes.api.probes.KeyedCounterProbeWithExecutionStatus;
import org.ow2.petals.probes.api.probes.KeyedResponseTimeProbe;
import org.ow2.petals.probes.api.probes.macro.ThreadPoolProbe;

public abstract class AbstractJBIManager<T> extends Assert {

    private static final String COMPONENT_NAME = "my-component";

    protected final ReferenceQueue<T> queueGC = new ReferenceQueue<T>();

    protected void assertThreadIsPresent(final String suffix) throws Exception {
        final String threadName = COMPONENT_NAME + suffix;

        final Thread[] threads = new Thread[50];
        Thread.currentThread().getThreadGroup().enumerate(threads);

        for (final Thread thread : threads) {
            if (thread != null && thread.getName().startsWith(threadName)) {
                return;
            }
        }

        fail("The Thread '" + threadName + "' is not present.");
    }

    protected void assertThreadIsFreed(final String suffix) throws Exception {

        try {
            // Wait thread context switches
            Thread.sleep(100);
        } catch (InterruptedException e) {
            // NOP
        }
        System.gc();

        final String threadName = COMPONENT_NAME + suffix;

        final Thread[] threads = new Thread[50];
        Thread.currentThread().getThreadGroup().enumerate(threads);

        for (final Thread thread : threads) {
            if (thread != null && thread.getName().startsWith(threadName)) {
                fail("The Thread '" + threadName + "' is not freed.");
            }
        }

    }

    protected void assertThreadGroupIsPresent(final String suffix) throws Exception {

        final String groupName = COMPONENT_NAME + suffix;

        final ThreadGroup[] childrenThreadGroup = new ThreadGroup[50];
        Thread.currentThread().getThreadGroup().enumerate(childrenThreadGroup);
        for (final ThreadGroup childThreadGroup : childrenThreadGroup) {
            if (childThreadGroup != null && childThreadGroup.getName().startsWith(groupName)) {
                return;
            }
        }

        fail("The Thread group '" + groupName + "' is not present.");
    }

    protected void assertThreadGroupIsFreed(final String suffix) throws Exception {

        try {
            // Wait thread context switches
            Thread.sleep(100);
        } catch (InterruptedException e) {
            // NOP
        }
        System.gc();

        final String groupName = COMPONENT_NAME + suffix;

        final ThreadGroup[] childrenThreadGroup = new ThreadGroup[50];
        Thread.currentThread().getThreadGroup().enumerate(childrenThreadGroup);
        for (final ThreadGroup childThreadGroup : childrenThreadGroup) {
            if (childThreadGroup != null && childThreadGroup.getName().startsWith(groupName)) {
                fail("The Thread group '" + groupName + "' is not freed.");
            }
        }
    }

    protected void assertIsGarbaged(final PhantomReference<T> refMgr) {
        try {
            boolean found = false;
            Reference<?> phantRef;
            do {
                System.gc();
                phantRef = this.queueGC.remove(1000);
                if (phantRef == refMgr) {
                    found = true;
                }
            } while (!found && phantRef != null);
            if (!found) {
                fail("JBI...Manager is not correctly garbaged");
            }
        } catch (final InterruptedException e) {
            fail("JBI...Manager is not correctly garbaged !!!!");
        }
    }

    protected AbstractComponent createComponentMock(
            final int _acceptorsCoreSize, final long _acceptorsStopMaxWait,
            final int _acceptorsRetryNb, final long _acceptorsRetryWait, 
            final int _processorCoreSize, final int _processorMaxSize,
            final int _processorKeepAliveTime, final long _processorStopMaxWait,
            final long _timeBetweenAsyncCleanerRuns) throws MessagingException,
            MultipleProbesFactoriesFoundException, NoProbesFactoryFoundException {

        final Runtimepositivestrictint acceptorPoolSize = new Runtimepositivestrictint();
            acceptorPoolSize.setValue(_acceptorsCoreSize);

        final Runtimepositivestrictlong acceptorsStopMaxWait = new Runtimepositivestrictlong();
        acceptorsStopMaxWait.setValue(_acceptorsStopMaxWait);

        final Settablepositivestrictint acceptorsRetryNb = new Settablepositivestrictint();
        acceptorsRetryNb.setValue(_acceptorsRetryNb);

        final Settablepositivestrictlong acceptorsRetryWait = new Settablepositivestrictlong();
        acceptorsRetryWait.setValue(_acceptorsRetryWait);

        final Runtimepositivestrictint processorPoolSize = new Runtimepositivestrictint();
        processorPoolSize.setValue(_processorCoreSize);

        final Runtimepositivestrictint processorMaxPoolSize = new Runtimepositivestrictint();
        processorMaxPoolSize.setValue(_processorMaxSize);

        final Runtimepositivestrictint processorKeepAliveTime = new Runtimepositivestrictint();
        processorKeepAliveTime.setValue(_processorKeepAliveTime);

        final Runtimepositivestrictlong processorStopMaxWait = new Runtimepositivestrictlong();
        processorStopMaxWait.setValue(_processorStopMaxWait);

        final Settablepositivestrictlong timeBetweenAsyncCleanerRuns = new Settablepositivestrictlong();
        timeBetweenAsyncCleanerRuns.setValue(_timeBetweenAsyncCleanerRuns);

        final Identification identification = new Identification();
        identification.setName(COMPONENT_NAME);

        final AbstractJBIListener jbiListener = EasyMock.createNiceMock(AbstractJBIListener.class);
        EasyMock.makeThreadSafe(jbiListener, true);
        jbiListener.init();
        EasyMock.expectLastCall();
        EasyMock.replay(jbiListener);

        final Jbi jbiCfg = EasyMock.createMock(Jbi.class);
        EasyMock.makeThreadSafe(jbiCfg, true);
        final Component componentCfg = EasyMock.createMock(Component.class);
        EasyMock.expect(componentCfg.getIdentification()).andReturn(identification).anyTimes();
        EasyMock.expect(componentCfg.getAcceptorPoolSize()).andReturn(acceptorPoolSize).anyTimes();
        EasyMock.expect(componentCfg.getAcceptorRetryNumber()).andReturn(acceptorsRetryNb)
                .anyTimes();
        EasyMock.expect(componentCfg.getAcceptorRetryWait()).andReturn(acceptorsRetryWait)
                .anyTimes();
        EasyMock.expect(componentCfg.getAcceptorStopMaxWait()).andReturn(acceptorsStopMaxWait)
                .anyTimes();
        EasyMock.expect(componentCfg.getProcessorPoolSize()).andReturn(processorPoolSize)
                .anyTimes();
        EasyMock.expect(componentCfg.getProcessorMaxPoolSize()).andReturn(processorMaxPoolSize)
                .anyTimes();
        EasyMock.expect(componentCfg.getProcessorKeepAliveTime()).andReturn(processorKeepAliveTime)
                .anyTimes();
        EasyMock.expect(componentCfg.getProcessorStopMaxWait()).andReturn(processorStopMaxWait)
                .anyTimes();
        EasyMock.expect(componentCfg.getTimeBeetweenAsyncCleanerRuns()).andReturn(timeBetweenAsyncCleanerRuns)
                .anyTimes();
        EasyMock.expect(componentCfg.getJbiListenerClassName())
                .andReturn(jbiListener.getClass().getName()).anyTimes();
        EasyMock.expect(jbiCfg.getComponent()).andReturn(componentCfg).anyTimes();
        EasyMock.replay(jbiCfg, componentCfg);

        // we can't use easymock for that because the concurrent calls to accept() does not work well
        final DeliveryChannel deliveryChannel = new DeliveryChannel() {

            @Override
            public boolean sendSync(MessageExchange exchange, long timeoutMS) throws MessagingException {
                return false;
            }

            @Override
            public boolean sendSync(MessageExchange exchange) throws MessagingException {
                return false;
            }

            @Override
            public void send(MessageExchange exchange) throws MessagingException {
            }

            @Override
            public MessageExchangeFactory createExchangeFactoryForService(QName serviceName) {
                return null;
            }

            @Override
            public MessageExchangeFactory createExchangeFactory(ServiceEndpoint endpoint) {
                return null;
            }

            @Override
            public MessageExchangeFactory createExchangeFactory(QName interfaceName) {
                return null;
            }

            @Override
            public MessageExchangeFactory createExchangeFactory() {
                return null;
            }

            @Override
            public void close() throws MessagingException {

            }

            @Override
            public MessageExchange accept(long timeoutMS) throws MessagingException, IllegalArgumentException {
                try {
                    synchronized (this) {
                        // this is meant to be interrupted on stop for the tests and that's all
                        this.wait(timeoutMS);
                    }
                } catch (final InterruptedException e) {
                    // that's how it is supposed to happen according to the JBI spec
                    throw new MessagingException("accept() interrupted", e);
                }
                return null;
            }

            @Override
            public MessageExchange accept() throws MessagingException {
                return accept(0);
            }
        };

        final ComponentContext context = EasyMock.createMock(ComponentContext.class);
        EasyMock.makeThreadSafe(context, true);
        EasyMock.expect(context.getComponentName()).andReturn(COMPONENT_NAME).anyTimes();
        // No transaction manager is needed by the test
        EasyMock.expect(context.getTransactionManager()).andReturn(null).anyTimes();
        EasyMock.replay(context);

        final Monitoring monitoringMock = EasyMock.createNiceMock(Monitoring.class);
        EasyMock.makeThreadSafe(monitoringMock, true);
        this.addProbeMocks(monitoringMock);
        EasyMock.replay(monitoringMock);

        final AbstractComponent componentMock = EasyMock.createMock(AbstractComponent.class);
        EasyMock.makeThreadSafe(componentMock, true);
        EasyMock.expect(componentMock.getJbiComponentDescriptor()).andReturn(jbiCfg)
                .anyTimes();
        EasyMock.expect(componentMock.getChannel()).andReturn(deliveryChannel).anyTimes();
        EasyMock.expect(componentMock.getMonitoringBean()).andReturn(monitoringMock).anyTimes();
        // No service unit manager is needed by the test
        EasyMock.expect(componentMock.getServiceUnitManager()).andReturn(null).anyTimes();
        // No async manager is needed by the test
        EasyMock.expect(componentMock.getAsyncManager()).andReturn(null).anyTimes();
        EasyMock.expect(componentMock.getContext()).andReturn(context).anyTimes();
        EasyMock.expect(componentMock.getMBeanAttributesNames()).andReturn(new ArrayList<String>()).anyTimes();
        EasyMock.replay(componentMock);

        return componentMock;
    }

    private void addProbeMocks(final Monitoring monitoringMock) {
        final MessageExchangeAcceptorPoolProbe messageExchangeAcceptorPoolProbeMock = EasyMock
                .createMock(MessageExchangeAcceptorPoolProbe.class);
        EasyMock.makeThreadSafe(messageExchangeAcceptorPoolProbeMock, true);
        EasyMock.expect(monitoringMock.getMessageExchangeAcceptorPoolProbe())
                .andReturn(messageExchangeAcceptorPoolProbeMock).anyTimes();

        final ThreadPoolProbe threadPoolProbeMock = EasyMock.createMock(ThreadPoolProbe.class);
        EasyMock.makeThreadSafe(threadPoolProbeMock, true);
        EasyMock.expect(monitoringMock.getMessageExchangeProcessorThreadPoolProbe())
                .andReturn(threadPoolProbeMock).anyTimes();

        final DurationProbe mexProcessorDurationProbeMock = EasyMock
                .createMock(DurationProbe.class);
        EasyMock.makeThreadSafe(mexProcessorDurationProbeMock, true);
        EasyMock.expect(monitoringMock.getMessageExchangeProcessorDurationProbe())
                .andReturn(mexProcessorDurationProbeMock).anyTimes();

        final KeyedResponseTimeProbe<ServiceProviderKey> serviceProviderResponseTimesProbeMock = EasyMock
                .createMock(KeyedResponseTimeProbe.class);
        EasyMock.makeThreadSafe(serviceProviderResponseTimesProbeMock, true);
        EasyMock.expect(monitoringMock.getServiceProviderResponseTimesProbe())
                .andReturn(serviceProviderResponseTimesProbeMock).anyTimes();

        final KeyedCounterProbeWithExecutionStatus<ProbeKey> serviceProviderInvocationsProbeMock = EasyMock
                .createMock(KeyedCounterProbeWithExecutionStatus.class);
        EasyMock.makeThreadSafe(serviceProviderInvocationsProbeMock, true);
        EasyMock.expect(monitoringMock.getServiceProviderInvocationsProbe())
                .andReturn(serviceProviderInvocationsProbeMock).anyTimes();
    }


}
