/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.lang.ref.PhantomReference;
import java.util.logging.Logger;

import org.junit.Test;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;

import junit.framework.TestCase;

public class JBIAcceptorManagerTest extends AbstractJBIManager<JBIAcceptorManager> {
    
    private final static Logger LOG = Logger.getLogger(JBIAcceptorManagerTest.class.getName());

    private void assertThreadGroupIsPresent() throws Exception {
        super.assertThreadGroupIsPresent(JBIAcceptorManager.GROUP_SUFFIX);
    }

    private void assertAcceptorThreadGroupIsFreed() throws Exception {
        super.assertThreadGroupIsFreed(JBIAcceptorManager.GROUP_SUFFIX);
    }

    /**
     * Checks the JBIAcceptorManager is correctly garbaged.
     */
    @Test
    public void testGarbaged_001() throws Exception {

        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, 10, 10, 100, 100, 300);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(componentMock, null, null), LOG);

        acceptorMgr.start();

        assertThreadGroupIsPresent();

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
        acceptorMgr.stop();

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }

    /**
     * Checks the start against default values.
     * Expected result: all parameter values are defined with their default values.
     */
    @Test
    public void testStart_000() throws Exception {
        
        final int acceptorPoolSize = 10;
        final long acceptorStopMaxWait = 100;
        final AbstractComponent componentMock = this.createComponentMock(acceptorPoolSize, acceptorStopMaxWait, 40, 250,
                10, 10, 100, 100, 300);
    
        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(componentMock, null, null), LOG);
    
        acceptorMgr.start();

        assertThreadGroupIsPresent();
        
        TestCase.assertEquals("Acceptor pool size is not the expected one.", acceptorPoolSize,
                acceptorMgr.getPoolSize());
        TestCase.assertEquals("Acceptor stop max wait is not the expected one.", acceptorStopMaxWait,
                acceptorMgr.getStopMaxWait());
        
        acceptorMgr.stop();
    
        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);
    
        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();
    
    }

    /**
     * Checks the stop against a value of 'acceptor-stop-max-wait' not set.
     * Expected result: no error thrown and the duration of stop of the
     * JBIAcceptorManager is lower than the default value (500ms).
     */
    @Test
    public void testStop_001() throws Exception {

        final int acceptorCoreSize = 10;
        final long acceptorStopMaxWait = 100;
        final AbstractComponent componentMock = this.createComponentMock(acceptorCoreSize, acceptorStopMaxWait, 40, 250,
                10, 10, 100, 100, 300);

        final JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        JBIAcceptorManager acceptorMgr = new JBIAcceptorManager(componentMock, procMgr,
                new RuntimeConfigurationNotifier(componentMock, null, null), LOG);

        acceptorMgr.start();

        assertThreadGroupIsPresent();

        final long start = System.currentTimeMillis();

        acceptorMgr.stop();

        final long duration = System.currentTimeMillis() - start;
        // let's give it 10% more for the ending time
        final long max = (long) ((acceptorCoreSize * acceptorStopMaxWait) * 1.1);
        TestCase.assertTrue("The stop duration (" + duration + ")the acceptor pool is higher than the default value ("
                + max + "ms).", duration <= max);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIAcceptorManager> refAcceptorMgr = new PhantomReference<JBIAcceptorManager>(
                acceptorMgr, this.queueGC);

        acceptorMgr = null;
        this.assertIsGarbaged(refAcceptorMgr);
        this.assertAcceptorThreadGroupIsFreed();

    }
}
