/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.process;

import java.lang.ref.PhantomReference;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import org.junit.Ignore;
import org.junit.Test;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;

import junit.framework.TestCase;

public class JBIProcessorManagerTest extends AbstractJBIManager<JBIProcessorManager> {
    
    private final static Logger LOG = Logger.getLogger(JBIProcessorManagerTest.class.getName());

    private void assertThreadGroupIsPresent() throws Exception {
        super.assertThreadGroupIsPresent(JBIProcessorManager.GROUP_SUFFIX);
    }

    private void assertAcceptorThreadGroupIsFreed() throws Exception {
        super.assertThreadGroupIsFreed(JBIProcessorManager.GROUP_SUFFIX);
    }

    /**
     * Checks the start against valid configuration parameters of processor core
     * pool. Expected result: configuration parameter values are used
     */
    @Test
    public void testStart_000() throws Exception {

        final int acceptorCoreSize = 10;
        final long acceptorStopMaxWait = 500l;
        final int acceptorRetryNb = 15;
        final long acceptorRetryWait = 750;
        final int processorCoreSize = 10;
        final int processorMaxSize = 50;
        final int processorKeepAliveTime = 250;
        final long processorStopMaxWait = 670l;

        final AbstractComponent componentMock = this.createComponentMock(acceptorCoreSize, acceptorStopMaxWait,
                acceptorRetryNb, acceptorRetryWait, processorCoreSize, processorMaxSize, processorKeepAliveTime,
                processorStopMaxWait, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals("The processor core pool size is not set to the configured value.",
                processorCoreSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals("The processor max pool size is not set to the configured value.",
                processorMaxSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());
        TestCase.assertEquals("The processor keep alive time is not set to the configured value.",
                processorKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));
        TestCase.assertEquals("The processor retry policy max number is not set to the configured value.",
                acceptorRetryNb, procMgr.getRetryMaxTry());
        TestCase.assertEquals("The acceptor retry policy base duration is not set to the configured value.",
                acceptorRetryWait, procMgr.getRetryBaseDuration());
        TestCase.assertEquals("The acceptor stop max wait is not set to the configured value.", 
                processorStopMaxWait, procMgr.getProcessorStopMaxWait());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }

    /**
     * Checks the start against a value of processor core pool size higher than
     * the max pool size. Expected result: the core pool size is used as max
     * pool size.
     */
    @Test
    public void testStart_001() throws Exception {

        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, 10, 5, 100, 500, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);
        
        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals("The processor max pool size is not set to the core pool size.",
                procMgr.getProcessorThreadPool().getCorePoolSize(), procMgr.getProcessorThreadPool()
                        .getMaximumPoolSize());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }

    /**
     * Checks the stop against a value of 'processor-stop-max-wait' not set.
     * Expected result: no error thrown and the duration of stop of the
     * JBIProcessorManager is lower than the default value (15s).
     */
    @Test
    public void testStop_001() throws Exception {

        final long processorStopMaxWait = 100;
        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, 10, 10, 100,
                processorStopMaxWait, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        final long start = System.currentTimeMillis();

        procMgr.stop();

        final long end = System.currentTimeMillis() - start;

        // let's give it 10% more for the ending time
        final long max = (long) (processorStopMaxWait * 1.1);
        TestCase.assertTrue("The stop duration of the processor pool (" + end
                + ") is higher than the configured value (" + processorStopMaxWait + ").",
                end <= max);

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);

        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();

    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, increasing the core
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the core size of the message exchange processors thread pool is size
     * to the new value,</li>
     * <li>the min idle number of message exchange processors pool is set to the
     * new value,</li>
     * <li>the number of started threads corresponds to the new value.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetCorePoolSize_000() throws Exception {

        final int originalCorePoolSize = 5;
        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, originalCorePoolSize, 20,
                100, 500, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals(
                "The processor thread pool core size is not set to the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newCorePoolSize = 12;
        procMgr.setCorePoolSize(newCorePoolSize);

        TestCase.assertEquals(
                "The processor thread pool core size is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, decreasing the core
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the core size of the message exchange processors thread pool is sized
     * to the new value,</li>
     * <li>the min idle number of message exchange processors pool is set to the
     * new value,</li>
     * <li>the number of started threads corresponds to the new value.</li>
     * </ul>
     * </p>
     */
    @Ignore("The processor thread pool is not reduce because of JVM bug 6450200, 6576792: http://bugs.sun.com/view_bug.do?bug_id=6576792")
    @Test
    public void testSetCorePoolSize_001() throws Exception {

        final int originalCorePoolSize = 50;
        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, originalCorePoolSize, 55,
                100, 500, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals(
                "The processor thread pool core size is not set to the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not the expected value.",
                originalCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newCorePoolSize = 12;
        procMgr.setCorePoolSize(newCorePoolSize);

        TestCase.assertEquals(
                "The processor thread pool core size is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getCorePoolSize());
        TestCase.assertEquals(
                "The number of running thread of the message exchange processor thread pool is not correctly updated to the new value.",
                newCorePoolSize, procMgr.getProcessorThreadPool().getPoolSize());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, increasing the max
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the max size of the message exchange processors thread pool is size
     * to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetMaxCorePoolSize_000() throws Exception {

        final int originalMaxPoolSize = 50;
        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, 5, originalMaxPoolSize, 100,
                500, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals(
                "The processor thread pool max size is not set to the expected value.",
                originalMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newMaxPoolSize = 62;
        procMgr.setMaxCorePoolSize(newMaxPoolSize);

        TestCase.assertEquals(
                "The processor thread pool max size is not correctly updated to the new value.",
                newMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }

    /**
     * <p>
     * Checks the resize of the message exchange processors, decreasing the max
     * pool size.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the max size of the message exchange processors thread pool is sized
     * to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetMaxCorePoolSize_001() throws Exception {

        final int originalMaxPoolSize = 50;
        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, 5, originalMaxPoolSize, 100,
                500, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals(
                "The processor thread pool max size is not set to the expected value.",
                originalMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newMaxPoolSize = 42;
        procMgr.setMaxCorePoolSize(newMaxPoolSize);

        TestCase.assertEquals(
                "The processor thread pool max size is not correctly updated to the new value.",
                newMaxPoolSize, procMgr.getProcessorThreadPool().getMaximumPoolSize());

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }

    /**
     * <p>
     * Checks the reconfiguration of the message exchange processors, increasing
     * the keep alive time.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the keep alive time of the message exchange processors thread pool is
     * size to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetKeepAliveTime_000() throws Exception {

        final int originalKeepAliveTime = 500;
        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, 5, 10, originalKeepAliveTime,
                500, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not set to the expected value.",
                originalKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newKeepAliveTime = 999;
        procMgr.setPoolKeepAliveTime(newKeepAliveTime);

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not correctly updated to the new value.",
                newKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }

    /**
     * <p>
     * Checks the reconfiguration of the message exchange processors, decreasing
     * the keep alive time.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>the keep alive time of the message exchange processors thread pool is
     * size to the new value,</li>
     * <li>the max number of message exchange processors pool is set to the new
     * value added with the queue size.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testSetKeepAliveTime_001() throws Exception {

        final int originalKeepAliveTime = 500;
        final AbstractComponent componentMock = this.createComponentMock(10, 100, 40, 250, 5, 10, originalKeepAliveTime,
                500, 300);

        JBIProcessorManager procMgr = new JBIProcessorManager(componentMock, new RuntimeConfigurationNotifier(
                componentMock, null, null), LOG);

        procMgr.start();

        assertThreadGroupIsPresent();

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not set to the expected value.",
                originalKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Set a new value to the core pool size of the message exchange
        // processor thread pool
        final int newKeepAliveTime = 249;
        procMgr.setPoolKeepAliveTime(newKeepAliveTime);

        TestCase.assertEquals(
                "The processor thread pool keep alive time is not correctly updated to the new value.",
                newKeepAliveTime, procMgr.getProcessorThreadPool().getKeepAliveTime(TimeUnit.SECONDS));

        // Check the garbage collection of JBIProcessorManager
        final PhantomReference<JBIProcessorManager> refProcPgr = new PhantomReference<JBIProcessorManager>(
                procMgr, this.queueGC);
        procMgr.stop();
        procMgr = null;
        assertIsGarbaged(refProcPgr);
        assertAcceptorThreadGroupIsFreed();
    }
}
