/**
 * Copyright (c) 2010-2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.io.StringReader;
import java.util.Arrays;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.XMLComparator;
import com.ebmwebsourcing.easycommons.xml.XMLPrettyPrinter;

/**
 * Test MTOMUtil method.
 * 
 * TODO: testing method getMtomMapping(Exchange exchange, Element element) is
 *       not possible actually because we haven't a dependency on cdk-core.
 * @author Mathieu Carrolle - EBM WebSourcing
 */
public class MtomUtilTest {
    final static String expectedXopIncludeMessage = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?><xop:Include href=\"cid:contentId\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/>";

    final static String expectedSimpleMtomStructure1 = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?><file xmlns=\"http://petals.ow2.org\"><xop:Include href=\"cid:contentID1\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/></file>";

    final static String expectedSimpleMtomStructure2 = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?><rootNode><file xmlns=\"http://petals.ow2.org\"><xop:Include href=\"cid:contentID1\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/></file><file xmlns=\"http://petals.ow2.org\"><xop:Include href=\"cid:contentID2\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/></file></rootNode>";

    final static String expectedCustomMtomStructure1 = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?><attachments xmlns=\"http://petals.ow2.org\"><customElement xmlns=\"http://custom.ns\"><xop:Include href=\"cid:contentID1\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/></customElement></attachments>";

    final static String expectedCustomMtomStructure2 = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?><attachments xmlns=\"http://petals.ow2.org\"><customElement xmlns=\"http://custom.ns\"><xop:Include href=\"cid:contentID1\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/></customElement><customElement xmlns=\"http://custom.ns\"><xop:Include href=\"cid:contentID2\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/></customElement></attachments>";

    Document document;

    static final DocumentBuilder builder;

    static {
        try {
            builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
    }

    @Before
    public void setUp() throws Exception {
        this.document = builder.newDocument();
    }

    @After
    public void tearDown() throws Exception {
        this.document = null;
    }

    @Test
 	public void testGenerateSimpleMtomStructure1() throws TransformerException {
        final DocumentFragment attchElt1 = MtomUtil.generateMtomStructure(document,
                Arrays.asList("contentID1"), new QName("http://petals.ow2.org", "file"));
        assertNotNull(attchElt1);
        this.document.appendChild(attchElt1);
        assertTrue(
                "Check expected simple mtom structure 1",
                XMLComparator.isEquivalent(expectedSimpleMtomStructure1,
                        XMLPrettyPrinter.prettyPrint(document)));
    }

    @Test
 	public void testGenerateSimpleMtomStructure2() {
        final DocumentFragment attchElt1 = MtomUtil.generateMtomStructure(document, Arrays.asList(
                "contentID1", "contentID2"), new QName("http://petals.ow2.org", "file"));
        assertNotNull(attchElt1);
        final Node rootNode = this.document.appendChild(this.document.createElement("rootNode"));
        rootNode.appendChild(attchElt1);

        assertTrue(
                "Check expected simple mtom structure 2",
                XMLComparator.isEquivalent(expectedSimpleMtomStructure2,
                        XMLPrettyPrinter.prettyPrint(document)));
    }

    @Test
 	public void testGenerateComplexMtomStructure1() throws TransformerException {
        Node attchElt1 = MtomUtil.generateMtomStructure(document, Arrays.asList("contentID1"),
                new QName("http://petals.ow2.org", "attachments"), new QName("http://custom.ns",
                        "customElement"));
        assertNotNull(attchElt1);
        this.document.appendChild(attchElt1);
        assertTrue(
                "Check Expected custom mtom structure 1",
                XMLComparator.isEquivalent(expectedCustomMtomStructure1,
                        XMLPrettyPrinter.prettyPrint(document)));
    }

    @Test
 	public void testGenerateComplexMtomStructure2() throws TransformerException {
        Node attchElt1 = MtomUtil.generateMtomStructure(document, Arrays.asList("contentID1",
                "contentID2"), new QName("http://petals.ow2.org", "attachments"), new QName(
                "http://custom.ns", "customElement"));
        assertNotNull(attchElt1);
        this.document.appendChild(attchElt1);
        assertTrue(
                "Check expected custom mtom structure 2",
                XMLComparator.isEquivalent(expectedCustomMtomStructure2,
                        XMLPrettyPrinter.prettyPrint(document)));
    }

    @Test
 	public void testGenerateMtomXopElement() throws TransformerException {
        Node includeNode = MtomUtil.generateMtomXopElement(document, "contentId");
        assertNotNull(includeNode);
        this.document.appendChild(includeNode);
        assertTrue(XMLComparator.isEquivalent(expectedXopIncludeMessage,
                XMLPrettyPrinter.prettyPrint(document)));
    }

    /**
     * <p>
     * Check the return of {@link MtomUtil#extractAttachmentsIdFromPayload(org.w3c.dom.Element)} from a payload containg
     * MTOM reference.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>No error occurs,</li>
     * <li>MTOM references are returned</li>
     * </ul>
     * </p>
     */
    @Test
    public void extractAttachmentsIdFromPayload() throws SAXException, IOException {
        
        final String expectedId1 = "contentID1";
        final String expectedId2 = "contentID2";
        final StringBuilder strBuilder = new StringBuilder();
        strBuilder.append("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>");
        strBuilder.append("<attachments xmlns=\"http://petals.ow2.org\">");
        strBuilder.append("<customElement xmlns=\"http://custom.ns\">");
        strBuilder.append("<xop:Include href=\"cid:" + expectedId1
                + "\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/>");
        strBuilder.append("</customElement>");
        strBuilder.append("<customElement xmlns=\"http://custom.ns\">");
        strBuilder.append("<xop:Include href=\"cid:" + expectedId2
                + "\" xmlns:xop=\"http://www.w3.org/2004/08/xop/include\"/>");
        strBuilder.append("</customElement>");
        strBuilder.append("</attachments>");
        
        final DocumentBuilder docBuilder = DocumentBuilders.takeDocumentBuilder();
        try {
            final StringReader sr = new StringReader(strBuilder.toString());
            try {
                final Document doc = docBuilder.parse(new InputSource(sr));
                
                final Set<String> identifiers = MtomUtil.extractAttachmentsIdFromPayload(doc.getDocumentElement());
                assertEquals(2, identifiers.size());
                assertTrue(identifiers.contains(expectedId1));
                assertTrue(identifiers.contains(expectedId2));
                
            } finally {
                sr.close();
            }
        } finally {
            DocumentBuilders.releaseDocumentBuilder(docBuilder);
        }
    }

    /**
     * <p>
     * Check the return of {@link MtomUtil#extractAttachmentsIdFromPayload(org.w3c.dom.Element)} from a payload containg
     * no MTOM reference.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>No error occurs,</li>
     * <li>no MTOM reference is returned</li>
     * </ul>
     * </p>
     */
    @Test
    public void extractAttachmentsIdFromPayload_NoMtomRef() throws SAXException, IOException {

        final StringBuilder strBuilder = new StringBuilder();
        strBuilder.append("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>");
        strBuilder.append("<attachments xmlns=\"http://petals.ow2.org\"/>");

        final DocumentBuilder docBuilder = DocumentBuilders.takeDocumentBuilder();
        try {
            final StringReader sr = new StringReader(strBuilder.toString());
            try {
                final Document doc = docBuilder.parse(new InputSource(sr));

                final Set<String> identifiers = MtomUtil.extractAttachmentsIdFromPayload(doc.getDocumentElement());
                assertEquals(0, identifiers.size());

            } finally {
                sr.close();
            }
        } finally {
            DocumentBuilders.releaseDocumentBuilder(docBuilder);
        }
    }
}
